%% -*- erlang-indent-level: 2 -*-

-module(hipe_hot_cold).

-export([finalize/3,get_hotness/1,split_constants/1]).


get_hotness(MFA) ->
  {-1,[]}.
%%    {0.5,absolute_to_rel(hipe_profile:read_counters(MFA))}.


finalize(CFG,Hotness,Options) ->
  OptCFG = case property_lists:get_bool(sparc_peephole,Options) of
	     true  -> hipe_sparc_cfg:remove_dead_code(CFG);
	     false -> CFG
	   end,

  LinearCode = opt(hipe_sparc:sparc_code(hipe_sparc_cfg:linearize(OptCFG))),
  AllHotCode = fill_delay(LinearCode, Options),
  [{hot,lists:flatten(AllHotCode)},[{cold,[]}]].


opt(Code) ->
  lists:reverse(opt(Code,[])).

opt([G,L|Code], Acc) ->
  case hipe_sparc:is_goto(G) of
    true ->
      case hipe_sparc:is_label(L) of
	true ->
	  case hipe_sparc:goto_label(G) =:= hipe_sparc:label_name(L) of
	    true ->
	      opt(Code,[L|Acc]);
	    false ->
	      opt(Code,[L,G|Acc])
	  end;
	false ->
	  opt([L|Code], [G|Acc])
      end;
    false ->
      opt([L|Code], [G|Acc])
  end;
opt([I],Acc) ->
  [I|Acc];
opt([],Acc) ->
  Acc.

%%
%% Code is a list of instructions. 
%%

fill_delay(Code, Options) ->
  Code0 = 
    case property_lists:get_bool(sparc_peephole,Options) of
      true  -> peephole(Code);
      false -> Code
    end,
  Codes =  split_at_branch(Code0),
  case property_lists:get_bool(fill_delayslot,Options) of
    true  -> lists:map(fun fill_delay0/1, Codes);
    false -> lists:map(fun nofill_delay/1, Codes)
  end.

nofill_delay(Code) ->
  [Code | [hipe_sparc:nop_create([])]].

%%
%% Code is a list of instructions where a branch/jump 
%%

fill_delay0(Code) ->
  case catch find_delay(Code) of
    no_branch ->
      Code;
    {NewCode, _, _, _} ->
      [NewCode | [hipe_sparc:nop_create([])]];
    {NewCode, Delay} ->
      [NewCode | [Delay]]
  end.

%%
%% Extracts a delay instruction from a list 
%%

find_delay([Jmp]) ->
  case hipe_sparc:is_any_branch(Jmp) of
    true ->
      {[Jmp], 
       ordsets:from_list(hipe_sparc:uses(Jmp)), 
       ordsets:from_list(hipe_sparc:defines(Jmp)),
       []};
    false ->
      throw(no_branch)
  end;
find_delay([I|Is]) ->
  case find_delay(Is) of
    {NewIs, Uses, Defs, Loads} ->
      IUses = ordsets:from_list(hipe_sparc:uses(I)),
      IDefs = ordsets:from_list(hipe_sparc:defines(I)),
      NewUses = ordsets:union(Uses, IUses),
      NewDefs = ordsets:union(Defs, IDefs),
      IsStore = hipe_sparc:is_store(I),
      NewLoads = case hipe_sparc:is_load(I) of 
		   true -> [I|Loads];
		   false -> Loads
		 end,
      case is_delay_instr(I) of
	true ->
	  %% Make sure the instruction isn't defining a reg that is 
	  %% used later or uses a reg that is defined later or
	  %% defines a reg that is defined later
	  X = {ordsets:intersection(Uses, IDefs), 
	       ordsets:intersection(Defs, IUses),
	       ordsets:intersection(Defs, IDefs)},
	  case X of
	    {[], [], []} ->  %% No register conflicts
	      case {IsStore,Loads} of
		{_,[]} -> % No following loads
		  {NewIs, I}; % found a delay instr.
		{false,_} -> % This is not a store
		  {NewIs, I}; % found a delay instr.
		_ -> %% A store with following loads.
		  %% XXX: TODO check whether the store 
		  %%      and the load REALY conflicts.
		  %%      (Stack and Heap does not conflict)
		  {[I|NewIs], NewUses, NewDefs, NewLoads} %% Search more
	      end;
	    _ -> %% Register conflicts.
	      {[I|NewIs], NewUses, NewDefs, NewLoads}
	  end;
	false ->
	  {[I|NewIs], NewUses, NewDefs, NewLoads}
      end;
    {NewIs, Delay} ->
      {[I|NewIs], Delay}
  end.

%%
%% true if I is an instruction that can be moved to a delay slot
%%

is_delay_instr(I) ->
  case hipe_sparc:type(I) of
    label -> false;
    comment -> false;
    load_address -> false;
    load_atom -> false;
    load_word_index -> false;
    %% (Happi) Tests have indicated that puting loads 
    %%         in the delayslot can slow down code...
    %%         ... but it can also speed up code.
    %%         the impact is about 10 - 20 % on small bms
    %%         on the average you loose 1-2 % by not putting
    %%         loads in delayslots
    %% load -> false;
    _ -> true
  end.

%%
%% Split a list of instructions to a list of lists of instructions
%% Where each sublist ends with a branch.
%%

split_at_branch([]) ->
   [];
split_at_branch([I|Rest]) ->
  case Rest of
    [L|Is] ->
      case hipe_sparc:is_label(L) of
	true -> 
	  Lname = hipe_sparc:label_name(L),
	  split(I,Lname,Rest);
	false ->
	  split(I,[],Rest)
      end;
    _ ->
      split(I,[],Rest)
  end.

split(I,Lname,Is)->
  case hipe_sparc:type(I) of
    b -> 
      case hipe_sparc:b_false_label(I) of
	Lname ->
	  [[hipe_sparc:b_false_label_update(I,[])] |
	   split_at_branch(Is)];
	FL ->
	  [[hipe_sparc:b_false_label_update(I,[])], 
	   [hipe_sparc:goto_create(FL,[fallthrough])]|
	   split_at_branch(Is)]
      end;
    br -> 
      case hipe_sparc:br_false_label(I) of
	Lname ->
	  [[hipe_sparc:br_false_label_update(I,[])] |
	   split_at_branch(Is)];
	FL ->
	  [[hipe_sparc:br_false_label_update(I,[])], 
	   [hipe_sparc:goto_create(FL,[fallthrough])]|
	   split_at_branch(Is)]
      end;
    goto -> 
      [[I] | split_at_branch(Is)];
    jmp -> 
      [[I] | split_at_branch(Is)];
    jmp_link -> 
      [[I] | split_at_branch(Is)];
    call_link ->
      case hipe_sparc:call_link_continuation(I) of
	[] -> 
	  [Same|Lists] = split_at_branch(Is),
	  [[I|Same]|Lists];
	Lname -> 
	  [[hipe_sparc:call_link_continuation_update(I,[])]|
	   split_at_branch(Is)];
	CCL ->
	  [[hipe_sparc:call_link_continuation_update(I,[])],
	   [hipe_sparc:goto_create(CCL,[fallthrough])]|
	   split_at_branch(Is)]
      end;
    label ->
      [[I] | split_at_branch(Is)];
    _ -> 
      [Same|Lists] = split_at_branch(Is),
      [[I|Same]|Lists]
  end.

%%
%% 
%%

peephole([]) ->
  [];
peephole([I|Is]) ->
   case hipe_sparc:type(I) of
     move ->
       case hipe_sparc:move_src(I) =:= hipe_sparc:move_dest(I) of
	 true ->
	   peephole(Is);
	 false ->
	   [I | peephole(Is)]
       end;
      _ ->
       [I | peephole(Is)]
   end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Replaces big immediates with registers that are defined with
%% a 'sethi' and an 'or'. 
%%
%% NEWSFLASH! This also checks if arg1 to an alu operation is an
%% immediate and rectifies that situation.
%%

split_constants(CFG) ->
   Labels = hipe_sparc_cfg:labels(CFG),
   {Low, High} = hipe_sparc_cfg:var_range(CFG),
   hipe_gensym:set_var(sparc,High),
   NewCFG = split_bbs(Labels, CFG),
   hipe_sparc_cfg:var_range_update(NewCFG, {Low, hipe_gensym:get_var(sparc)}).


split_bbs([], CFG) ->
  CFG;
split_bbs([Lbl|Lbls], CFG) ->
  BB = hipe_sparc_cfg:bb(CFG, Lbl),
  Code = hipe_bb:code(BB),
  case split_instrs(Code, [], unchanged) of
    unchanged ->
      split_bbs(Lbls, CFG);
    NewCode ->
      NewCFG = hipe_sparc_cfg:bb_update(CFG, Lbl, hipe_bb:code_update(BB, NewCode)),
      split_bbs(Lbls, NewCFG)
  end.

split_instrs([], RevCode, unchanged) ->
  unchanged;
split_instrs([], RevCode, changed) ->
  lists:reverse(RevCode);
split_instrs([I|Is], RevCode, Status) ->
  case fuck_left_immediate(I) of
    unchanged ->
      case split_instr(I) of
	unchanged ->
	  split_instrs(Is, [I|RevCode], Status);
	NewCode ->
	  split_instrs(Is, NewCode++RevCode, changed)
      end;
    NewCode ->
      split_instrs(NewCode++Is, RevCode, changed)
  end.

fuck_left_immediate(I) ->
  case {hipe_sparc:is_alu(I), hipe_sparc:is_alu_cc(I)} of
    {true, _} ->
      Src1 = hipe_sparc:alu_src1(I),
      case hipe_sparc:is_imm(Src1) of
	true ->
	  case is_commutative(hipe_sparc:alu_operator(I)) of
	    true ->
	      Src2 = hipe_sparc:alu_src2(I),
	      I0 = hipe_sparc:alu_src1_update(I, Src2),
	      [hipe_sparc:alu_src2_update(I0, Src1)];
	    false ->
	      Tmp = hipe_sparc:mk_new_reg(),
	      Mov = hipe_sparc:move_create(Tmp, Src1, []),
	      NewI = hipe_sparc:alu_cc_src1_update(I, Tmp),
	      [Mov, NewI]
	  end;
	false ->
	  unchanged
      end;
    {_, true} ->
      Src1 = hipe_sparc:alu_cc_src1(I),
      case hipe_sparc:is_imm(Src1) of
	true ->
	  case is_commutative(hipe_sparc:alu_cc_operator(I)) of
	    true ->
	      Src2 = hipe_sparc:alu_cc_src2(I),
	      I0 = hipe_sparc:alu_cc_src1_update(I, Src2),
	      [hipe_sparc:alu_cc_src2_update(I0, Src1)];
	    false ->
	      Tmp = hipe_sparc:mk_new_reg(),
	      Mov = hipe_sparc:move_create(Tmp, Src1, []),
	      NewI = hipe_sparc:alu_cc_src1_update(I, Tmp),
	      [Mov, NewI]
	  end;
	false ->
	  unchanged
      end;
    {false, false} ->
      unchanged
  end.

is_commutative(Op) ->
  case Op of
    '+' -> true;
    'or' -> true;
    'and' -> true;
    'xor' -> true;
    _ -> false
  end.

split_instr(I) ->
  Uses = hipe_sparc:imm_uses(I),
  case big_constants(Uses) of
    [] -> unchanged;
    {Code, Subst} -> [hipe_sparc:subst(I, Subst) | Code]
  end.

big_constants([]) ->
  {[], []};
big_constants([V|Vs]) ->
  C = hipe_sparc:imm_value(V),
  case is_big(C) of
    true ->
      NewVar = hipe_sparc:mk_new_reg(),
      Low = low10(C),
      Code =
	if Low =:= 0 ->
	    [hipe_sparc:sethi_create(
	       NewVar,
	       hipe_sparc:mk_imm(high22(C)),[])];
	   true ->	     
	    [hipe_sparc:alu_create(NewVar, NewVar, 'or', 
				   hipe_sparc:mk_imm(Low), []),
	     hipe_sparc:sethi_create(NewVar,hipe_sparc:mk_imm(high22(C)),[])]
	end,
      {MoreCode, MoreSubst} = big_constants(Vs),
      {Code++MoreCode, [{V, NewVar} | MoreSubst]};
    false ->
      big_constants(Vs)
  end.

is_big(X) ->
  if X > 16#fff ->
      true;
     true ->
      false
  end.

high22(X) -> X bsr 10.
low10(X) -> X band 16#3ff.


