%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Copyright (c) 2001 by Erik Johansson.  All Rights Reserved 
%% -*- erlang-indent-level: 2 -*-
%% ====================================================================
%%  Filename : 	hipe_sparc_caller_saves.erl
%%  Module   :	hipe_sparc_caller_saves
%%  Purpose  :  To add save and restore code for caller save regs
%%              at call sites.
%%  Notes    :  LastSpillPos is a bad name in this module,
%%               since it denotes the postion after the last used
%%               position.
%%  History  :	* 2001-11-16 Erik Johansson (happi@csd.uu.se): 
%%               Created.
%%  CVS      :
%%              $Author: happi $
%%              $Date: 2001/12/17 20:56:19 $
%%              $Revision: 1.4 $
%% ====================================================================
%%  Exports  :
%%
%%  Description:
%%
%%   1. Calculate liveness. 
%%      (Consider reusing the liveness calculated by the regalloc.)
%%
%%   2. Go through each basic block bottom up.
%%      For each call c:
%%         For each live out temp t. (that is not a retval of c)
%%             If t is spilled: Do nothing.
%%             If t is not spilled:
%%                a) Get a spillposition p for t.
%%                b) Add code for spilling t to p before the call.
%%                c) add code for unspilling t from p after the call.
%%                d) If any spillcode was added rewrite the call
%%                     If there is an exception handler: add the unspill
%%                       code to both branches of the call.
%%
%%   3. Optimization pass: Place the spillcode "optimally".
%%                        Implemented in hipe_sparc_opt_frame
%%
%%  TODO: Either make this module completely target independent or
%%        completely target dependent.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-module(hipe_sparc_caller_saves).
-export([rewrite/4]).
-include("../main/hipe.hrl").
-include("../rtl/hipe_literals.hrl").

%% XXX: Make target independent.
-define(TARGET, hipe_sparc_specific).

%% XXX: Make target independent.
rewrite(Cfg, TempMap, LastSpillPos, Options) ->
  %% hipe_sparc_cfg:pp(Cfg),  
  ?opt_start_timer("Caller Saves"),
  %% Get Info.
  Liveness = liveness(Cfg, ?TARGET),
  State = new_state(Liveness,  Cfg, TempMap, LastSpillPos),
  BBs = hipe_sparc_cfg:predictionorder(Cfg),

  %% Rewrite.
  {NewCode, NextPos} = traverse(BBs, State),

  %% Make new CFG.
  Sparc0 = hipe_sparc_cfg:linearize(Cfg),
  Sparc1 = hipe_sparc:sparc_code_update(Sparc0, NewCode),
  Sparc2 = hipe_sparc:sparc_var_range_update(
	     Sparc1, {0,hipe_gensym:get_var(sparc)}),
  Sparc3 = hipe_sparc:sparc_label_range_update(
	     Sparc2, {0,hipe_gensym:get_label(sparc)}),  
  Cfg1 = hipe_sparc_cfg:init(Sparc3),

  %% Minimize the number of loads and stores.
  ?opt_stop_timer("Caller Saves"),
  ?opt_start_timer("Opt Frame"),
  Cfg2 =   hipe_sparc_opt_frame:cfg(Cfg1),
  ?opt_stop_timer("Opt Frame"),
  {Cfg2, NextPos}.

%% Go through each BasicBlock
traverse(BBs, State) ->
  State1 = lists:foldl(fun handle_bb/2,State,BBs),  
  {instrs(State1),spill_index(State1)}.
  
%% For each BasicBlock go through the code bottom-up.
handle_bb(BB,State) ->
  %% io:format("--------------\n",[]),
  Code =  hipe_bb:code(bb(cfg(State),BB, ?TARGET)),
  State1 = enter_bb(BB, State),
  State2 = bottom_up(Code, State1),
  %% io:format("-- Block ~w --\n",[BB]),
  add_instrs(hipe_sparc:label_create(BB,[]), State2).

%% Go through each instruction from the last to the first.
%% (Collect liveness from live out and backwards.)
bottom_up(Is, State) ->
  lists:foldr(fun handle/2,State,Is).

%% For each instruction rewrite calls so that live temps are saved
handle(I,State) ->

  
  Defs = defines(I, ?TARGET),
  Uses = uses(I, ?TARGET),
  LiveOut = live(State)-- hipe_sparc_registers:global(),
  Live = (LiveOut--Defs), %% Remove killed
  LiveIn = ordsets:union(ordsets:from_list(Live),ordsets:from_list(Uses))
    -- hipe_sparc_registers:global(),
    %% io:format("~nLiveOut: ~w~n",[LiveOut-- hipe_sparc_registers:global()]),
    %% hipe_sparc_pp:pp_instr(I),
    %% io:format("LiveIn: ~w~n",[LiveIn]),


  %% Add save and restores if neccessary.
  State1 = update_call(I,State,Live, LiveOut),

  %% Update the liveness information in the state.
  State2 = update_live(LiveIn, State1).

%% XXX: Make this target independent.
update_call(I,State,Live,LiveOut) ->
  case hipe_sparc:type(I) of
    call_link ->
	case live_in_regs(Live, State) of
	  [] -> 
	    %% All live temps are allready on the stack.
	    %% No need for extra saving.
	    %% Just add the live-info to the stack descriptor 
	    %%  of the call.
	    add_instrs(set_live_slots(I,State,Live),State);
	  
	  ToSpill ->
	    %% We have at least one live temp that is not on the stack
	    handle_call(I,State,ToSpill,Live)
	end;

    _ ->
      %% This is not a call, nothing to save.
      %% Just add the instruction to the state.
      add_instrs(I,State)
  end.



%% Ugly function that finds the stackslots of all
%% live temps and adds these to the stack descriptor
%% of the call. 
set_live_slots(Call,State,LiveOut) ->
  Live =
    [stack_pos(T,State) || T <- LiveOut],
  SD = hipe_sparc:call_link_stack_desc(Call),
  NewSD = hipe_sparc_stack_descriptors:set_live(SD, Live),
  hipe_sparc:call_link_stack_desc_update(Call,NewSD).

%% Helper function to set_live_slots.
%% Maps a Temp to its stack pos.
stack_pos(T,State) ->
  TempMap = tempmap(State),
  case hipe_temp_map:find(T, TempMap) of
    {spill, P} ->
      P;
    _ -> 
      element(1,get_pos(T,extramap(State)))
  end.
	      
%% XXX: Make target independent.
live_in_regs(Live, State) ->

  TempMap = tempmap(State),
  [T || T <- Live, not_spilled(T, TempMap)]
    -- hipe_sparc_registers:global().




handle_call(I,State,ToSpill,LiveOut)->
  %% a) Get a spillposition p for t.
  %% b) Add code for spilling t to p before the call.
  %% c) add code for unspilling t from p after the call.
  %% d) If any spillcode was added rewrite the call
  %%      If there is an exception handler: add the unspill
  %%        code to both branches of the call.


  ExtraMap = extramap(State),
  {SpillPositions, NewExtraMap} =
    get_spill_positions(ToSpill, ExtraMap),
  {Saves, Restores} = gen_save_restore(SpillPositions),
  ContLab = hipe_sparc:call_link_continuation(I),
  FailLab = hipe_sparc:call_link_fail(I),
  NewContLab = hipe_sparc:label_create_new(),
  GotoCont =  hipe_sparc:goto_create(ContLab,[]),
  NewI =
    hipe_sparc:call_link_continuation_update(
      I,
      hipe_sparc:label_name(NewContLab)),

  
  State1 = update_extramap(NewExtraMap, State),

  %% Update stack descriptor
  NewCall = set_live_slots(NewI,State1,LiveOut),

  NewCode = 
    case FailLab of
      [] ->
	[Saves, 
	 NewCall,
	 NewContLab,
	 Restores,
	 GotoCont];
      _ ->
	NewFailLab = hipe_sparc:label_create_new(),
	GotoFail =  hipe_sparc:goto_create(FailLab,[]),
	NewI2 = 
	  hipe_sparc:call_link_fail_update(
	    NewCall,
	    hipe_sparc:label_name(NewFailLab)),
	[Saves, 
	 NewI2,
	 NewContLab,
	 Restores,
	 GotoCont,
	 NewFailLab,
	 Restores,
	 GotoFail]
    end,

  add_instrs(NewCode,State1).




not_spilled(Temp, Map) ->
  not hipe_temp_map:is_spilled(Temp, Map).


get_spill_positions([T|Temps], ExtraMap) ->
  {Pos, NewMap} = get_pos(T,ExtraMap),
  {Positions, Map} = get_spill_positions(Temps, NewMap),
  {[{T,Pos}|Positions], Map};
get_spill_positions([],Map) ->
  {[],Map}.



gen_save_restore([{T,Pos}|Positions]) ->
  SparcTemp = hipe_sparc:mk_reg(T),
  SparcPos = hipe_sparc:mk_imm(Pos),
  {Saves, Restores} = gen_save_restore(Positions),
  Save = hipe_sparc:pseudo_spill_create(SparcTemp,SparcPos),
  Restore = hipe_sparc:pseudo_unspill_create(SparcTemp,SparcPos),
  {[Save | Saves], [Restore | Restores]};
gen_save_restore([]) -> {[],[]}.



%% ================================================================= %%
%%                             A  D  T s                             %%
%% ================================================================= %%


%% ================================================================= %%
%% The ExtraMap structure
%% Maps unspilled temps to spillpositions.
%% 

%% ____________________________________________________________________
%%  new_extramap(LastSpillPos)    
%% Returns: 	An empty ExtraMap.
%% Arguments:	LastSpillPos - The next spill will be to this pos.
%% Description:	Creates an empty ExtraMap 
%% ____________________________________________________________________
new_extramap(LastSpillPos) ->
  {LastSpillPos, empty()}.

index(Map) ->
  %% io:format("Map: ~w\n",[Map]),
  element(1,Map).

%% ____________________________________________________________________
%%  get_pos(T,ExtraMap)    
%% Returns: 	
%% Arguments:	
%% Description:	Finds the spill position of the temp T,
%%              If T is not in the map T will be added. 
%% ____________________________________________________________________
get_pos(T,{Next,Map}) ->
  case lookup(T,Map) of
    none ->
      {Next, {Next+1,insert(T,Next,Map)}};
    {value, Pos} ->
      {Pos, {Next,Map}}
  end.


%% ____________________________________________________________________
%% 
%% The Map kernel.


empty() ->
  gb_trees:empty().
lookup(T,Map) ->
  gb_trees:lookup(T,Map).
insert(T,Next,Map) ->
  gb_trees:insert(T,Next,Map).



%% ================================================================= %%
%%
%% The state structure
%%
%% ================================================================= %%

%% ____________________________________________________________________
%% The state record.
-record(state, {instrs, extramap, live, liveness, cfg, tempmap}).

%% Create a new state.
new_state(Liveness,  Cfg, TempMap, LastSpillPos) ->
  #state{instrs=[], 
	 extramap = new_extramap(LastSpillPos), 
	 live = [],
	 liveness=Liveness,  
	 cfg=Cfg,
	 tempmap=TempMap}.  

%% ____________________________________________________________________
%% Selectors
instrs(State) -> 
  lists:flatten(State#state.instrs).
extramap(State)->
  State#state.extramap.
live(State)->
  State#state.live.
liveness(State)->
  State#state.liveness.
cfg(State)->
  State#state.cfg.
tempmap(State)->
  State#state.tempmap.
spill_index(State) ->
  index(State#state.extramap).

%% ____________________________________________________________________
%% Updates

enter_bb(BB, State) ->
  State#state{live=ordsets:from_list(
		     liveout(
		       liveness(State), BB, hipe_sparc_specific))}.
  
add_instrs(I,State) ->
  State#state{instrs=[I|State#state.instrs]}.

update_extramap(NewExtraMap, State) ->
  State#state{extramap=NewExtraMap}.

update_live(NewLive, State) ->
  State#state{live=NewLive}.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Interface to target specific external functions.
%% XXX: Make this efficient somehow...
%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

liveness(CFG, Target) ->
   Target:analyze(CFG).

bb(CFG, L, Target) ->
  Target:bb(CFG,L).

livein(Liveness,L, Target) ->
  regnames(Target:livein(Liveness,L), Target).

liveout(Liveness,L, Target)->
  regnames(Target:liveout(Liveness,L), Target).

uses(I, Target)->
  regnames(Target:uses(I), Target).

defines(I, Target) ->
  regnames(Target:defines(I), Target).

regname(Reg, Target) ->
  Target:reg_nr(Reg).

regnames(Regs, Target) ->
  [Target:reg_nr(X) || X <- Regs]. 
