%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Copyright (c) 2001 by Erik Johansson.  All Rights Reserved 
%% -*- erlang-indent-level: 2 -*-
%% ====================================================================
%%  Filename : 	hipe_sparc_frame.erl
%%  Module   :	hipe_sparc_frame
%%  Purpose  :  
%%  Notes    :  
%%              The code may only have spilled registers in
%%              the pseudo instructions: 
%%
%%                pseudo_spill   t1, t2 ; spills t1 to spillpos of t2
%%                pseudo_unspill t1, t2 ; loads t1 from spillpos of t2 
%%
%%              The TempMap must contain mappings from all used temps
%%              to physical registers or spill positions.   
%%
%%              XXX: New immediates are introduced as stack offsets.
%%                   If more than 1024 stack slots are needed,
%%                   the immediate offsets will be to big for a
%%                   sparc instruction.
%%                   This can happen with the naive allocator.
%%                   A special pass called from hipe_main 
%%                   rectifies this situation.
%% 
%%  History  :	* 2001-10-25 Erik Johansson (happi@csd.uu.se): 
%%               Created.
%%  CVS      :
%%              $Author: happi $
%%              $Date: 2002/02/07 16:26:18 $
%%              $Revision: 1.10 $
%% ====================================================================
%%  Exports  :
%%
%% TODO:
%%  Optimize 
%%  Handle multiple return values
%%
%% Testcases:
%% hipe:c({len,len,2},[pp_rtl,{regalloc,naive}]).
%% hipe:c({len,len,2},[pp_sparc,{regalloc,lfls}]).
%% hipe:c({late,t,0},[pp_sparc,{regalloc,lfls}]).
%% hipe:c({late,seven,7},[pp_sparc,{regalloc,lfls}]).
%% hipe:c({late,test_enter,7},[pp_sparc,{regalloc,lfls}]).
%% hipe:c({trivial_19,test,0},[pp_sparc,{regalloc,lfls}]).
%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-module(hipe_sparc_frame).
-export([frame/4]).
-include("../main/hipe.hrl").
-include("../rtl/hipe_literals.hrl").


frame(Cfg, TempMap, NextSpillPos, Options) ->
  %%  io:format("~p\n",[TempMap]),
  rewrite(Cfg, TempMap, NextSpillPos).


rewrite(Cfg, TempMap, NextSpillPos) ->
  %%  hipe_sparc_cfg:pp(Cfg),

  Sparc = hipe_sparc_cfg:linearize(Cfg),
  {M,F,A} = hipe_sparc:sparc_fun(Sparc),
  Arity = hipe_sparc:sparc_arity(Sparc),
  Code = hipe_sparc:sparc_code(Sparc),

  %% Need is stack positions for spills
  Need = NextSpillPos +1, %% +1 for CP, not realy needed for leaf
                          %%            functions... 

  %% ExtraNeed is room for argumets in calls. 
  ExtraNeed = extra_stack_need(Code,Need,Arity),
  StackNeed = Need + ExtraNeed,

  CP = hipe_sparc:mk_reg(hipe_sparc_registers:return_address()),
  SP = hipe_sparc:mk_reg(hipe_sparc_registers:stack_pointer()),
  NewStart = hipe_sparc:label_create_new(),
  TempReg = hipe_sparc:mk_reg(hipe_sparc_registers:temp0()), 

  {StackTestCode, StackIncCode} =  
    gen_stack_test(StackNeed, SP, CP, Arity, Cfg),
  AllocCode = gen_alloc_code(Need*4, SP),

  EntryCode = 
    [NewStart,
     StackTestCode,  %% (We will flatten the list later.)
     AllocCode,
     gen_save_cp(SP, CP, Need)],

  {RetCode,RetLabel} = gen_ret(TempMap, Need, Arity),
  
  NewCode = 
    [EntryCode,
     %% Rewrite instructions to use physical registers 
     %% instead of temps.
     %% Also, handle pseudo instructions.
     [rewrite_instr(I, TempMap, Need, Arity, RetLabel) || I <- Code],
     RetCode,
     %% code to call increment stack.
     StackIncCode],

  %% Create a new CFG
  NewSparc = hipe_sparc:sparc_code_update(Sparc, lists:flatten(NewCode)),
  CFG0 = hipe_sparc_cfg:init(NewSparc),
  CFG1 = hipe_sparc_cfg:var_range_update(CFG0, {0, 31}),
  %%   hipe_sparc_cfg:pp(CFG1),
  CFG1.


rewrite_instr(I, TempMap, Need, Arity, RetLabel) ->
  %% io:format("\n\n~w -> \n",[I]),
  I2 = rewrite_instr2(I, TempMap, Need, Arity, RetLabel),
  %% io:format("~w\n",[I2]),
  I2.

rewrite_instr2(I, TempMap, Need, Arity, RetLabel) ->

  case hipe_sparc:type(I) of
    pseudo_push ->
      ?EXIT({pseudo_push, not_supportet_any_more,I});
    pseudo_pop ->
      SP = hipe_sparc:mk_reg(hipe_sparc_registers:stack_pointer()),
      Offset = 
	offset(
	  %% Arg position on stack = arg index + need 
	  (hipe_sparc:imm_value(hipe_sparc:pseudo_pop_index(I)) +
	   Need)),
      Temp = hipe_sparc:pseudo_pop_reg(I),
      Reg = map(Temp,TempMap),
      [hipe_sparc:load_create(Reg, uw, SP, Offset, hipe_sparc:info(I))];
    move ->
      rewrite_move(I, TempMap, Need, Arity);
    pseudo_spill ->
      rewrite_spill(I, TempMap, Need, Arity);
    pseudo_unspill ->
      rewrite_unspill(I, TempMap, Need, Arity);

    call_link ->
      rewrite_call(I, TempMap, Need, Arity);

    pseudo_return ->
      %% Note we assume that the arguments to return have been 
      %% mapped to physical register in the translation RTL->SPARC.

      %% XXX: Fix if we start using multiple return values
      %%      Acctually just fix gen_ret/3 to annotate the
      %%      jmp with the live registers...
      case length(hipe_sparc:pseudo_return_regs(I)) == 1 of
	true -> %% Just one retval all is ok
	  hipe_sparc:goto_create(RetLabel,[return]);
	false ->
	  %% Number of retvalues not 1.
	  ?EXIT({multiple_retvals_not_handled,I})
      end;

    pseudo_enter ->
      %% Calculate number of args (to current fun) on stack
      %% Consider moving this out of the loop.
      NoRegArgs = hipe_sparc_registers:register_args(),
      ArgsOnStack = 
	if Arity > NoRegArgs -> Arity - NoRegArgs; 
	   true -> 0 
	end,

      %% Get arguments to enter.
      Args = hipe_sparc:pseudo_enter_args(I), 
      RegArgs = lists:sublist(Args,1,NoRegArgs),
      %% Get target (closure or address).
      {Target,GetTInstr}
	= case hipe_sparc:pseudo_enter_type(I) of
	    closure ->
	      {map(hipe_sparc:pseudo_enter_target(I), TempMap),[]};
	    _ ->
	      TL = hipe_sparc:pseudo_enter_target(I),
	      TR = hipe_sparc:mk_reg(hipe_sparc_registers:temp0()),
	      TI = hipe_sparc:load_address_create(TR,TL,
						  function,[]),
	      {TR,TI}
	  end,

      CP = hipe_sparc:mk_reg(hipe_sparc_registers:return_address()),
      SP = hipe_sparc:mk_reg(hipe_sparc_registers:stack_pointer()),

      %% The code for enter.
      [
       %% Restore CP from stack
       %% TODO: Do this only in non-leaf functions.
       hipe_sparc:load_create(
	 CP, uw, SP, offset(Need), []), 

       %% If arguments to the called function spill over to the 
       %% stack, do some shuffling.
       %% Otherwise just adjust the stackpointer.
       case length(Args) > NoRegArgs of
	 true ->
	   enter_stack_args(Args, Need, TempMap, ArgsOnStack);
	 false ->
	   hipe_sparc:alu_create(
	      SP,SP,'-',hipe_sparc:mk_imm((ArgsOnStack+Need)*4),[remove_frame])
       end,

       %% Jump to the target function.
       GetTInstr,
       hipe_sparc:jmp_create(Target, hipe_sparc:mk_imm(0), 
			     [map(R,TempMap)||R <- RegArgs] , [])];


    _ -> 
      %% All other instructions.
      %% No special handling, just map the temps to regs or spill positions.
      Defs = hipe_sparc:defines(I),
      NewI = remap(remap(I, Defs, TempMap), hipe_sparc:uses(I), TempMap)
  end.


rewrite_call(Call, TempMap, Need, Arity) ->
  I = set_stack_size(Call,Need, Arity),
  %% io:format("Call ~w~n",[I]),
  Defs = hipe_sparc:defines(I),
  NewI = remap(remap(I, Defs, TempMap), hipe_sparc:uses(I), TempMap),
  
  {PushCode, NoArgs} = push_call_args(hipe_sparc:call_link_args(I), TempMap),
  SP = hipe_sparc:mk_reg(hipe_sparc_registers:stack_pointer()),


  case hipe_sparc:call_link_fail(I) of
    [] -> %% Not in a catch, all is well.
      if NoArgs > 0 -> %% Some args must be pushed...
	  [
	   hipe_sparc:alu_create(
	     SP,SP,'+',hipe_sparc:mk_imm(4*NoArgs),[]),
	   PushCode,
	   NewI];
	 true -> %% All args in regs.
	  [NewI]
      end;


    FailL -> 
      Temp =  hipe_sparc:mk_reg(hipe_sparc_registers:temp0()),
      RetLab = hipe_sparc:label_create_new(),
      GotoCont = hipe_sparc:goto_create(
		   hipe_sparc:call_link_continuation(NewI),[]),
      CallI =  hipe_sparc:call_link_continuation_update(
		 NewI, 
		 hipe_sparc:label_name(RetLab)),
      [
       hipe_sparc:alu_create(
	 SP,SP,'+',hipe_sparc:mk_imm((NoArgs)*4),[]),
       PushCode,
       CallI,
       RetLab,
       GotoCont
	   
      ]
  end.


%% ____________________________________________________________________
%%      
%% Returns: 	
%% Arguments:	
%% Description:	 
%% ____________________________________________________________________
%%  hipe:c({test13,test,0},[late_frames,{regalloc,lfls},pp_sparc]).
rewrite_move(I, TempMap, Need, Arity) ->
  Src = hipe_sparc:move_src(I),
  Dest = hipe_sparc:move_dest(I),
  SP = hipe_sparc:mk_reg(hipe_sparc_registers:stack_pointer()),
  DestReg = hipe_sparc:reg_nr(Dest),

  case hipe_sparc:is_reg(Src) of
    false ->
      case hipe_temp_map:is_spilled(DestReg,TempMap) of
	true ->
	  Pos = spill_offset(hipe_temp_map:find(DestReg, TempMap)),
	  Reg = hipe_sparc:mk_reg(hipe_sparc_registers:temp0()), 
	  [hipe_sparc:move_dest_update(I,Reg),
	   hipe_sparc:store_create(
	     SP, Pos,uw,Reg, hipe_sparc:info(I))];
	false ->
	  remap(I, [Dest], TempMap)
      end;
    true ->
      SrcReg = hipe_sparc:reg_nr(Src),

  
      case {hipe_temp_map:is_spilled(SrcReg,TempMap),
	    hipe_temp_map:is_spilled(DestReg,TempMap)} of
	{false,true} -> %% Spill
	  Pos = spill_offset(hipe_temp_map:find(DestReg, TempMap)),
	  Reg = map(Src,TempMap),
	  hipe_sparc:store_create(
	    SP, Pos,uw,Reg, hipe_sparc:info(I));
	{true, false} -> %% unspill
	  Pos = spill_offset(hipe_temp_map:find(SrcReg, TempMap)),
	  Reg = map(Dest,TempMap),
	  hipe_sparc:load_create(
	    Reg, uw, SP, Pos, 
	    hipe_sparc:info(I));
	{false,false} ->
	  remap(I, [Src,Dest], TempMap);
	{true,true} ->
	  Pos1 = spill_offset(hipe_temp_map:find(SrcReg, TempMap)),
	  Pos2 = spill_offset(hipe_temp_map:find(DestReg, TempMap)),
	  Reg = hipe_sparc:mk_reg(hipe_sparc_registers:temp0()), 
	  [hipe_sparc:load_create(
	     Reg, uw, SP, Pos1, 
	     hipe_sparc:info(I)),
	   hipe_sparc:store_create(
	     SP, Pos2,uw,Reg, hipe_sparc:info(I))]
      end
  end.

%% ____________________________________________________________________
%%      
%% Returns: 	
%% Arguments:	
%% Description:	 
%% ____________________________________________________________________
rewrite_spill(I, TempMap, Need, Arity) ->
  SP = hipe_sparc:mk_reg(hipe_sparc_registers:stack_pointer()),
  Reg = map(hipe_sparc:pseudo_spill_reg(I),TempMap),
  Pos = spill_offset(hipe_sparc:pseudo_spill_pos(I)),
  hipe_sparc:store_create(
    SP,Pos,uw,Reg, hipe_sparc:info(I)).

%% ____________________________________________________________________
%%      
%% Returns: 	
%% Arguments:	
%% Description:	 
%% ____________________________________________________________________
rewrite_unspill(I, TempMap, Need, Arity) ->
  SP = hipe_sparc:mk_reg(hipe_sparc_registers:stack_pointer()),
  Reg = map(hipe_sparc:pseudo_unspill_reg(I),TempMap),
  Pos = spill_offset(hipe_sparc:pseudo_unspill_pos(I)),
  hipe_sparc:load_create(
    Reg, uw, SP, Pos, 
    hipe_sparc:info(I)).






offset(O) ->
   hipe_sparc:mk_imm(-4*(O)).

spill_pos(P) ->
  offset(P+1).

spill_offset(T) ->
 spill_pos(element(2,T)).

enter_stack_args(Args, Need, TempMap, ArgsOnStack) ->
  enter_stack_args(0, hipe_sparc_registers:register_args(), Args,
		   Need, TempMap, ArgsOnStack).

%%
%% The stack on SPARC grows uppwards (At least now...)
%%
%% Assume a function f with Arity = j is doing a tail-call to
%% a function g with arity k. 
%% Also assume the current function has spilled i temps.
%%
%%  Current stack frame:
%%
%%        |         |
%% SP ->  |         |
%%        | Spill 0 |
%%        |  ...    |
%%        | Spill i |
%%        | Dead CP |  (Read to %o7)    
%%        | Arg j   |
%%        | ...     |
%%        | Arg 6   |  (Currently 5 args in reg)
%%        |OLD FRAME|
%%
%%  Stack just before call to target:
%%        |         |
%% sp ->  |         |
%%        |         |
%% SP ->  |         |                   
%%        | Arg' k  |
%%        | ...     |
%%        | ...     |
%%        | ...     |
%%        | Arg' 6  |  (Currently 5 args in reg)
%%        |OLD FRAME|
%%
%% The arguments to g might be on the stack
%% If they are in the dangezone (the area overwritten by the arguments)
%% They have to be moved.
%% e.g. ( d marks the dangerzone)
%%           f/7               g/10
%%        |         |
%% SP ->  |         |
%%        | Spill 1 |   SP ->|         |
%%   d    | Spill 2 |        | Spill 1 |
%%   d    | Spill 3 |        | Spill 2 |
%%   d    | Dead CP |        | Spill 3 |
%%   d    | Arg 7   |        | Arg 6   |
%%   d    | Arg 6   |        | t1      |
%%        |OLD FRAME|        |OLD FRAME|
%%
%%   The DangerZone is SP - 2 to SP - 6
%%  We go through the arguments to g bottom up:
%%
%%   DZ     ARG      POSITION       SAVE  
%%
%%    < -6, t1     : in reg -> OK
%%    < -5, Arg 6  : SP-6 -> in DZ, Save = [SP-6]  
%%    < -4, Spill 2: SP-2 -> OK
%%    < -3, Spill 3: SP-1 -> OK
%%    < -2, Spill 1: SP-3 -> in DZ, Save = [SP-3, SP-6]
%%
%%  Then we copy the positions in Save to the top of the stack.
%%        | Arg 6   |
%% SP ->  | Spill 1 |
%%        | Spill 3 |
%%   d    | Spill 2 |
%%   d    | Spill 1 |
%%   d    | Dead CP |
%%   d    | Arg 7   |
%%   d    | Arg 6   |
%%        |OLD FRAME|
%%
%%  Now we can rewrite the stack safely:
%%        | A g 6   |
%% sp ->  | S i l 1 |
%% SP ->  | s i l 2 |
%%        | Spill 1 |
%%        | Spill 2 |
%%        | Spill 3 |
%%        | Arg 6   |
%%        | t1      |
%%        |OLD FRAME|

%% TODO: Verify that this realy is correct.

enter_stack_args(ArgNo, ArgNo, Args, Need, TempMap, ArgsOnStack) ->
  %% How many args has to be spilled to the stack?
  NoNewArgsOnStack = length(Args),

  %% Where does the first arg end up?
  %% SP - (Need + ArgsOnStack)*4
  Base = -(Need + ArgsOnStack),

  {ToSave,RealArgs} = args_in_danger(Args, Base-1, [], 0, [], TempMap),

  SP = hipe_sparc:mk_reg(hipe_sparc_registers:stack_pointer()),
  TempReg = hipe_sparc:mk_reg(hipe_sparc_registers:temp0()),

  SaveCode = save(ToSave,0, SP, TempReg),
  PushCode = pushem(RealArgs, Base, SP, TempReg),

  AdjustSP = 
    hipe_sparc:alu_create(
      SP,SP,'-',hipe_sparc:mk_imm((ArgsOnStack + Need - NoNewArgsOnStack)*4),[remove_frame]),
  [ SaveCode, PushCode, AdjustSP];
enter_stack_args(No, NoRegArgs, [Arg|Args], Need, TempMap, ArgsOnStack) ->
  enter_stack_args(No+1, NoRegArgs, Args, Need, TempMap, ArgsOnStack).

args_in_danger([Arg|Args], DangerZone, ToSave, SavePos, MappedArgs, TempMap) ->
  case hipe_sparc:is_reg(Arg) of
    true ->
      case hipe_temp_map:find(hipe_sparc:reg_nr(Arg), TempMap) of
	{reg, Reg} -> %% Arg in reg all cool.
	  args_in_danger(Args, DangerZone+1, ToSave, SavePos, 
			 [{reg,Reg}|MappedArgs], TempMap);
	{spill, Pos} ->
	  if -Pos < DangerZone -> %% Spill in dangerzone
	      args_in_danger(Args, DangerZone+1, [-Pos|ToSave], SavePos+1,
			     [{spill, SavePos}| MappedArgs], TempMap);
	     true ->
	      args_in_danger(Args, DangerZone+1, ToSave, SavePos,
			     [{spill, -(Pos+1)}| MappedArgs], TempMap)
	  end
      end;
    false ->
      case hipe_sparc:is_imm(Arg) of 
	true ->
	  args_in_danger(Args, DangerZone+1, ToSave, SavePos, 
			 [{imm,	hipe_sparc:imm_value(Arg)}|MappedArgs], TempMap);
	false ->
	  %% TODO: XXX: Dont handle fpreg ...
	  ?EXIT({do_no_handle,Arg})
      end
  end;
args_in_danger([],_,ToSave,_,Args,_) ->
  {lists:reverse(ToSave),lists:reverse(Args)}.

save([Pos|Poses], Offset, SP, TempReg) ->
  ToPos = hipe_sparc:mk_imm(Offset*4),
  FromPos = offset(Pos+1),
  [
   hipe_sparc:load_create(TempReg, uw, SP, FromPos, [save1]),
   hipe_sparc:store_create(SP, ToPos,  uw, TempReg, [save2])
   | save(Poses, Offset+1, SP, TempReg)];
save([],_,_,_) -> [].

pushem([{reg,R}|Args], Pos, SP, Temp) ->
  ToPos = hipe_sparc:mk_imm(Pos*4),
  [hipe_sparc:store_create(SP, ToPos,  uw, hipe_sparc:mk_reg(R), [push_arg])|
   pushem(Args,Pos+1,SP, Temp)];
pushem([{spill,P}|Args], Pos, SP, Temp) ->
  FromPos = hipe_sparc:mk_imm(P*4), 
  ToPos = hipe_sparc:mk_imm(Pos*4),
  [ hipe_sparc:load_create(Temp, uw, SP, FromPos, [get_arg]),
    hipe_sparc:store_create(SP, ToPos,  uw, Temp, [push_arg])|
    pushem(Args,Pos+1,SP, Temp)];
pushem([{imm,I}|Args], Pos, SP, Temp) ->
  ToPos = hipe_sparc:mk_imm(Pos*4),
  [ hipe_sparc:move_create(Temp, hipe_sparc:mk_imm(I), [get_arg]),
    hipe_sparc:store_create(SP, ToPos,  uw, Temp, [push_arg])|
    pushem(Args,Pos+1,SP, Temp)];
pushem([],_,_,_) -> [].




remap(I, Temps, Map) ->
%%  io:format("\n\n~w~w\n~w\n",[I,Temps,Map]),
  Substs = [{Temp, hipe_sparc:mk_reg(element(2,hipe_temp_map:find(hipe_sparc:reg_nr(Temp), Map)))}
	    || Temp <- Temps],
  hipe_sparc:subst(I, Substs).

map(Temp, Map) ->
  case hipe_sparc:is_reg(Temp) of
    true ->
      hipe_sparc:mk_reg(element(2,hipe_temp_map:find(hipe_sparc:reg_nr(Temp), Map)));
    false ->
      Temp
  end.

	 

max(P,A) ->
  if P > A -> P;
     true -> A
  end.
min(P,A) ->
  if P < A -> P;
     true -> A
  end.



push_call_args(Args, TempMap) ->
  ArgsOnStack = args_on_stack(hipe_sparc_registers:register_args(), Args),
  NoArgs = length(ArgsOnStack),      
  MappedArgs = 
  [
   case hipe_sparc:is_reg(Arg) of
     true ->
       case hipe_temp_map:find(hipe_sparc:reg_nr(Arg), TempMap) of
	 {reg, Reg} -> {reg,Reg};
	 {spill, Pos} -> {spill, -(1+Pos+NoArgs)}
       end;
     false ->
       case hipe_sparc:is_imm(Arg) of 
	 true ->
	   {imm,hipe_sparc:imm_value(Arg)};
	 false ->
	   %% TODO: XXX: Dont handle fpreg ...
	   ?EXIT({do_no_handle,Arg})
       end
   end
   || Arg <- ArgsOnStack],
  SP = hipe_sparc:mk_reg(hipe_sparc_registers:stack_pointer()),
  TempReg = hipe_sparc:mk_reg(hipe_sparc_registers:temp0()),

  {pushem(MappedArgs,-NoArgs,SP,TempReg), NoArgs}.


args_on_stack(0, Args) ->
  Args;
args_on_stack(N,[Arg|Args]) ->
  args_on_stack(N-1,Args);
args_on_stack(N,[]) -> 
  [].


extra_stack_need(Code,Need,Arity) ->
  %% Calculate number of args (to current fun) on stack
  NoRegArgs = hipe_sparc_registers:register_args(),
  ArgsOnStack = 
    if Arity > NoRegArgs -> Arity - NoRegArgs; 
       true -> 0 
    end,
  PrevSPLevel = ArgsOnStack + Need,  
  lists:foldl(fun (I,Max) ->
		  max(stack_need(I,PrevSPLevel), Max)
	      end, 
	      0,
	      Code).

stack_need(I, PrevSPLevel) ->
  %% io:format("\n\n~w\n~w\n",[I,TempMap]),
  case hipe_sparc:type(I) of
    call_link ->
      Argneed = call_args_need(hipe_sparc:call_link_args(I)),
      Argneed;

    pseudo_enter ->
      NoArgs = length(hipe_sparc:pseudo_enter_args(I)),
      NoRegArgs = hipe_sparc_registers:register_args(),
      %% stack, do some shuffling.
      %% Otherwise just adjust the stackpointer.
      case NoArgs > NoRegArgs of
	true ->        
	  %% Arguments to the called function spill 
	  NewOnStack = NoArgs - NoRegArgs,

	  case NewOnStack > PrevSPLevel of
	    true -> NewOnStack - PrevSPLevel;
	    false -> 0
	  end;
	false -> %% All arguments on stack...
	  0
      end;
    _ -> 
      %% All other instructions no extra need.
      0
  end.

call_args_need(Args) ->
  length(args_on_stack(hipe_sparc_registers:register_args(), Args)).




set_stack_size(Call,Size, Arity) ->
  SD = hipe_sparc:call_link_stack_desc(Call),
  NewSD = hipe_sparc_stack_descriptors:set_size(SD, Size),
  NewSD2 = hipe_sparc_stack_descriptors:set_arity(NewSD, Arity),
  hipe_sparc:call_link_stack_desc_update(Call,NewSD2).


gen_stack_test(StackNeed, SP, CP, Arity, Cfg) ->
  Leaf =  hipe_sparc_cfg:is_leaf(Cfg),
  TempReg = hipe_sparc:mk_reg(hipe_sparc_registers:temp0()),
  TempReg1 = hipe_sparc:mk_reg(hipe_sparc_registers:temp1()),  


  if (StackNeed) =:= 0 -> {[],[]};
     (StackNeed) < ?HIPE_SPARC_LEAF_WORDS, Leaf =:= true -> {[],[]};
     true -> 
      Zero = hipe_sparc:mk_reg(hipe_sparc_registers:zero()),
      SL = hipe_sparc:mk_reg(hipe_sparc_registers:stack_limit()),
      Pred = 0.01,
      OverflowLbl = hipe_sparc:label_create_new(),
      StartLbl = hipe_sparc:label_create_new(),
      TestLbl = hipe_sparc:label_create_new(),
      RetLbl = hipe_sparc:label_create_new(),
      OverflowName = hipe_rtl:label_name(OverflowLbl), 
      StartName = hipe_rtl:label_name(StartLbl),
      TestName = hipe_rtl:label_name(TestLbl),
      ByteNeed = (?HIPE_SPARC_LEAF_WORDS+StackNeed)*4, 
      TestCode = 
	if ByteNeed > 16#fff -> %% Max in simm13.
	    [
	     hipe_sparc:alu_create(TempReg, SL, '-', SP, []),
	     hipe_sparc:sethi_create(TempReg1,
				     hipe_sparc:mk_imm(
				       high22(ByteNeed)),[]),
	     hipe_sparc:alu_create(TempReg1, TempReg1, 'or', 
				   hipe_sparc:mk_imm(
				     low10(ByteNeed)), 
				   []),
	     hipe_sparc:alu_cc_create(Zero, TempReg, '-', TempReg1,
				      []),
	     hipe_sparc:b_create(l, OverflowName, 
				 StartName, 0.01, na, []),
	     StartLbl];
	   
	   true ->
	    [hipe_sparc:alu_create(TempReg, SL, '-', SP, []),
	     hipe_sparc:alu_cc_create(Zero, TempReg, '-', 
				      hipe_sparc:mk_imm(ByteNeed),
				      []),
	     hipe_sparc:b_create(l, OverflowName, 
				 StartName, 0.01, na, []),
	     StartLbl]
	end,
	
      
      CPSAVE = hipe_sparc:mk_reg(hipe_sparc_registers:cpsave()),
      {[hipe_sparc:goto_create(TestName,[]),
	TestLbl|TestCode],
       [OverflowLbl,
	hipe_sparc:move_create(CPSAVE, CP,[savecp]),
	hipe_sparc:call_link_create(inc_stack_fun(Arity),
				   CP,
				   [CPSAVE],
				    hipe_rtl:label_name(RetLbl), [], c, []), 
	RetLbl,
	hipe_sparc:move_create(CP, CPSAVE,[restorecp]),
	hipe_sparc:goto_create(TestName,[])]
      };
     false ->
      {[],[]}
  end.

inc_stack_fun(Arity) ->
  ArgsInRegs = 
    min(Arity,
	hipe_sparc_registers:register_args()), 
  list_to_atom("inc_stack_" ++ 
	       integer_to_list(ArgsInRegs) ++ "args_0").


gen_alloc_code(NeededBytes,SP) ->
  TempReg = hipe_sparc:mk_reg(hipe_sparc_registers:temp0()), 
  %% Make room on stack for spills.

  %% Alloc space on stack.
  if NeededBytes > 4096 -> %% To big for imm
      [hipe_sparc:sethi_create(TempReg,
			       hipe_sparc:mk_imm(
				 high22(NeededBytes)),[frame]),
       hipe_sparc:alu_create(TempReg, TempReg, 'or', 
			     hipe_sparc:mk_imm(
			       low10(NeededBytes)), 
			     []),
       hipe_sparc:alu_create(
	 SP,SP,'+',TempReg,[frame])];
     true ->
      [hipe_sparc:alu_create(
	 SP,SP,'+',hipe_sparc:mk_imm(NeededBytes),[frame])]
  end.
 
					
gen_save_cp(SP, CP, Need)->
  [hipe_sparc:store_create(
     SP,offset(Need), uw, CP, [save_cp])].

gen_ret(TempMap, Need, Arity) ->
  %% Generate code for stack cleanup and return.

  %% XXX: Fix if we start using multiple return values
  RegArgs = [hipe_sparc:mk_reg(hipe_sparc_registers:arg(0))], 
  CP = hipe_sparc:mk_reg(hipe_sparc_registers:return_address()),
  SP = hipe_sparc:mk_reg(hipe_sparc_registers:stack_pointer()),
  %% Calculate number of args (to current fun) on stack
  %% Consider moving this out of the loop.
  NoRegArgs = hipe_sparc_registers:register_args(),
  ArgsOnStack = 
    if Arity > NoRegArgs -> Arity - NoRegArgs; 
       true -> 0 
    end,
  RetLabel = hipe_sparc:label_create_new(),
  
  StackAdjust = hipe_sparc:mk_imm((ArgsOnStack+Need)*4),
  {[RetLabel,
    hipe_sparc:load_create(CP, uw, SP, offset(Need), []), 
    hipe_sparc:alu_create(SP,SP,'-', StackAdjust, [remove_frame]),
    hipe_sparc:jmp_create(CP, hipe_sparc:mk_imm(8), RegArgs, [])
   ],
   hipe_sparc:label_name(RetLabel)}.
	 

high22(X) -> X bsr 10.
low10(X) -> X band 16#3ff.
