%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%%
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%%
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%%
%%     $Id$
%%
%%----------------------------------------------------------------------
%% Purpose: Measure megaco codec's encoding & decoding time's
%%
%% Measurement process consists of:
%%   For each message in a directory:
%%     Pre:         Read message from the file, close the file
%%     Measurement: 1) measure decode
%%                  2) measure encode (of the previously decoded message)
%%     Post:        Print average
%%   For each directory: 
%%     A summery is written, both to the console and to a file, 
%%     in an excel compatible format.
%%
%% megaco_codec_meas:t().
%% megaco_codec_meas:t([pretty, compact]).
%% megaco_codec_meas:t([per, pretty, compact]).
%%
%%----------------------------------------------------------------------

-module(megaco_codec_meas).


%% API
-export([t/0, t/1]).

%% Internal exports
-export([do_measure_codec/5, do_measure_codec_loop/6]).
-export([flex_scanner_handler/1]).


-include_lib("kernel/include/file.hrl").


-define(MEASURE_TIME, 10000).



t() ->
    t([pretty, compact, per, ber, erlang]).


%% Dirs is a list of directories containing files,
%% each with a single megaco message. 
%%
%% Note that it is a requirement that each dir has
%% the name of the codec with which the messages has
%% been encoded: 
%%
%%    pretty | compact | ber | per | erlang
%%

t(Dirs) when list(Dirs) ->
%     put(verbose,true),
    EDirs = expand_dirs(Dirs, []),
    Results = t1(EDirs, []),
    store_results(Results).


t1([], Results) ->
    lists:reverse(Results);
t1([{Dir, Codec, Conf} = EDir|Dirs], Results) ->
    case (catch measure(EDir)) of
	{'EXIT', Reason} ->
	    error("measure of codec ~p exited: ~n~p", [Codec, Reason]),
	    t1(Dirs, Results);
	{error, Reason} ->
	    error("skipping codec ~p: ~n~p", [Codec, Reason]),
	    t1(Dirs, Results);
	{ok, Res} ->
	    t1(Dirs, [{Dir, Conf, Res}| Results])
    end.


measure({Dir, Codec, Conf}) when list(Dir) ->
    io:format("measure using codec ~p ~p~n  ", [Codec, Conf]),
    {Init, Conf1} = measure_init(Conf),
    Res = measure(Dir, Codec, Conf1, read_files(Dir), [], ?MEASURE_TIME),
    measure_fin(Init),
    Res.


expand_dirs([], EDirs) ->
    lists:reverse(lists:flatten(EDirs));
expand_dirs([Dir|Dirs], EDirs) when atom(Dir) ->
    EDir = expand_dir(atom_to_list(Dir)),
    expand_dirs(Dirs, [EDir|EDirs]);
expand_dirs([Dir|Dirs], EDirs) when list(Dir) ->
    EDir = expand_dir(Dir),
    expand_dirs(Dirs, [EDir|EDirs]).

expand_dir(Dir) ->
    case Dir of
	"pretty" ->
	    [{Dir, megaco_pretty_text_encoder, []},
	     {Dir, megaco_pretty_text_encoder, [flex_scanner]}];
	"compact" ->
	    [{Dir, megaco_compact_text_encoder, []},
	     {Dir, megaco_compact_text_encoder, [flex_scanner]}];
	"ber" ->
	    [{Dir, megaco_ber_bin_encoder, []},
	     {Dir, megaco_ber_bin_encoder, [native]}];
	"per" ->
	    [{Dir, megaco_per_bin_encoder, []},
	     {Dir, megaco_per_bin_encoder, [native]}];
	"erlang" ->
	    [{Dir, megaco_erl_dist_encoder, []}];
	Else ->
	    exit({error, {invalid_codec, Else}})
    end.


read_files(Dir) ->
    case file:list_dir(Dir) of
        {ok, Files} ->
            Files;
        Error ->
            exit(Error)
    end.

measure_init([flex_scanner]) ->
    start_flex_scanner();
measure_init(Conf) ->
    {undefined, Conf}.


measure_fin(Pid) when pid(Pid) ->
    stop_flex_scanner(Pid),
    ok;
measure_fin(_) ->
    ok.
	    

measure(Dir, Codec, Conf, [], Results, _Time) ->
    ECountSum = lists:sum([Ecount || {_, Ecount, _, _, _} <- Results]),
    ETimeSum  = lists:sum([Etime  || {_, _, Etime, _, _}  <- Results]),

    DCountSum = lists:sum([Dcount || {_, _, _, Dcount, _} <- Results]),
    DTimeSum  = lists:sum([Dtime  || {_, _, _, _, Dtime}  <- Results]),

    io:format("~n  Measurment on ~p messages:"
	      "~n  Average encode: ~w microsec, decode: ~w microsec~n", 
	      [length(Results), ETimeSum/ECountSum, DTimeSum/DCountSum]),

    {ok, Results};

measure(Dir, Codec, Conf, [File|Files], Results, Time) ->
    case (catch do_measure(Dir, Codec, Conf, File, Time)) of
	{ok, {Dcount, Dtime, Ecount, Etime}} ->
	    io:format("o", []),
	    measure(Dir, Codec, Conf, Files, 
		    [{File, Dcount, Dtime, Ecount, Etime}|Results], Time);

	{error, S} ->
	    case get(verbose) of
		true ->
		    io:format("~n", []),
		    error(S,[]);
		_ ->
		    io:format("e", [])
	    end,
	    measure(Dir, Codec, Conf, Files, Results, Time);

	{info, S} ->
	    case get(verbose) of
		true ->
		    io:format("~n", []),
		    info(S,[]);
		_ ->
		    io:format("i", [])
	    end,
	    measure(Dir, Codec, Conf, Files, Results, Time)
    
    end.


do_measure(Dir, Codec, Conf, File, Time) ->
    %% io:format("  measure ~s: ", [File]),
    BinMsg               = read_message(Dir, File),
    {Msg, Dcount, Dtime} = measure_decode(Codec, Conf, BinMsg, Time),
    {Ecount, Etime}      = measure_encode(Codec, Conf, Msg,    Time),

%     io:format(" decode: ~w microsec, encode: ~w microsec~n", 
% 	      [Dtime/Dcount, Etime/Ecount]),

    {ok, {Ecount, Etime, Dcount, Dtime}}.


read_message(Dir, FileName) ->
    File = filename:join([Dir, FileName]),
    case file:read_file_info(File) of
        {ok, #file_info{size = Sz, type = regular}} when Sz > 0 ->
            case file:read_file(File) of
                {ok, Msg} ->
		    %% io:format(".", []),
                    Msg;
                {error, Reason} ->
                    S = format("failed reading file ~s: ~p", [File, Reason]),
                    throw({error, S})
            end;

        {ok, #file_info{type = regular}} ->
            S = format("skipping empty file ~s", [File]),
            throw({info, S});
            
        {ok, #file_info{type = Type}} ->
            S = format("skipping ~p file ~s", [Type, File]),
            throw({info, S});

        {ok, Info} ->
            S = format("skipping file ~s~n~p", [File, Info]),
            throw({info, S});
                                                 
        {error, Reason} ->
            S = format("failed reading file info for ~s: ~p", [File, Reason]),
            throw({error, S})

    end.


measure_decode(Codec, Conf, Bin, Time) ->
    case measure_codec(Codec, decode_message, Conf, Bin, Time) of
	{ok, {Dcount, Dtime, M}} ->
	    %% io:format(".", []),
	    {M, Dcount, Dtime};
	{error, Reason} ->
	    S = format("decode failed for ~p:~n~p", [Codec, Reason]),
	    throw({error, S})
    end.
	

measure_encode(Codec, Conf, Bin, Time) ->
    case measure_codec(Codec, encode_message, Conf, Bin, Time) of
	{ok, {Ecount, Etime, _}} ->
	    %% io:format(".", []),
	    {Ecount, Etime};
	{error, Reason} ->
	    S = format("encode failed for ~p:~n~p", [Codec, Reason]),
	    throw({error, S})
    end.
	

measure_codec(Codec, Func, Conf, Bin, Time) ->
    Pid = spawn_link(?MODULE, do_measure_codec, 
                     [self(), Codec, Func, Conf, Bin]),
    receive
	Else ->
	    {error, {premature_result, Else}}

    after Time ->
	    Pid ! stop,
	    receive
		{measure_result, Pid, Func, {MCount, MTime, M}} ->
		    {ok, {MCount, MTime, M}};
		{error, Pid, Error} ->
		    {error, Error};
		Else ->
		    {error, {unexpected_result, Else}}
	    after 1000 ->
		    {error, timeout}
	    end
    end.
	

do_measure_codec(Parent, Codec, Func, Conf, Bin) ->
    Res = timer:tc(?MODULE, do_measure_codec_loop, 
		   [Codec, Func, Conf, Bin, 0, dummy]),
    case Res of
	{Time, {Count, M}} when integer(Count) ->
	    Parent ! {measure_result, self(), Func, {Count, Time, M}};
	{Time, Error} ->
	    Parent ! {error, self(), Error}
    end,
    unlink(Parent). % Make sure Parent don't get our exit signal


do_measure_codec_loop(Codec, Func, Conf, Bin, Count, M0) ->
    receive 
	stop ->
	    {Count, M0}
    after 0 ->
	    {ok, M1} = apply(Codec, Func, [Conf, Bin]),
	    do_measure_codec_loop(Codec, Func, Conf, Bin, Count + 1, M1)
    end.
      

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

store_results(Results) ->
    io:format("store: ", []),    
    store_excel_decode_time(Results),
    store_excel_encode_time(Results),
    store_excel_total_time(Results),
    io:format("~n", []),
    ok.


store_excel_encode_time(Res) ->
    io:format(".", []),
    {ok, Fd} = file:open("encode_time.xls",[write]),
    Encodes = enc_times(Res, []),
    store_excel_tab(Fd, Encodes),
    ok.

store_excel_decode_time(Res) ->
    io:format(".", []),
    {ok, Fd} = file:open("decode_time.xls",[write]),
    Decodes = dec_times(Res, []),
    store_excel_tab(Fd, Decodes),
    ok.

store_excel_total_time(Res) ->
    io:format(".", []),
    {ok, Fd} = file:open("total_time.xls",[write]),
    Totals = tot_times(Res, []),
    store_excel_tab(Fd, Totals),
    ok.


enc_times([], Times) ->
    lists:reverse(Times);
enc_times([{Dir, Conf, Res}|T], Acc) ->
    Times = [Etime/Ecount || {_, Ecount, Etime, _, _} <- Res],
    enc_times(T, [{Dir, Conf, Times}|Acc]).
    
dec_times([], Times) ->
    lists:reverse(Times);
dec_times([{Dir, Conf, Res}|T], Acc) ->
    Times = [Dtime/Dcount || {_, _, _, Dcount, Dtime} <- Res],
    dec_times(T, [{Dir, Conf, Times}|Acc]).
    
tot_times([], Times) ->
    lists:reverse(Times);
tot_times([{Dir, Conf, Res}|T], Acc) ->
    Times = [(Etime+Dtime)/(Ecnt+Dcnt) || {_,Ecnt,Etime,Dcnt,Dtime} <- Res],
    tot_times(T, [{Dir, Conf, Times}|Acc]).


store_excel_tab(Fd, Res) ->
    %% For all elements of this list, the Values is of the same length...
    [{_, _, Values}|_] = Res,
    store_excel_tab_header(Fd, length(Values), 1),
    store_excel_tab1(Fd, Res).

store_excel_tab1(Fd, []) ->
    io:format(Fd, "~n", []);
store_excel_tab1(Fd, [{Dir, [], Values}|T]) ->
    io:format(Fd, "~s", [filename:basename(Dir)]),
    store_excel_tab_row(Fd, Values),
    store_excel_tab1(Fd, T);
store_excel_tab1(Fd, [{Dir, [Conf], Values}|T]) when atom(Conf) ->
    io:format(Fd, "~s_~w", [filename:basename(Dir), Conf]),
    store_excel_tab_row(Fd, Values),
    store_excel_tab1(Fd, T).

store_excel_tab_header(Fd, 0, _) ->
    io:format(Fd, "~n", []);
store_excel_tab_header(Fd, N, M) ->
    io:format(Fd, "\t~w", [M]),
    store_excel_tab_header(Fd, N-1, M+1).

store_excel_tab_row(Fd, []) ->
    io:format(Fd, "~n", []);
store_excel_tab_row(Fd, [Value|Values]) ->
    io:format(Fd, "\t~w", [round(Value)]),
    store_excel_tab_row(Fd, Values).
    

    

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

start_flex_scanner() ->
    Pid = proc_lib:spawn(?MODULE, flex_scanner_handler, [self()]),
    receive
        {flex_scanner_started, Pid, Conf} ->
            {Pid, [Conf]};
        {flex_scanner_error, {failed_loading_flex_scanner_driver, Reason}} ->
            throw({error, {failed_loading_flex_scanner_driver, Reason}});
        {flex_scanner_error, Reason} ->
            throw({error, {failed_loading_flex_scanner_driver, Reason}})
    after 10000 ->
            exit(Pid, kill),
            throw({error, {failed_starting_flex_scanner, timeout}})
    end.

stop_flex_scanner(Pid) ->
    Pid ! stop_flex_scanner.

flex_scanner_handler(Pid) ->
    case (catch megaco_flex_scanner:start()) of
        {ok, Port} when port(Port) ->
            Pid ! {flex_scanner_started, self(), {flex, Port}},
            flex_scanner_handler(Pid, Port);
        {error, {load_driver, {open_error, Reason}}} ->
            Error = {failed_loading_flex_scanner_driver, Reason},
            Pid ! {flex_scanner_error, Error},
            exit(Error);
        Else ->
            Error = {unknown_result_from_start_flex_scanner, Else},
            Pid ! {flex_scanner_error, Error},
            exit(Error)
    end.

flex_scanner_handler(Pid, Port) ->
    receive
        {ping, Pinger} ->
            Pinger ! {pong, self()},
            flex_scanner_handler(Pid, Port);
        {'EXIT', Port, Reason} ->
            Pid ! {flex_scanner_exit, Reason},
            exit({flex_scanner_exit, Reason});
        stop_flex_scanner ->
            megaco_flex_scanner:stop(Port),
            exit(normal);
        Other ->
            info("flex scanner handler got something:~n~p", [Other]),
            flex_scanner_handler(Pid, Port)
    end.
            

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


info(F, A) ->
    io:format(F ++ "~n", A).


error(F, A) -> 
    io:format("ERROR: " ++ F ++ "~n", A).


format(F, A) ->    
    lists:flatten(io_lib:format(F, A)).
