/* ``The contents of this file are subject to the Erlang Public License,
 * Version 1.1, (the "License"); you may not use this file except in
 * compliance with the License. You should have received a copy of the
 * Erlang Public License along with this software. If not, it can be
 * retrieved via the world wide web at http://www.erlang.org/.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Original Code is Ericsson Utvecklings AB.
 * Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
 * AB. All Rights Reserved.''
 * 
 *     $Id$
 */

/*
 * Purpose: ODBC (Open Database Connectivity).
 * This version of ODBC use open_port communication to erlang 
 * and encode information with ei functions. 
 * The size of the input/output buffert is dynamic
*/

#ifdef WIN32
#include<windows.h> 
#include<io.h>
#endif

#include<stdio.h>
#include<fcntl.h>
#include<stdlib.h>
#include<string.h>
#include<memory.h>
#include<stdarg.h>
#ifndef WIN32
#include<unistd.h>
#endif
#include "ei.h"
#include "ei_x_encode.h"

#define ODBCVER 0x0300
#ifdef WIN32
#include<sql.h>
#include<sqlext.h>
#else 
#include "sql.h"
#include "sqlext.h"
#endif

/* Debug function */
#define DBG( proto ) if (dbgflag) Log proto


#define COMMANDLEN 6    /* Maximum characters in a odbc-command */
#define MAXCOLSIZE 4096 /* Size of a column with a datatypes without */
                        /* length. */


typedef unsigned char byte;
typedef int (*RHANDLER)(char *);
typedef int (*XRHANDLER)(char *, ei_x_buff *);

typedef struct {
  int selected;                   /* The column is selected */
  char columnName[100];
  char *buf;
  SQLSMALLINT type;
  char *typename;                 /* columntype as readible name */
  SQLINTEGER len;
  SQLINTEGER resultLen;
} ColumnDef;

typedef struct {
  char name[COMMANDLEN];
  RHANDLER handler;
} HandlerRecord;

typedef struct {
  char name[COMMANDLEN];
  XRHANDLER handler;
} XHandlerRecord;

/* Global variabel definition */
static int dbgflag = 0;           /* debug flag */
static char *buf;
static SQLHDBC hdbc1;             /* ConnectionHandle */
static SQLHENV henv1;             /* EnvironmentHandle */
static SQLHSTMT hstmt;            /* StatementHandle */
static ColumnDef *columns; 
static int connected;
static int numberColumn = 0;

int dbDebug(char *in);
int dbOpen(char *in);
int dbExecExt(char *in, ei_x_buff *x);
int dbCloseConnection(char *in);
int dbEndTran(char *in);
int dbExecDir(char *in);
int dbNumResultCols(char *in);
int dbRowCount(char *in);
int dbBindCol(char *in);
int dbDescribeCol(char *in);
int dbFetch(char *in);
int dbreadbuffer(char *in);
int dbSetConnectAttr(char *in);
int dbCloseCursor(char *in);
int write_cmd(byte *buf, int len);

void Log(char *s, ...){
  FILE *f;
  va_list args;
  va_start(args, s); /* The first arg without name */
  f = fopen("odbc_debug.log", "a");
  vfprintf(f, s, args);
  fclose(f);
  va_end(args);
}


/* Translete the requested instruction to a function */ 
HandlerRecord requestHandlers[] = {
  {"debugs", dbDebug},
  {"opendb", dbOpen},
  {"closdb", dbCloseConnection},
  {"bincol", dbBindCol},
  {"descol", dbDescribeCol},
  {"endtra", dbEndTran},
  {"exedir", dbExecDir},
  {"fetchd", dbFetch},
  {"rescol", dbNumResultCols},
  {"rowcon", dbRowCount},
  {"setatr", dbSetConnectAttr},
  {"readbu", dbreadbuffer},
  {"closec", dbCloseCursor}
};

XHandlerRecord xRequestHandles[] = {
  {"execdb", dbExecExt},
  {"", NULL}
};  

/* HandleRequest do a request */
/* Input is a requeststring */
/* HandleRequest returns the length of data which should be sent or */
/* zero if data have allready been sent or minus one if request is */
/* unknown. */
/* Normally does function return zero if there is a error but this */
/* case does the function return a length if it returns zero or higher */
/* and error if below zero */
int handleRequest(char *reqstring) {
  char request[COMMANDLEN], *tmp1;
  int i, result;
  XHandlerRecord *xrp;
  
  for(i = 0; i < COMMANDLEN; i++) 
    request[i] = reqstring[i];
  request[COMMANDLEN] = '\0';
  /* move the pointer in reqstring */
  tmp1 = reqstring + COMMANDLEN;
  DBG(("handleRequest ->requeststring: %s \n", request));
  i = 0;
  while( i < sizeof(requestHandlers) / sizeof(requestHandlers[0])) {
    if(strcmp(requestHandlers[i].name, request) == 0) {
      result = requestHandlers[i].handler(tmp1);
      return result;
    }
    i++;
  }
  for (xrp = xRequestHandles; xrp->handler != NULL; ++xrp) {
    if (strcmp(xrp->name, request) == 0) {
      ei_x_buff x;
      ei_x_new_with_version(&x);
      xrp->handler(tmp1, &x);
      write_cmd(x.buff, x.index);
      ei_x_free(&x);
      return 0;
    }
  }

  /* There are no such command */
  return -1;
}
/* receive buffer with a len byte  from stdio */ 
int read_exact(byte *buf, int len) {
  int i, got=0;

  do {
    if ((i = read(0, buf+got, len-got)) <= 0)
      return(i);
    got += i;
  } while (got<len);
  return len;
}

int write_exact(byte *buf, int len) {
  int i, wrote = 0;

  do {
    if ((i = write(1, buf+wrote, len-wrote)) <= 0)
      return (i);
    wrote += i;
  } while (wrote<len);
  return len;
}

/* receive the length of a receiving message and */
/* change the received length from big endian to little endian */
int receivedlength() {
  int len, i, tmp;
  byte lbuf[4];

  len = 0;
  tmp = read_exact(lbuf, 4);
  if(tmp != 4)
    return -1;
  for(i=0; i < 4; i++) {
    len <<= 8;
    len |= lbuf[i];
  }
  return len;
}

/* Write buffer to stdout */
int write_cmd(byte *buf, int len) {
  byte li;

  li = (len >> 24) & 0x000000FF;
  write_exact(&li, 1);
  li = (len >> 16) & 0x000000FF;
  write_exact(&li, 1);
  li = (len >> 8) & 0x000000FF;
  write_exact(&li, 1);
  li = len &  0x000000FF;
  write_exact(&li, 1);
  return write_exact(buf, len);
}

int sql_success(int r) {
    return r == SQL_SUCCESS || r == SQL_SUCCESS_WITH_INFO;
}

int sql_error(SQLSMALLINT err_handle, ei_x_buff* x, const char* func_name)
{
  char msg[1000];
  SQLCHAR sqlState[6];
  SQLINTEGER nativeError;
  SQLSMALLINT i, msgLen;
  ei_x_buff x_msg;
  
  ei_x_new(&x_msg);
  i = 1; 
  while (SQLGetDiagRec(err_handle, hstmt, i, sqlState, &nativeError,
		       msg, sizeof(msg), &msgLen) 
	 == SQL_SUCCESS) {
    DBG(("sql_error -> error Message code %d\n", nativeError));
    DBG(("sql_error -> error Message %s\n", msg));
    DBG(("sql_error -> sqlState %s\n",sqlState));
    ei_x_append_buf(&x_msg, msg, msgLen);
    ++i;
  }
  ei_x_encode_tuple_header(x, 3);
  ei_x_encode_atom(x, "error");
  ei_x_encode_string(x, func_name);
  ei_x_encode_string_len(x, x_msg.buff, x_msg.index);
  ei_x_free(&x_msg);
  return -1;
}

/* loockup the SQL STATE, ignored any errormessages */
int getDiagnos(SQLSMALLINT handleType, SQLHANDLE handle, 
		char *state) {
  SQLCHAR sqlState[6];
  SQLINTEGER nativeError;
  SQLSMALLINT msgLen;
  SQLSMALLINT i;
  int j, len;
  char *msg;

  msg = malloc(200);
  if(msg == NULL)
    return -1;
  len = 199;
  sqlState[5] = '\0';
  i = 1;
  j = 0;
  while(SQLGetDiagRec(handleType, handle, i, sqlState, &nativeError,
		      msg, sizeof(msg), &msgLen) == SQL_SUCCESS) {
    if (j < (len - 2)) {
      j = j + msgLen;
      msg[j] = '\0';
      j++;
      i++;
    }
    i++;
    DBG(("getDiagnos -> sqlState %s\n",sqlState));
  }
  msg[j] = '\0';
  strcpy(state, sqlState);
  free(msg);
  return j;
}

/* Get Errormessages */
int getError(SQLSMALLINT handleType, SQLHANDLE handle, 
		char *msg, int len) {
  SQLCHAR sqlState[6];
  SQLINTEGER nativeError;
  SQLSMALLINT msgLen;
  SQLSMALLINT i;
  int j;

  sqlState[5] = '\0';
  i = 1;
  j = 0;
  while(SQLGetDiagRec(handleType, handle, i, sqlState, &nativeError,
		      msg, sizeof(msg), &msgLen) == SQL_SUCCESS) {
    if (j < (len - 2)) {
      j = j + msgLen;
      msg[j] = '\r';
      j++;
      msg[j] = '\0';
      DBG(("getError -> error Message number %d\n",i));
      DBG(("getError -> error Message %s\n",msg));
      j++;
      i++;
    }
    i++;
    DBG(("getError -> sqlState %s\n",sqlState));
  }
  msg[j] = '\0';
  if(j == 0)
    strcpy(msg, "unknown error\r\n");
  return j;
}

/* switch on or off the debuging function */
int dbDebug(char *in) {
  if (strcmp(in,"on") == 0)
    dbgflag = 1;
  else
    dbgflag = 0;
  return 0;
}

/* connect to the database */
int dbOpen(char *in) {
  char connStr[256];
  SQLRETURN	ret;
  SQLSMALLINT si, connlen;
  int eimes, index,errMsgLen;
  char *conn, *mes, *tmp;
  
  errMsgLen = 0;
  conn = in;
  DBG(("dbOpen -> The connection string is: %s\n",conn));
  if(connected) {
    mes = "error";
    tmp ="already_connected";
    goto exit;
  }

  /* allocate Environment handle and register version */
  ret = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &henv1);
  if (!sql_success(ret)) {
    mes = "error";
    tmp = "SQLAllocHandle";
    goto exit;
  }
  /* Set the odbc-version to 3.0 */
  ret = SQLSetEnvAttr(henv1, SQL_ATTR_ODBC_VERSION, 
		      (SQLPOINTER)SQL_OV_ODBC3,0);
  if (!sql_success(ret)){
    mes = "error";
    tmp = "SQLSetEnvAttr";
    goto exit;			
  }

  SQLAllocHandle(SQL_HANDLE_DBC, henv1, &hdbc1);
  if (!sql_success(ret)) {
    mes = "error";
    tmp = "SQLAllocHandle";
    goto exit;
  }

  connlen = (SQLSMALLINT)strlen(conn);
  ret = SQLDriverConnect(hdbc1, NULL, conn, connlen, connStr, 
			sizeof(connStr), &si, 
			SQL_DRIVER_NOPROMPT);
  if (sql_success(ret)) {
    mes = "ok";
    tmp = "connected";
    connected = 1;
  }
  else {
    mes = "error";
    tmp = malloc(128);
    errMsgLen = getError(SQL_HANDLE_DBC, hdbc1, tmp, 128);
    SQLFreeHandle(SQL_HANDLE_DBC, hdbc1);
    SQLFreeHandle(SQL_HANDLE_ENV, henv1);
    connected = 0;
  }

 exit:
  index = 0;
  buf = malloc(200 + errMsgLen);
  eimes = ei_encode_version(buf, &index);
  eimes = ei_encode_tuple_header(buf, &index, 3);
  eimes = ei_encode_atom(buf, &index, mes);
  eimes = ei_encode_string(buf, &index, tmp);
  eimes = ei_encode_long(buf, &index, ret);
  DBG(("dbOpen -> returned mes %s\n", mes));
  DBG(("dbOpen -> returned returncode %d\n", ret));
  if(errMsgLen > 0)
    free(tmp);
  DBG(("dbOpen -> returned index %d\n", index));
  return index;
}

/* Transform SQL columntype to C columntype */
int column_type(int typ, int sz, int decimals, ColumnDef* column) {
    switch(typ) {
    case SQL_UNKNOWN_TYPE:
      column->len = 0;
      return 0;
      break;
    case SQL_CHAR:
      column->len = sz + 1;
      column->type = SQL_C_CHAR;
      column->typename = "SQL_CHAR";
      break;
    case SQL_NUMERIC:
      column->len = sizeof(double);
      column->type = SQL_C_DOUBLE;
      column->typename = "SQL_NUMERIC";
      break;
    case SQL_DECIMAL:
      column->len = 50;
      column->type = SQL_C_CHAR;
      column->typename = "SQL_DECIMAL";
      break;
    case SQL_INTEGER:
      column->len = sizeof(int);
      column->type = SQL_C_SLONG;
      column->typename = "SQL_INTEGER";
      break;
    case SQL_SMALLINT:
      column->len = sizeof(int);
      column->type = SQL_C_SLONG;
      column->typename = "SQL_SMALLINT";
      break;
    case SQL_FLOAT:
      column->len = sizeof(double);
      column->type = SQL_C_DOUBLE;
      column->typename = "SQL_FLOAT";
      break;
    case SQL_REAL:
      column->len = sizeof(double);
      column->type = SQL_C_DOUBLE;
      column->typename = "SQL_REAL";
      break;
    case SQL_DOUBLE:
      column->len = sizeof(double);
      column->type = SQL_C_DOUBLE;
      column->typename = "SQL_DOUBLE";
      break;
    case SQL_VARCHAR:
      column->len = sz + 1;
      column->type = SQL_C_CHAR;
      column->typename = "SQL_VARCHAR";
      break;
    case SQL_TYPE_DATE:
    case SQL_TYPE_TIME:
    case SQL_TYPE_TIMESTAMP:
      /* erlang dates instead? */
      column->len= 30;
      column->type = SQL_C_CHAR;
      column->typename = "SQL_TYPE_DATE/TIME";
      break;
    case SQL_LONGVARCHAR:
      column->len = sz + 1;
      column->type = SQL_C_BINARY;
      column->typename = "SQL_LONGVARCHAR";
      break;
    case SQL_BINARY:
      column->len = sz + 1;
      column->type = SQL_C_BINARY;
      column->typename = "SQL_BINARY";
      break;
    case SQL_VARBINARY:
      column->len = sz + 1;
      column->type = SQL_C_BINARY;
      column->typename = "SQL_VARBINARY";
      break;
    case SQL_LONGVARBINARY:
      column->len = sz + 1;
      column->type = SQL_C_BINARY;
      column->typename = "SQL_LONGVARBINARY";
      break;
    case SQL_BIGINT:
      column->len = 100;
      column->type = SQL_C_CHAR;
      column->typename = "SQL_BIGINT";
      break;
    case SQL_TINYINT:
      column->len = sizeof(int);
      column->type = SQL_C_SLONG;
      column->typename = "SQL_TINYINT";
      break;
    case SQL_BIT:
      column->len = 7;
      column->type = SQL_C_BIT;
      column->typename = "SQL_BIT";
      break;

    default:
      column->len = 0;
      return 0;
      break;
    } /* switch */
    return 1;
}

int dbExecExt(char *in, ei_x_buff *x)
{
    const char* odbc_err_func = "";
    char *sql;
    char *atom, *prevptr;
    char sqlState[6];
    int blocklen, prevlen;
    int rowCount;
    int elements, i, eimes;
    int update;
    ColumnDef *columns;
    SQLTCHAR cName[255];
    SQLSMALLINT cNameLen, cType, cDecDigits, cNullable, NumColumn;
    SQLUINTEGER cSize; 
    SQLRETURN ret;
    SQLINTEGER RowCountPtr;

    numberColumn = 0;
    hstmt = NULL;
    columns = NULL;

    DBG(("dbExecExt -> the complete command string is: %s\n",in));
    /* DebugBreak(); */
    /* alloc stmt handle */
    ret = SQLAllocHandle(SQL_HANDLE_STMT, hdbc1, &hstmt);
    DBG(("dbExecExt -> SQLAllocHandle returnvalue: %d\n", ret));
    if (!sql_success(ret)) {
	odbc_err_func = "SQLAllocHandle";
	goto exit; 
    }


    /* find size of column buffer */
    sql = in;
    ret = SQLExecDirect(hstmt, sql, SQL_NTS);
    DBG(("dbExecExt -> SQLExecDirect returnvalue: %d\n", ret));
    if (!sql_success(ret)) {
	odbc_err_func = "SQLExecDirect";
	goto exit; 
    }

    /* count columns */
    DBG(("dbExecExt -> translate columns to sql values\n"));
    ret = SQLNumResultCols(hstmt, &NumColumn);
    if (!sql_success(ret)) {
	odbc_err_func = "SQLNumResultCols";
	goto exit; 
    }
    update = 0;
    DBG(("dbExecExt -> Number of columns: %d\n", NumColumn));
    if (NumColumn == 0) {
	elements = 2;
	atom = "updated";
	update = 1;
    } else {
	elements = 3;
	atom = "selected";
    }
    ret = SQLRowCount(hstmt, &RowCountPtr); 
    if(!sql_success(ret)) {
      odbc_err_func = "SQLRowCount";
      goto exit;
    }
    rowCount =(int)RowCountPtr;
    DBG(("dbExecExt -> Number of rows: %d\n", rowCount)); 
   ei_x_encode_tuple_header(x, elements);
    ei_x_encode_atom(x, atom);
    if (update == 1)
      goto exit;
    ei_x_encode_list_header(x, NumColumn);

    columns = (ColumnDef*)malloc(NumColumn * sizeof(ColumnDef));
    if (columns == NULL) {
      odbc_err_func = "malloc_no_mem_left";
      goto exit;
    }
    DBG(("dbExecExt -> allocate memory for a row \n"));
    numberColumn = NumColumn;
    for (i = 0; i < NumColumn; ++i) {
      DBG(("dbExecExt -> transf SQL-type to C-type for column %d\n", i +1)); 
      ret = SQLDescribeCol(hstmt, (SQLSMALLINT)(i+1), cName,
			   sizeof(cName), &cNameLen, &cType, 
			   &cSize, &cDecDigits, &cNullable);
      DBG(("dbExecExt -> SQLDescribeCol returns: %d\n", ret));
	if (!sql_success(ret)) {
	  odbc_err_func = "SQLDescribeCol";
	  goto exit; 
	}
	/* column has a unknown datatype */
	if (cType == SQL_UNKNOWN_TYPE) {
	  odbc_err_func = "Column_has_Unknown_data_type";
	  goto exit;
	}
	/* The cType SQL_LONGVARCHAR or SQL_LONGVARBINARY has a */
	/* cSize 2 GByte which must be limited to MAXCOLSIZE */
       	if(cType == SQL_LONGVARCHAR || cType == SQL_LONGVARBINARY)
	  cSize = MAXCOLSIZE;
	/* map column type */
	if (!column_type(cType, cSize, cDecDigits, &columns[i])) {
	  odbc_err_func = "unknown_columntype";
	  goto exit;
	}
	DBG(("dbExecExt -> columntype %d\n",cType));
	DBG(("dbExecExt -> columntypename %s\n",columns[i].typename));
	DBG(("dbExecExt -> columnlength %ld \n", columns[i].len)); 
	if (columns[i].len > 0) {
	    columns[i].buf = malloc(columns[i].len);
	    if (columns[i].buf == NULL) {
	      odbc_err_func = "malloc_No_memory_left";
	      goto exit;
	    }
       	    if (cType != SQL_LONGVARCHAR && cType != SQL_LONGVARBINARY) { 
	      ret = SQLBindCol(hstmt, (SQLSMALLINT)(i+1), columns[i].type, 
			       columns[i].buf,
			       columns[i].len, &columns[i].resultLen);
	      DBG(("dbExecExt -> SQLBindCol returns: %d\n", ret));
	      if (!sql_success(ret)) {
	    	odbc_err_func = "SQLBindCol";
		goto exit; 
	      }
	    } 
	/* encode column name */
	ei_x_encode_string_len(x, cName, cNameLen);
	} 
	else {
	  columns[i].len = 0;
	  columns[i].buf = NULL;
	}
    }  /* for */
    ei_x_encode_empty_list(x); /* end of column name list */
    
    /* now fetch data and encode it */
    for (;;) {      
      /* fetch it */
      ret = SQLFetch(hstmt);
      DBG(("dbExecExt -> SQLFetch returns: %d\n", ret));
      if (ret == SQL_NO_DATA)
	break;
      if (!sql_success(ret)) {
	odbc_err_func = "SQLFetch";
	goto exit;
      }
      if (columns[i].resultLen == SQL_NO_TOTAL) {
	odbc_err_func = "No_length_on_data";
	goto exit;
      }
      ei_x_encode_list_header(x, 1);
      /* encode it */
      ei_x_encode_list_header(x, NumColumn);
      for (i = 0; i < NumColumn; i++) {
	DBG(("dbExecExt -> encode column number: %d\n", i + 1));
	if (columns[i].len == 0 || columns[i].resultLen == SQL_NULL_DATA) {
	  DBG(("dbExecExt -> encode SQL_NULL_DATA \n"));
	  ei_x_encode_atom(x, "null");
	} 
	else {
	  switch(columns[i].type) {
	  case SQL_C_CHAR:
	    DBG(("dbExecExt -> encode SQL_C_CHAR \n"));
	    ei_x_encode_string(x, columns[i].buf);
	    break;
	  case SQL_C_SLONG:
	    DBG(("dbExecExt -> encode SQL_C_SLONG \n"));
	    ei_x_encode_long(x, *(long*)columns[i].buf);
	    break;
	  case SQL_C_DOUBLE:
	    DBG(("dbExecExt -> encode SQL_C_DOUBLE \n"));
	    ei_x_encode_double(x, *(double*)columns[i].buf);
	    break;
	  case SQL_C_BIT:
	    DBG(("dbExecExt -> encode SQL_C_BIT \n"));
	    ei_x_encode_atom(x, columns[i].buf[0]?"true":"false");
	    break;
	  case SQL_C_BINARY:
	    /*      DebugBreak();*/
	    DBG(("dbExecExt -> encode SQL_C_BINARY \n"));
	    blocklen = columns[i].len;
	    prevptr = columns[i].buf;
	    ret = SQLGetData(hstmt, (SQLSMALLINT)(i+1), SQL_C_CHAR, prevptr,
			     blocklen, &columns[i].resultLen);
	    DBG(("dbExecExt -> SQLGetData returns: %d\n", ret));
	    DBG(("dbExecExt -> columns buffer %s\n", columns[i].buf));

	    if (ret == SQL_SUCCESS_WITH_INFO) 
	      do {
		eimes = getDiagnos(SQL_HANDLE_STMT, hstmt,sqlState);
		/* increase columnbuffer with blocklen */
		if(columns[i].resultLen == SQL_NO_TOTAL ||
		   strcmp(sqlState, "01004") == 0) {
		  prevlen = columns[i].len - 1;
		  columns[i].len = prevlen +  blocklen;
		  prevptr = realloc(columns[i].buf, columns[i].len);
		  if (prevptr == NULL) {
		    free(columns[i].buf);
		    odbc_err_func = "malloc_No_memory_left";
		    goto exit;
		  }
		  columns[i].buf = prevptr;
		  prevptr = columns[i].buf + prevlen;
		  DBG(("dbExecExt -> columns buffer size %d\n",
		       columns[i].len));
		  ret = SQLGetData(hstmt, (SQLSMALLINT)(i+1), SQL_C_CHAR, 
				   prevptr, blocklen, &columns[i].resultLen);
		  DBG(("dbExecExt -> SQLGetData returns %d\n", ret));
		  DBG(("dbExecExt -> columns buffer %s\n", columns[i].buf));
		}
	      } while (ret == SQL_SUCCESS_WITH_INFO && ret != SQL_ERROR);
	    if (ret == SQL_NO_DATA) {
	      DBG(("dbExecExt -> ret == SQL_NO_DATA \n"));
	      ei_x_encode_string(x, columns[i].buf);
		break;
	    }
	    if (ret == SQL_SUCCESS) {
	      DBG(("dbExecExt -> ret == SQL_SUCCESS \n"));
	      DBG(("dbExecExt -> columns buffer %s\n", columns[i].buf));
	      ei_x_encode_string(x, columns[i].buf);
		break;
	    }
	    if (ret == SQL_ERROR) {
	      odbc_err_func = "SQLGetData";
	      goto exit;
	    }
	    break;
	  default:
	    DBG(("dbExecExt -> encode default\n"));
	    ei_x_encode_atom(x, "error");
	    break;
	  } /*switch */
	} /* else */
      } /* column for */
      ei_x_encode_empty_list(x);
    } /* for(;;) */
    ei_x_encode_empty_list(x);

 exit:
    DBG(("dbExecExt -> exit\n"));
    DBG(("dbExecExt -> return column buffers\n"));
    for (i = 0; i < numberColumn; ++i)
      if (columns[i].buf != NULL && columns[i].len > 0) {
	DBG(("dbExecExt -> return column buffer number %d \n",i));
	free(columns[i].buf);
      }
    DBG(("dbExecExt -> return columns\n"));
    if(columns != NULL)
      free(columns);
    if(update == 1)
      ei_x_encode_long(x, rowCount);
    DBG(("dbExecExt -> update \n"));
    if(odbc_err_func[0] != '\0') {
      x->index = 0;
      ei_x_encode_version(x);
      sql_error(SQL_HANDLE_STMT, x, odbc_err_func);
    }
    DBG(("dbExecExt -> errfunc \n"));
    if (hstmt != NULL) {
      ret = SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
      DBG(("dbExecExt -> SQLFreeHandle returns: %d\n", ret));
      if (!sql_success(ret))
        odbc_err_func = "SQLFreeStmt";
    }
    numberColumn = 0; /* since we've freed them all */
    update = 0;
    return 0;
}

int dbExecDir(char *in) {
  char *mes, *result;
  char *sql;
  int index, errMsgLen; 
  int eimes;
  SQLRETURN ret;	
  
  errMsgLen = 0;
  ret = SQLAllocHandle(SQL_HANDLE_STMT, hdbc1, &hstmt);
  DBG(("dbExecDir -> SQLAllocHandle returnvalue: %d\n", ret));
  if(ret == SQL_ERROR) {
    result = malloc(1024);
    if (result == NULL) {
      mes = "error";
      result = "No_mem_left";
      goto exit;
    }
    DBG(("dbExecDir -> allocate memory for errmessage \n"));
    errMsgLen = getError(SQL_HANDLE_DBC, hdbc1, result, 1024);
    mes = "error";
    goto exit;
  }
  if(ret == SQL_INVALID_HANDLE) {
    mes = "error";
    result = "SQLAllocHandle_Invalid_handle";
    goto exit;
  }

  sql = in;
  DBG(("dbExecDir -> the command string is: %s\n",sql));
  ret = SQLExecDirect(hstmt, sql, SQL_NTS);
  DBG(("dbExecDir -> SQLExecDirect returnvalue: %d\n", ret)); 
  if(ret == SQL_ERROR) {
    result = malloc(1024);
    if (result == NULL) {
      mes = "error";
      result = "No_mem_left";
      goto exit;
    }
    DBG(("dbExecDir -> allocate memory for errmessges \n"));
    errMsgLen = getError(SQL_HANDLE_STMT, hstmt, result, 1024);
    mes = "error";
    goto exit;
  }
  if(ret == SQL_INVALID_HANDLE) {
    mes = "error";
    result = "SQLExecDirect_Invalid_handle";
    goto exit;
  }

  /* numberColumn = 0;*/
  mes = "ok";
  result = "dbExecDir"; 
 exit:
  DBG(("dbExecDir -> mes %s\n", mes));
  buf = malloc(250 + errMsgLen); 
  index = 0;
  eimes = ei_encode_version(buf, &index);
  eimes = ei_encode_tuple_header(buf, &index, 3);
  eimes = ei_encode_atom(buf, &index, mes);
  eimes = ei_encode_string(buf, &index, result);
  eimes = ei_encode_long(buf, &index, ret); 
  DBG(("dbExecDir -> result %s\n", result));
  if (errMsgLen > 0)
    free(result);
  DBG(("dbExecDir -> returned index %d\n", index));
  return index;
}

int dbEndTran(char *in) {
  char *mes, *result, attr[20];
  int index, eimes;
  SQLRETURN ret;
  SQLSMALLINT intattr;
  int i;

  DBG(("dbEndTran -> inparameter:%s\n", in));

  i = 0;   /* string index */
  while(in[i] != '\0' && i < 20) {
      if(in[i] == 59) /* if in[i] == ";" */
	  attr[i] = '\0';
      else 
	  attr[i] = in[i];
      i++;
  }

  intattr = atoi(attr);
  ret = SQLEndTran(SQL_HANDLE_DBC, hdbc1, intattr);
  DBG(("dbEndTran -> SQLEndTrans returned %d\n", ret));
  buf = malloc(200);
  index = 0;
  if (!sql_success(ret)) {
    mes = "error";
    result = "SQLEndTran";
    eimes = ei_encode_version(buf, &index);
    eimes = ei_encode_tuple_header(buf, &index, 3);
    eimes = ei_encode_atom(buf, &index, mes);
    eimes = ei_encode_string(buf, &index, result);
    eimes = ei_encode_long(buf, &index, ret);
    DBG(("dbEndTran -> mes %s\n", mes));
    DBG(("dbEndTran -> returned result %s \n", result));
    DBG(("dbEndTran -> returned index %d\n", index));
    return index;
  }
  else {
    mes = "ok";
    result = "dbEndTran";
    eimes = ei_encode_version(buf, &index);
    eimes = ei_encode_tuple_header(buf, &index, 3);
    eimes = ei_encode_atom(buf, &index, mes);
    eimes = ei_encode_string(buf, &index, result);
    eimes = ei_encode_long(buf, &index, ret);
    DBG(("dbEndTran -> mes %s\n", mes));
    DBG(("dbEndTran -> rowcount %s\n",result)); 
    DBG(("dbEndTran -> returned index %d\n", index));
    return index;
  }
}

int dbNumResultCols(char *in) {
  char *mes, *result;
  int index;
  SQLSMALLINT NumColumn;
  int eimes;
  SQLRETURN ret;	
  
  buf = malloc(200); 
  index = 0;
  ret = SQLNumResultCols(hstmt, &NumColumn);
  DBG(("dbNumResultCols -> returned ret %d\n", ret));
  if (!sql_success(ret)) {
    mes = "error";
    result = "SQLNumResultCols";
    SQLCloseCursor(hstmt); 
    SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
    eimes = ei_encode_version(buf, &index);
    eimes = ei_encode_tuple_header(buf, &index, 3);
    eimes = ei_encode_atom(buf, &index, mes);
    eimes = ei_encode_string(buf, &index, result);
    eimes = ei_encode_long(buf, &index, ret);
    DBG(("dbNumResultCols -> mes %s\n", mes));
    DBG(("dbNumResultCols -> returned result %s \n", result));
    DBG(("dbNumResultCols -> returned index %d\n", index));
    return index;
  }
  else {
    numberColumn = (int)NumColumn;
    mes = "ok";
    eimes = ei_encode_version(buf, &index);
    eimes = ei_encode_tuple_header(buf, &index, 3);
    eimes = ei_encode_atom(buf, &index, mes);
    eimes = ei_encode_long(buf, &index, numberColumn);
    eimes = ei_encode_long(buf, &index, ret);
    DBG(("dbNumResultCols -> mes %s\n", mes));
    DBG(("dbNumResultCols -> Number of Column %d\n", numberColumn)); 
    DBG(("dbNumResultCols -> returned index %d\n", index));
  return index;
  }
}

int dbRowCount(char *in) {
  char *mes, *result;
  int index;
  SQLINTEGER rowCount;
  int eimes;
  SQLRETURN ret;
  
  buf = malloc(200); 
  index = 0;
  ret = SQLRowCount(hstmt, &rowCount);
  DBG(("dbRowCount -> returned ret %d\n", ret));
  if (!sql_success(ret)) {
    mes = "error";
    result = "SQLRowCount";
    SQLCloseCursor(hstmt); 
    SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
    eimes = ei_encode_version(buf, &index);
    eimes = ei_encode_tuple_header(buf, &index, 3);
    eimes = ei_encode_atom(buf, &index, mes);
    eimes = ei_encode_string(buf, &index, result);
    eimes = ei_encode_long(buf, &index, ret);
    DBG(("dbRowCount -> mes %s\n", mes));
    DBG(("dbRowCount -> returned result %s \n", result));
    DBG(("dbRowCount -> returned index %d\n", index));
    return index;
  }
  else {
    mes = "ok";
    eimes = ei_encode_version(buf, &index);
    eimes = ei_encode_tuple_header(buf, &index, 3);
    eimes = ei_encode_atom(buf, &index, mes);
    eimes = ei_encode_long(buf, &index, rowCount);
    eimes = ei_encode_long(buf, &index, ret);
    DBG(("dbRowCount -> mes %s\n", mes));
    DBG(("dbRowCount -> rowcount %d\n",rowCount)); 
    DBG(("dbRowCount -> returned index %d\n", index));
    return index;
  }
}

/* Set connectionattributes */
int dbSetConnectAttr(char *in) {
    char *mes, *result;
    int index, eimes;
    SQLRETURN ret;
    SQLINTEGER attr1, intattr2;
    int sel;
    char *array[3];
    int i,j,k;

    DBG(("dbSetConnectAttr -> inparameter:%s\n", in));
  /* This need some comments: */
  /* SQLSetConnectAttr needs two attr from Erlang. */
  /* The first arguments is a integer, second a decoder for the third */
  /* argument and finaly the third argument which */
  /* can be string or integer. Between each argument is delimeter */
  /* character ;. The decoder decode 0 for integer and 1 for string. */
  /* So a string as third argument can look like "0;1;nisse;" and */
  /* and integer like "0;0;12;" */
  i = 0;   /* in string index */
  j = 0;   /* which attr */
  k = 0;   /* which char in the attr */
  
  /* Reserve memory for each argument */
  while(in[i] != '\0' && i < 40) {
    if(in[i] == 59) { /* if in[i] == ";" */
      array[j] = malloc(k + 1);
      j++;
      k = 0;
    }
    else
      k++;
    i++;
  }
  i = 0; j = 0; k = 0;
  while(in[i] != '\0' && i < 40) {
    if(in[i] == 59) { /* if in[i] == ";" */
      array[j][k] = '\0';
      j++;
      k = 0;
    }
    else {
      array[j][k] = in[i];
      k++;
    }
    i++;
  }
  attr1 = atoi(array[0]);
  sel = atoi(array[1]);
  if(sel == 0) { /* second attribute integer */
    intattr2 = atoi(array[2]);
    ret = SQLSetConnectAttr(hdbc1, attr1, (SQLPOINTER)intattr2, 0);
  }
  else 
    ret = SQLSetConnectAttr(hdbc1, attr1, (SQLPOINTER)array[2], SQL_NTS);
  DBG(("dbSetConnectAttr -> SQLSetConnectAttr returned %d\n", ret));
  for(i = 0; i < 3; i++) 
    free(array[i]);
  buf = malloc(200);
  index = 0;
  if (!sql_success(ret)) {
    mes = "error";
    result = "SQLSetConnectAttr";
    eimes = ei_encode_version(buf, &index);
    eimes = ei_encode_tuple_header(buf, &index, 3);
    eimes = ei_encode_atom(buf, &index, mes);
    eimes = ei_encode_string(buf, &index, result);
    eimes = ei_encode_long(buf, &index, ret);
    DBG(("dbSetConnectAttr -> mes %s\n", mes));
    DBG(("dbSetConnectAttr -> returned result %s \n", result));
    DBG(("dbSetConnectAttr -> returned index %d\n", index));
    return index;
  }
  else {
    mes = "ok";
    result = "dbSetConnectAttr";
    eimes = ei_encode_version(buf, &index);
    eimes = ei_encode_tuple_header(buf, &index, 3);
    eimes = ei_encode_atom(buf, &index, mes);
    eimes = ei_encode_string(buf, &index, result);
    eimes = ei_encode_long(buf, &index, ret);
    DBG(("dbSetConnectAttr -> mes %s\n", mes));
    DBG(("dbSetConnectAttr -> rowcount %s\n",result)); 
    DBG(("dbSetConnectAttr -> returned index %d\n", index));
    return index;
  }
}

/* dbBindCol creates memory for the columns */
/* Inparmeter: Column number (starts with 1) */
int dbBindCol(char *in) {
  char *mes, *result, *col;
  int index, j;
  SQLUSMALLINT i, NumColumn;
  SQLTCHAR cName[255];
  SQLSMALLINT cNameLen, cType, cDecDigits, cNullable;
  SQLUINTEGER cSize; 
  int resultListLen;
  int eimes;
  SQLRETURN ret;	

  DBG(("dbBindCol -> translate...\n"));
  DBG(("dbBindCol -> inparameter:%s\n", in));

  col = in;
  /* selected column */
  j = atoi(col);
  DBG(("dbBindCol -> selected column:%d\n", j));

  /* The columnarray starts at index 0 */
  i = j - 1;
  /* If Number of Column are not known */
  if (numberColumn == 0) {
    ret = SQLNumResultCols(hstmt, &NumColumn);
    DBG(("dbBindCol -> SQLNumResultCols return %d\n", ret));
    if(!sql_success(ret)) {
      mes = "error";
      result = "SQLNumResultCols";
      goto exit;
    }
    numberColumn = (int)NumColumn;
    DBG(("dbBindCol -> number of Columns %d\n", numberColumn));
  }
  if (columns == NULL) {
    columns = malloc(numberColumn * sizeof(ColumnDef));
    if (columns == NULL) {
      mes = "error";
      result = "No_mem_left";
      goto exit;
    }
    DBG(("dbBindCol -> allocate memory for rows \n"));
  }
  columns[i].selected = 1;
  index = 0;
  ret = SQLDescribeCol(hstmt, (SQLSMALLINT)j, cName, sizeof(cName),
		       &cNameLen, &cType, &cSize, &cDecDigits, 
		       &cNullable);
  DBG(("dbBindCol -> SQLDescribeCol return %d\n", ret));
  DBG(("dbBindCol -> columnname %s\n", cName));
  DBG(("dbBindCol -> columntype %d\n", cType));
  if(!sql_success(ret)) {
    mes = "error";
    result = "SQLDescribeCol";
    goto exit;
  }
  /* If cType is unknown */
  if (cType == SQL_UNKNOWN_TYPE) {
    mes = "error";
    result = "Column_has_Unknown_data_type";
    goto exit;
  }
  /* The cType SQL_LONGVARCHAR or SQL_LONGVARBINARY has a */
  /* cSize 2 GByte which must be limited to MAXCOLSIZE */
  if(cType == SQL_LONGVARCHAR || cType == SQL_LONGVARBINARY) {
    cSize = MAXCOLSIZE;
  }
  if(!column_type(cType, cSize, cDecDigits, &columns[i])) {
    mes = "error";
    result = "unknown_columntype";
    goto exit;
  }
  DBG(("dbBindCol -> columntypename %s\n",columns[i].typename));
  if (columns[i].len > 0) {
    columns[i].buf = malloc(columns[i].len);
    if (columns[i].buf == NULL) {
      mes = "error";
      result = "No_memory_left";
      goto exit;
    }
    if(cType != SQL_LONGVARCHAR && cType != SQL_LONGVARBINARY) {
      ret = SQLBindCol(hstmt, (SQLSMALLINT)(i+1), columns[i].type, 
		       columns[i].buf,
		       columns[i].len, &columns[i].resultLen);
      if (!sql_success(ret)) {
	mes = "error";
	result = "SQLBindCol";
	goto exit;
      }
    } 
  }
  else {
    columns[i].len = 0;
    columns[i].buf = NULL;
  }
  DBG(("dbBindCol -> columntype  %d\n",cType));

  eimes = ei_encode_string(NULL, &index, cName);
  resultListLen = index + 200;
  mes = "ok";
  buf = malloc(resultListLen);
  if (buf == NULL) {
    mes = "error";
    result = "No_mem_left";
    goto exit;
  }
  DBG(("dbBindCol -> allocate memory  \n")); 
  index = 0;
  eimes = ei_encode_version(buf, &index);
  eimes = ei_encode_tuple_header(buf, &index, 3);
  eimes = ei_encode_atom(buf, &index, mes);
  eimes = ei_encode_string(buf, &index, "dbBindCol");
  eimes = ei_encode_long(buf, &index, ret); 
  DBG(("dbBindCol -> returned index %d\n", index));
  return index;

 exit:
  DBG(("dbBindCol -> mes %s\n", mes));
  SQLCloseCursor(hstmt); 
  SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
  buf = malloc(200); 
  eimes = ei_encode_version(buf, &index);
  eimes = ei_encode_tuple_header(buf, &index, 3);
  eimes = ei_encode_atom(buf, &index, mes);
  eimes = ei_encode_string(buf, &index, result);
  eimes = ei_encode_long(buf, &index, ret);
  DBG(("dbBindCol -> returned result %s \n", result));
  DBG(("dbBindCol -> returned index %d\n", index));
  return index;
}

/* dbDescribeCol returns column name for one column in */
/* the result set. */
/* Inparmeter: Column number (starts with 1) */
int dbDescribeCol(char *in) {
  char *mes, *result;
  int index, j, resultListLen;
  SQLTCHAR cName[255];
  SQLSMALLINT cNameLen, cType, cDecDigits, cNullable;
  SQLUINTEGER cSize; 
  int eimes;
  SQLRETURN ret;	

  DBG(("dbDescribeCol -> translate...\n"));
  DBG(("dbDescribeCol -> inparameter:%s\n", in));
  /* selected column */
  j = atoi(in);
  DBG(("dbDescribeCol -> selected column:%d\n", j));
  /* The columnarray starts at index 0 */

  index = 0;
  ret = SQLDescribeCol(hstmt, (SQLSMALLINT)j, cName, sizeof(cName),
		       &cNameLen, &cType, &cSize, &cDecDigits, 
		       &cNullable);
  DBG(("dbDescribeCol -> SQLDescribeColumn return %d\n", ret));
  if(!sql_success(ret)){ /* error */
    mes = "error";
    result = "SQLDescribeCol";
    goto exit;
  }
  if (sql_success(ret)) {
    DBG(("dbDescribeCol -> columnname %s\n",cName));
    eimes = ei_encode_string(NULL, &index, cName);
    resultListLen = index + 200;
    mes = "ok";
    buf = malloc(resultListLen); 
    if (buf == NULL) {
      mes = "error";
      result = "No_mem_left";
      goto exit;
    }
    DBG(("dbDescribeCol -> allocate memory \n"));
    index = 0;
    eimes = ei_encode_version(buf, &index);
    eimes = ei_encode_tuple_header(buf, &index, 4);
    eimes = ei_encode_atom(buf, &index, mes);
    eimes = ei_encode_long(buf, &index, ret);
    eimes = ei_encode_string(buf, &index, cName);
    eimes = ei_encode_long(buf, &index, cNullable);
    DBG(("dbDescribeCol -> returned index %d\n", index));
    return index;
  } /* if */
  
 exit:
  DBG(("dbDescribeCol -> mes %s\n", mes));
  SQLCloseCursor(hstmt); 
  SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
  buf = malloc(200); 
  eimes = ei_encode_version(buf, &index);
  eimes = ei_encode_tuple_header(buf, &index, 3);
  eimes = ei_encode_atom(buf, &index, mes);
  eimes = ei_encode_string(buf, &index, result);
  eimes = ei_encode_long(buf, &index, ret);
  DBG(("dbDescribeCol -> returned result %s \n", result));
  DBG(("dbDescribeCol -> returned index %d\n", index));
  return index;
}

/* Fetch data from database */  
int dbFetch(char *in) {
  char *mes, *result;
  int eimes, index;
  SQLRETURN ret;	

  index = 0;
  buf = malloc(150); 
  ret = SQLFetch(hstmt);
  DBG(("dbFetch -> SQLFetch return: %d \n", ret));
  if(!sql_success(ret) && ret != SQL_NO_DATA) {
    mes = "error";
    result = "SQLFetch";
    goto exit;
  }
  else {
    mes = "ok";
    eimes = ei_encode_version(buf, &index);
    eimes = ei_encode_tuple_header(buf, &index, 2);
    eimes = ei_encode_atom(buf, &index, mes);
    eimes = ei_encode_long(buf, &index, ret);
    DBG(("dbFetch -> mes %s\n", mes));
    DBG(("dbFetch -> returned index %d\n", index));
    return index;
  }

 exit:
  DBG(("dbFetch -> exit: ...\n"));
  SQLCloseCursor(hstmt); 
  SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
  eimes = ei_encode_version(buf, &index);
  eimes = ei_encode_tuple_header(buf, &index, 3);
  eimes = ei_encode_atom(buf, &index, mes);
  eimes = ei_encode_string(buf, &index, result);
  eimes = ei_encode_long(buf, &index, ret);
  DBG(("dbFetch -> mes %s\n", mes)); 
  DBG(("dbFetch -> returned result %s \n", result));
  DBG(("dbFetch -> returned index %d\n", index));
  return index;
  }


/* Read data from Database */
int dbreadbuffer(char *in) {
  char *mes, *result;
  int index,i, j;
  char *prevptr;
  char sqlState[6];
  int blocklen, prevlen;
  int resultListLen;
  int eimes, ret;

  DBG(("dbreadbuffer -> start \n"));
  /* First find the length of the buf */
  index = 0;
  /* Column number */
  j = atoi(in);
  /* Columnarray start at index 0 */
  i = j - 1;
  DBG(("dbreadbuffer -> selected column %d \n", j));
  DBG(("dbreadbuffer -> columntype: %d \n", columns[i].type));
  DBG(("ddreadbuffer -> columntypename %s\n",columns[i].typename));
  DBG(("dbreadbuffer -> columnbufferlength: %d \n", columns[i].len));
  if(columns[i].len == 0 || columns[i].resultLen == SQL_NULL_DATA) 
    eimes = ei_encode_atom(NULL, &index,"null");
  else {
    switch(columns[i].type) {
    case SQL_C_CHAR:
      eimes = ei_encode_string(NULL, &index, columns[i].buf);
      break;
    case SQL_C_DOUBLE:
      eimes = ei_encode_double(NULL, &index, 
			       *((SQLDOUBLE *) columns[i].buf));
      break;
    case SQL_C_SLONG:
      eimes = ei_encode_long(NULL, &index, 
			     *((SQLINTEGER *) columns[i].buf));
      break;
    case SQL_C_BIT:
      ei_encode_atom(NULL, &index,  columns[i].buf[0]?"true":"false");
      break;
    case SQL_C_BINARY:
      ei_encode_string(NULL, &index,  columns[i].buf);
      break;
    default:
      eimes = ei_encode_atom(NULL, &index,"error");
      break;
    } /*switch */
  } /* else */
  
  /* Putting data into the buffer */
  DBG(("dbreadbuffer -> calculated length of buf %d\n", index));
  resultListLen = index + 84;
  buf = malloc(resultListLen);
  if (buf == NULL) {
    mes = "error";
    result = "No_mem_left";
    ret = -99;
    goto exit;
  }
  DBG(("dbreadbuffer -> allocate memory \n"));
  index = 0;
  eimes = ei_encode_version(buf, &index);
  eimes = ei_encode_tuple_header(buf, &index, 2);
  eimes = ei_encode_atom(buf, &index, "ok");

  if(columns[i].len == 0 || columns[i].resultLen == SQL_NULL_DATA) {
    eimes = ei_encode_atom(buf, &index,"null");
    DBG(("dbreadbuffer -> encode SQL_NULL_DATA \n"));
  }
  else {
    switch(columns[i].type) {
    case SQL_C_CHAR:
      DBG(("dbreadbuffer -> encode SQL_C_CHAR \n"));
      eimes = ei_encode_string(buf, &index, columns[i].buf);
      break;
    case SQL_C_SLONG:
      DBG(("dbreadbuffer -> encode SQL_C_SLONG \n"));
      eimes = ei_encode_long(buf, &index, 
			     *((SQLINTEGER *) columns[i].buf));
      break;
   case SQL_C_DOUBLE:
      DBG(("dbreadbuffer -> encode SQL_C_DOUBLE \n"));
      eimes = ei_encode_double(buf, &index, 
			       *((SQLDOUBLE *) columns[i].buf));
      break;
    case SQL_C_BIT:
      DBG(("dbreadbuffer -> encode SQL_C_BIT \n"));
      ei_encode_atom(buf, &index,  columns[i].buf[0]?"true":"false");
      break;
    case SQL_C_BINARY:
      DBG(("readBuffer -> encode SQL_C_BINARY \n"));
      blocklen = columns[i].len;
      prevptr = columns[i].buf;
      ret = SQLGetData(hstmt, (SQLSMALLINT)(i+1), SQL_C_CHAR, prevptr,
		       blocklen, &columns[i].resultLen);
      DBG(("dbreadbuffer -> SQLGetData returns: %d\n", ret));
      DBG(("dbreadbuffer -> columns buffer %s\n", columns[i].buf));
      DBG(("dbreadbuffer -> SQLGetData colsize %d\n",
	   blocklen ));
      DBG(("dbreadbuffer -> SQLGetData resultLen %ld\n",
	   columns[i].resultLen));
      if (ret == SQL_SUCCESS_WITH_INFO) 
	do {
	  eimes = getDiagnos(SQL_HANDLE_STMT, hstmt,sqlState);
	  /* increase columnbuffer with blocklen */
	  if(columns[i].resultLen == SQL_NO_TOTAL ||
	     strcmp(sqlState, "01004") == 0) {
	    prevlen = columns[i].len - 1;
	    columns[i].len = prevlen +  blocklen;
	    prevptr = realloc(columns[i].buf, columns[i].len);
	    if (prevptr == NULL) {
	      free(columns[i].buf);
	      mes = "error";
	      result = "malloc_No_memory_left";
	      ret = -99;
	      goto exit;
	    }
	    columns[i].buf = prevptr;
	    prevptr = columns[i].buf + prevlen;
	    DBG(("dbreadbuffer -> columns buffer size %d\n",
		 columns[i].len));
	    ret = SQLGetData(hstmt,(SQLSMALLINT)(i+1), SQL_C_CHAR, prevptr,
			     blocklen, &columns[i].resultLen);
	    DBG(("dbreadbuffer -> SQLGetData returns %d\n", ret));
	    DBG(("dbreadbuffer -> columns buffer %s\n", columns[i].buf));
	  }
	} while (ret == SQL_SUCCESS_WITH_INFO && ret != SQL_ERROR);
      if (ret == SQL_NO_DATA) {
	DBG(("dbreadbuffer -> ret == SQL_NO_DATA \n"));
	ei_encode_string(buf, &index, columns[i].buf);
	break;
      }
      if (ret == SQL_SUCCESS) {
	DBG(("dbreadbuffer -> ret == SQL_SUCCESS \n"));
	
	DBG(("dbreadbuffer -> columns buffer %s\n", columns[i].buf));
	ei_encode_string(buf, &index, columns[i].buf);
	break;
      }
      if (ret == SQL_ERROR) {
	mes = "error";
	result = "SQLGetData";
	goto exit;
      }
      break;
    default:
      eimes = ei_encode_atom(buf, &index,"error");
      break;
    } /* switch */
  } /* else */
  DBG(("dbreadbuffer -> index  %d\n", index));
  return index;
 exit:
  DBG(("dbreadbuffer -> returned mes %s\n", mes));
  index = 0;
  buf = malloc(200);
  eimes = ei_encode_version(buf, &index);
  eimes = ei_encode_tuple_header(buf, &index, 3);
  eimes = ei_encode_atom(buf, &index, mes);
  eimes = ei_encode_string(buf, &index, result);
  eimes = ei_encode_long(buf, &index, ret); 
  DBG(("dbreadbuffer -> returned result %s \n", result));
  DBG(("dbreadbuffer -> returned index %d\n", index));
  return index;
}

int dbCloseCursor(char *in) {
  char *mes, *result;
  int i, index;
  int eimes;
  SQLRETURN ret;	

  ret = SQLCloseCursor(hstmt);
  DBG(("dbCloseCursor -> SQLCloseCursor returns %d\n", ret));
  if (!sql_success(ret)) {
    mes = "error";
    result = "SQLCloseCursor";
    goto exit;
  }
  if (hstmt != NULL) {
    ret = SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
    DBG(("dbCloseCursor -> SQLFreeHandle returns %d\n", ret));
    if (!sql_success(ret)) {
      mes = "error";
      result = "SQLSQLFreeHandle";
      goto exit;
    }
  }
  mes = "ok";
  result = "dbCloseCursor";
 exit:
  DBG(("dbCloseCursor -> exit: ...\n"));
  DBG(("dbCloseCursor -> deallocate memory \n"));
  if (columns != NULL) {
    for (i = 0; i < numberColumn; i++) {
      if(columns[i].selected && columns[i].len > 0 
	&& columns[i].buf != NULL) 
	free(columns[i].buf);
    }
    free(columns);
  }
  numberColumn = 0;
  DBG(("dbCloseCursor -> returned mes %s\n", mes));
  index = 0;
  buf = malloc(200);
  eimes = ei_encode_version(buf, &index);
  eimes = ei_encode_tuple_header(buf, &index, 3);
  eimes = ei_encode_atom(buf, &index, mes);
  eimes = ei_encode_string(buf, &index, result);
  eimes = ei_encode_long(buf, &index, ret); 
  DBG(("dbCloseCursor -> returned result %s \n", result));
  DBG(("dbCloseCursor -> returned index %d\n", index));
  return index;
}

/* close the connection to the database */
int dbCloseConnection(char *in) {

  char *mes, *result;
  int index;
  int eimes;
  SQLRETURN ret;

  if(!connected) {
    mes = "error";
    result = "allready_disconnected";
    ret = -1;
    goto exit;
  }
    ret = SQLDisconnect(hdbc1);
    DBG(("dbCloseConnection -> SQLDisconnect return %d \n", ret));
    if (!sql_success(ret)) {
      mes = "error";
      result = "SQLDisconnect";
      goto exit;
    }
    ret = SQLFreeHandle(SQL_HANDLE_DBC, hdbc1);
    DBG(("dbCloseConnection -> SQLFreeHandle return %d \n", ret));
    if (!sql_success(ret)) {
      mes = "error";
      result = "SQLFreeHandle";
      goto exit;
    }
    ret = SQLFreeHandle(SQL_HANDLE_ENV, henv1);
    DBG(("dbCloseConnection -> SQLFreeHandle return %d \n", ret));
    if (!sql_success(ret)) {
      mes = "error";
      result = "SQLFreeHandle";
      goto exit;
    }
    connected = 0;
    mes = "ok";
    result = "disconnected";
 exit:
    index = 0;
    buf = malloc(200);
    eimes = ei_encode_version(buf, &index);
    eimes = ei_encode_tuple_header(buf, &index, 3);
    eimes = ei_encode_atom(buf, &index, mes);
    eimes = ei_encode_string(buf, &index, result);
    eimes = ei_encode_long(buf, &index, ret); 
    DBG(("dbCloseConnection -> mes %s \n", mes));
    DBG(("dbCloseConnection -> returned result %s \n", result));
    DBG(("dbCloseConnection -> returned index %d\n", index));
    return index;
}


int main() {
  int len,lensend, index, ermes,writerr;
  char *cmdbuf;

#ifdef WIN32
  _setmode(_fileno( stdin),  _O_BINARY);
  _setmode(_fileno( stdout), _O_BINARY);
#endif

  DBG(("main() -> we start mow \n"));
  /* As long as there is something to read from Erlang */
  /* Read first two bytes, length indicator */
  while ((len = receivedlength()) > 0) {
    /* make the buffert of size len byte */
    buf = malloc(len); 
    if (buf == NULL) {
      DBG(("main() -> could not get any memory \n"));
      break;
    }
    DBG(("main() -> alloc buf of size  %d \n", len));
    cmdbuf = malloc(len);
    if (cmdbuf == NULL) {
      DBG(("main() -> could not get any memory \n"));
      break;
    }
    DBG(("main() -> alloc buf of size  %d \n", len));
    if (read_exact(buf, len) <= 0) {
      DBG(("main() -> received a messages with length: %d \n", len));
      break;
    }
    index = 0;
    /* Decode data in buf to a string and put in cmdbuf */
    ermes = ei_decode_version(buf, &index, &len);
    ermes = ei_decode_string(buf, &index, cmdbuf); 
	  
    DBG(("main() -> received messages: %s \n",cmdbuf));
    free(buf); 
    lensend = handleRequest(cmdbuf);
    free(cmdbuf);
    DBG(("main() -> Lensend is %d\n",lensend));
    if (lensend > 0) {
      writerr = write_cmd(buf, lensend);
      DBG(("main() -> write_cmd returns: %d\n",writerr));
      free(buf);
      if(writerr <= 0) {
	DBG(("main() -> could not write to erlang. \n"));
	break;
      }
    }
    if (lensend < 0) {
      DBG(("main() -> error do not know the command \n"));
      break;
    }
  } /* while */
  DBG(("main() -> after while loop: len; %d\n",len));
  return EXIT_SUCCESS;

}


