%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%

-module(odbc).

-behaviour(gen_server).


%% External exports------------------------------------
%% Basic API
-export([sqlBindColumn/3, sqlBindColumn/4]).
-export([sqlCloseCursor/1, sqlCloseCursor/2]).
-export([sqlConnect/4, sqlConnect/5]).
-export([sqlDescribeCol/2, sqlDescribeCol/3]).
-export([sqlDisConnect/1, sqlDisConnect/2]).
-export([sqlEndTran/2, sqlEndTran/3]).
-export([sqlExecDirect/2, sqlExecDirect/3]).
-export([sqlFetch/1, sqlFetch/2]).
-export([sqlNumResultCols/1, sqlNumResultCols/2]).
-export([sqlRowCount/1, sqlRowCount/2]).
-export([sqlSetConnectAttr/3, sqlSetConnectAttr/4]).
-export([columnRef/0]).
-export([readData/2, readData/3]).

%% Utility API
-export([erl_connect/2, erl_connect/3, erl_connect/4, erl_connect/5]).
-export([erl_executeStmt/2, erl_executeStmt/3]).
-export([erl_disconnect/1, erl_disconnect/2]).

%% Start/stop
-export([start_link/2, start_link/3]).
-export([stop/1, stop/2]).

%%------------------------------------------------------
%% Internal exports (for debug)
-export([get_state/1]).
%% debuge start/stop debugging of erlang
-export([debugerl/2]).
%%  debugc start/stop debugging of c-program 
-export([debugc/2])
.
%% Internal exports (for supervisor).
-export([start_link_sup/2, start_link_sup/3]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, 
	 terminate/2, code_change/3]).

%% Internal state
-record(state, {port,                        % The port to the c-program
		connected = false,           % Connected to database
		connection_string,           % connectionsstring 
		fetchdata = false,           % fetched data before read
		from,			     % 
		clientpid,                   % Client Pid
		columnvector = []            % memoryref for columns
	       }).

%%------------------------------------------------------
%% Internal macros
%% Path to the c-program.
-define(ServerDir, filename:nativename(
		     filename:join(code:priv_dir(odbc), "bin"))).

%% Name of the C program 
-define(ServerProg, "odbcserver").

%% Default timeout.
-define(DefaultTimeout, 5000).


%% These functions are deprecated and is included only for 
%% compatibility reason. Some of them do nothing and some 
%% are replaced by new function.
%% Deprecated functions
-export([sql_alloc_handle/3, sql_alloc_handle/4]).
-export([sql_bind_col/4, sql_bind_col/5]).
-export([sql_close_cursor/2, sql_close_cursor/3]).
-export([sql_connect/5, sql_connect/6]).
-export([sql_describe_col/4, sql_describe_col/5]).
-export([sql_disconnect/2, sql_disconnect/3]).
-export([sql_driver_connect/5, sql_driver_connect/6]).
-export([sql_end_tran/4, sql_end_tran/5]).
-export([sql_exec_direct/3, sql_exec_direct/4]).
-export([sql_fetch/2, sql_fetch/3]).
-export([sql_free_handle/3, sql_free_handle/4]).
-export([sql_get_connect_attr/4, sql_get_connect_attr/5]).
-export([sql_get_diag_rec/5, sql_get_diag_rec/6]).
-export([sql_num_result_cols/2, sql_num_result_cols/3]).
-export([sql_row_count/2, sql_row_count/3]).
-export([sql_set_connect_attr/5, sql_set_connect_attr/6]).
-export([sql_set_env_attr/5, sql_set_env_attr/6]).
-export([alloc_buffer/3, alloc_buffer/4]).
-export([dealloc_buffer/2, dealloc_buffer/3]).
-export([read_buffer/2, read_buffer/3]).
-export([init_env/1, init_env/2]).
-export([connect/3, connect/4, connect/5, connect/6]).
-export([execute_stmt/3, execute_stmt/4]).
-export([disconnect/2, disconnect/3]).
-export([terminate_env/2, terminate_env/3]).
-export([display_size/2]).


%%------------------------------------------------------
%% Info and debug function
%% to do use debug in runtime tools for tracing
i(F, A) ->
    i(get(dbge), F, A).
i(true, F, A) ->
    io:format(F ++ "~n", A);
i(_, _, _) ->
    ok.

%% Debug Erlang-program start with Cmd = on and stop with Cmd = off.
debugerl(Server,Cmd) ->
    debugerl(Server, Cmd, ?DefaultTimeout).
debugerl(Server, Cmd, Timeout) ->
    Ret = gen_server:call(Server,{debuge, Cmd}, Timeout),
    case Ret of
	debugeok ->
	    ok;
	Other ->
	    Other
    end.

%% Debug C-program start with Cmd = on and stop with Cmd = off.
debugc(Server,Cmd) ->
    debugc(Server, Cmd, ?DefaultTimeout).
debugc(Server, Cmd, Timeout) ->
    CmdS = 
	case Cmd of
	    on -> "on";
	    off -> "off"
	end,
    CmdStr = lists:flatten(["debugs", CmdS]),
    Ret = gen_server:call(Server,{debugc, CmdStr}, Timeout),
    case Ret of
	debugcok ->
	    ok;
	Other ->
	    Other
    end.

%%%----------------------------------------------------------
%%% Start and Stop
%%%----------------------------------------------------------

%% start_link/[2, 3]
%%
%% Starts the ODBCE server and the C node.
%% Links to the server to the calling process.
%% Registers the new server with the supervisor.
%%
%% start_link(ServerName, Args, Options, Client) ->
%% start_link(Args, Options, Client) ->

start_link(Args, Options) ->
    supervisor:start_child(odbc_sup, [[{client, self()} | Args], Options]).

start_link(ServerName, Args, Options) ->
    supervisor:start_child(odbc_sup,
			   [ServerName, [{client, self()} | Args], Options]).


%% start_link_sup/[2, 3]
%%
%% Called by the supervisor to start a new server instance.
%% (start_link calls the supervisor which calls start_link_sup.)

start_link_sup(Args, Options) ->
    gen_server:start_link(?MODULE, Args, Options).

start_link_sup(ServerName, Args, Options) ->
    gen_server:start_link(ServerName, ?MODULE, Args, Options).

%% stop/[1, 2]
%%
%% Stops the ODBCE server and the C program.
%%
%% stop(Server) ->
%% stop(Server, Timeout) ->

stop(Server) ->
    stop(Server, ?DefaultTimeout).

stop(Server, Timeout) ->
    gen_server:call(Server, stop, Timeout).



%%%--------------------------------------------------------------------------
%%% Basic API
%%%
%%%--------------------------------------------------------------------------

%% sqlBindColumn/[3, 4]
%% Assigns a referens to a column
%% (binds a referens to a  column).
%%
%% sqlBindColumn(Server, ColNum, Ref) ->
%% sqlBindColumn(Server, ColNum, Ref, Timeout) ->

sqlBindColumn(Server, ColNum, Ref) ->
    sqlBindColumn(Server, ColNum, Ref, ?DefaultTimeout).
sqlBindColumn(Server, ColNum, Ref, Timeout) 
  when integer(ColNum), reference(Ref) ->
    CmdStr = "bincol",
    Ret = gen_server:call(Server,{bindcolumn, CmdStr, ColNum, Ref},
			  Timeout),
    i("sqlBindColumn gen_server return: ~p",[Ret]),
    case Ret of
	{ok, RetMsg, RetCode} ->
	    RetCode;
	{error, ErrMsg, ErrCode} ->
	    {error, ErrMsg, ErrCode};
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end;
sqlBindColumn(Server, ColNum, Ref, Timeout)
  when integer(ColNum) ->
    exit({badarg,
	  sqlBindColumn, 3,
	  {"Not a reference", Ref}});
sqlBindColumn(Server, ColNum, Ref, Timeout)
  when reference(Ref) ->
    exit({badarg,
	  sqlBindColumn, 2,
	  {"Not a integer", ColNum}}).


%% sqlCloseCursor/[1, 2]
%% Closes a cursor that has been opened on a statement and discards pending 
%% results.
%%
%% sqlCloseCursor(Server) ->
%% sqlCloseCursor(Server, Timeout) ->

sqlCloseCursor(Server) ->
    sqlCloseCursor(Server, ?DefaultTimeout).

sqlCloseCursor(Server, Timeout) ->
    CmdStr = "closec",
    Ret = gen_server:call(Server,{closecursor, CmdStr},Timeout),
    i("sqlCloseCursor gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, ErrCode} ->
	    {error, ErrMsg, ErrCode};
	{ok, Result, ReturnCode} ->
	    ReturnCode;
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.


%% sqlConnect/[4, 5]
%% Establishes a connection to a driver and a database.
%%
%% sqlConnect(Server, DSN, UID, PWD) ->
%% sqlConnect(Server, DSN, UID, PWD, Timeout) ->

sqlConnect(Server, DSN, UID, PWD) ->
    sqlConnect(Server, DSN, UID, PWD, ?DefaultTimeout).

sqlConnect(Server, DSN, UID, PWD, Timeout) 
  when list(DSN), list(UID), list(PWD) ->
%% When there are not any user passsword
    ConnectStr =  lists:flatten(["DSN=",DSN,";UID=",UID,";PWD=",PWD]),
    CmdStr = "opendb",
    Ret = gen_server:call(Server,{openconnection, CmdStr, ConnectStr},
			  Timeout),
    i("sqlConnect gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, ErrCode} ->
	    {error, ErrMsg, ErrCode};
	{ok, Result, ReturnCode} ->
	    ReturnCode;
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end;
sqlConnect(Server, DSN, UID, PWD, Timeout) 
  when list(DSN), list(UID) ->
    exit({badarg,
	  sqlConnect, 4,
	  {"Not a string", PWD}});    
sqlConnect(Server, DSN, UID, PWD, Timeout) 
  when list(DSN) ->
    exit({badarg,
	  sqlConnect, 3,
	  {"Not a string", UID}});
sqlConnect(Server, DSN, UID, PWD, Timeout) ->
    exit({badarg,
	  sqlConnect, 2,
	  {"Not a string", DSN}}).

%% sqlDescribeCol/[2, 3]
%% Returns the result descriptor -- column name, type, column size,
%% decimal digits, and nullability -- for one column in the result set.
%%
%% sqlDescribeCol(Server, ColNum) ->
%% sqlDescribeCol(Server, ColNum, Timeout) ->

sqlDescribeCol(Server, ColNum) ->
    sqlDescribeCol(Server, ColNum, ?DefaultTimeout).

sqlDescribeCol(Server, ColNum, Timeout) when integer(ColNum) ->
    CmdStr = lists:flatten(["descol", integer_to_list(ColNum)]),
    Ret = gen_server:call(Server,{describeColumn, CmdStr},Timeout),
    i("sqlDescribeCol gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, ErrCode} ->
	    {error, ErrMsg, ErrCode};
	{ok, ReturnCode, CName, CNullable} ->
	    {ReturnCode, CName, CNullable};
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end;
sqlDescribeCol(Server, ColNum, Timeout) ->
    exit({badarg,
	  sqlDescribeCol, 2,
	  {"Not a Integer", ColNum}}).    


%% sqlDisConnect/[1, 2]
%% Close a database connection.
%%
%% sqlDisConnect(Server) ->
%% sqlDisconnect(Server, Timeout) ->

sqlDisConnect(Server) ->
    sqlDisConnect(Server, ?DefaultTimeout).

sqlDisConnect(Server, Timeout) ->
    CloseStr = "closdb",
    Ret = gen_server:call(Server,{closeconnection, CloseStr}, Timeout),
    i("sqlDisConnect gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded disconnecting  
	{ok, Mes,RetCode} ->
	    RetCode;
	% Did not succeded disconnecting (may already be disconnected)
	{error, ErrMsg, ErrCode} ->
	    {error, ErrMsg, ErrCode};
	Other ->
	    Other
    end.


%% sqlEndTran/[2, 3]
%%
%% Requests a commit or rollback operation for all active operations on all 
%% statement handles associated with a connection. It can only request a 
%% commit or rollback operation for a connection.
%%
%% sqlEndTran(Server, ComplType) ->
%% sqlEndTran(Server, ComplType, Timeout) ->

sqlEndTran(Server, ComplType) ->
    sqlEndTran(Server, ComplType, ?DefaultTimeout).

sqlEndTran(Server, ComplType, Timeout) ->
    AttrStr = lists:flatten([integer_to_list(ComplType), ";"]),
    CmdStr = lists:flatten(["endtra", AttrStr]),

    Ret = gen_server:call(Server,{endTran, CmdStr},Timeout),  
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, ErrCode} ->
	    {error, ErrMsg, ErrCode};
	{ok, ReturnMsg, ReturnCode} ->
	    ReturnCode;
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.


%% sqlExecDirect/[2, 3]
%% Executes a preparable statement.
%%
%% sqlExecDirect(Server, Stmt) ->
%% sqlExecDirect(Server, Stmt, Timeout) ->

sqlExecDirect(Server, Stmt) ->
    sqlExecDirect(Server, Stmt, ?DefaultTimeout).

sqlExecDirect(Server, Stmt, Timeout) when list(Stmt) ->
    SqlStmt = lists:flatten(["exedir", Stmt]),
    Ret = gen_server:call(Server, {execdir,SqlStmt}, Timeout),
    i("sqlExecDirect gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, Errcode} ->
	    {error,ErrMsg,Errcode};
	{ok, Result, ReturnCode} ->
	    ReturnCode;
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end;
sqlExecDirect(Server, Stmt, Timeout) ->
    exit({badarg,
	  sqlExecDirect, 2,
	  {"Not a string", Stmt}}).    


%% sqlFetch/[1, 2]
%% Fetches a row of data from a result set. The driver returns data for 
%% all columns that were bound to a referens with sqlBindCol/[3, 4].
%%
%% sqlFetch(Server) ->
%% sqlFetch(Server, Timeout) ->

sqlFetch(Server) ->
    sqlFetch(Server, ?DefaultTimeout).

sqlFetch(Server, Timeout) ->
    CmdStr = "fetchd",
    Ret = gen_server:call(Server,{fetch, CmdStr},Timeout),
    i("sqlFetct gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, ErrCode} ->
	    {error, ErrMsg, ErrCode};
	{ok, ReturnCode} ->
	    ReturnCode;
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.


%% sqlNumResultCols/[1, 2]
%% Returns the number of columns in a result set.
%%
%% sqlNumResultCols(Server) ->
%% sqlNumResultCols(Server, Timeout) ->

sqlNumResultCols(Server) ->
    sqlNumResultCols(Server, ?DefaultTimeout).

sqlNumResultCols(Server, Timeout) ->
    CmdStr = "rescol",
    Ret = gen_server:call(Server,{numResultCols, CmdStr},Timeout),
    i("sqlNumResultCols gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, ErrCode} ->
	    {error, ErrMsg, ErrCode};
	{ok, NumberColumn, ReturnCode} ->
	    {ReturnCode, NumberColumn};
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.


%% sqlRowCount/[1, 2]
%% Returns the number of rows affected by an UPDATE, INSERT, or DELETE 
%% statement.
%%
%% sqlRowCount(Server) ->
%% sqlRowCount(Server, Timeout) ->

sqlRowCount(Server) ->
    sqlRowCount(Server, ?DefaultTimeout).

sqlRowCount(Server, Timeout) ->
    CmdStr = "rowcon",
    Ret = gen_server:call(Server,{rowCount, CmdStr},Timeout),
    i("sqlRowCount gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, ErrCode} ->
	    {error, ErrMsg, ErrCode};
	{ok, NumberColumn, ReturnCode} ->
	    {ReturnCode, NumberColumn};
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.
    

%% sqlSetConnectAttr/[3, 4]
%%
%% Sets attributes that govern aspects of connections.
%% Driver-specific
%% attributes are not supported through macros, but can be set, if  they
%% are strings or signed/unsigned long integers.
%%
%% sqlSetConnectAttr(Server,Attr, Value) ->
%% sqlSetConnectAttr(Server, Attr, Value, Timeout) ->

sqlSetConnectAttr(Server, Attr, Value) ->
    sqlSetConnectAttr(Server, Attr, Value, ?DefaultTimeout).

sqlSetConnectAttr(Server, Attr, Value, Timeout) ->
    AttrStr = lists:flatten([integer_to_list(Attr), ";"]),
    ValueStr = if 
		   number(Value) ->
		       lists:flatten(["0;", integer_to_list(Value)]);
		   list(Value) ->
		       lists:flatten(["1;", Value]);
		   true ->
		       exit({sqlSetConnectAttr, wrong_Value_type})
	       end,
    CmdStr = lists:flatten(["setatr", AttrStr, ValueStr, ";"]),

    Ret = gen_server:call(Server,{setConnectAttr, CmdStr},Timeout),  
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, ErrCode} ->
	    {error, ErrMsg, ErrCode};
	{ok, Result, ReturnCode} ->
	    ReturnCode;
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.


%% columnRef/[0]
%% Generate a referens for a column. 
%%
%% columnRef() ->

columnRef() ->
    Ref = erlang:make_ref(),
    {ok, Ref}.


%% readData/[2, 3]
%% Returns data from the sqlFetch/[1,2] call.
%%
%% readData(Server, Ref) ->
%% readData(Server, Ref, Timeout) ->

readData(Server, Ref) ->
    readData(Server, Ref, ?DefaultTimeout).

readData(Server, Ref, Timeout) ->
    CmdStr = "readbu",
    Ret = gen_server:call(Server, {readbuffer, CmdStr, Ref},Timeout),
    i("readData gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in get data
	{ok, Data} ->
	    {ok, Data};
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.


%%%----------------------------------------------------------------------
%%% Utility API
%%%
%%%----------------------------------------------------------------------

%% erl_connect/[2, 3, 4, 5]
%% Opens a connection to a database. There can be only one open database 
%% connection per server.
%%
%% erl_connect(Server, ConnectStr) ->
%% erl_connect(Server, ConnectStr, Timeout) ->
%% erl_connect(Server, DSN, UID, PWD) ->
%% erl_connect(Server, DSN, UID, PWD, Timeout) ->

erl_connect(Server, ConnectStr) ->
    erl_connect(Server, ConnectStr, ?DefaultTimeout).

erl_connect(Server, ConnectStr, Timeout) when list(ConnectStr) ->
    CmdStr = "opendb",
    Ret = gen_server:call(Server,{openconnection, CmdStr, ConnectStr}, 
			  Timeout),
    i("erl_connect gen_server call returns: ~p",[Ret]),
    case Ret of
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	% Succeeded connected 
	{ok, Msg, RetCode} ->
	    ok;
	% Did not succed connected
	{error, ErrMsg, ErrCode} ->
	    {error,ErrMsg, ErrCode};
	Other ->
	    Other
    end;
erl_connect(Server, ConnectStr, Timeout) ->
    exit({badarg,
	  erl_connect, 2,
	  {"Not a string", ConnectStr}}).    

erl_connect(Server, DSN, UID, PWD) ->
    erl_connect(Server, DSN, UID, PWD, ?DefaultTimeout).

erl_connect(Server, DSN, UID, PWD, Timeout) 
  when list(DSN), list(UID), list(PWD) ->
    ConnectStr = lists:flatten(["DSN=",DSN,";UID=",UID,";PWD=",PWD]),
    erl_connect(Server, ConnectStr, Timeout);

erl_connect(Server, DSN, UID, PWD, Timeout) 
  when list(DSN), list(UID) ->
    exit({badarg,
	  erl_connect, 4,
	  {"Not a string", PWD}});

erl_connect(Server, DSN, UID, PWD, Timeout) 
  when list(DSN) ->
        exit({badarg,
	  erl_connect, 3,
	  {"Not a string", UID}});

erl_connect(Server, DSN, UID, PWD, Timeout) ->
        exit({badarg,
	  erl_connect, 2,
	  {"Not a string", DSN}}).


%% erl_executeStmt/[2, 3]
%% Executes a single SQL statement. All changes to the data source are
%% automatically committed if successful.
%%
%% erl_executeStmt(Server, Stmt) ->
%% erl_executeStmt(Server, Stmt, Timeout) ->

erl_executeStmt(Server, Stmt) ->
    erl_executeStmt(Server, Stmt, ?DefaultTimeout).
erl_executeStmt(Server, Stmt, Timeout) when list(Stmt) ->
    SqlStmt = lists:flatten(["execdb", Stmt]),
    % SqlStmt = {execdb, Stmt},
    gen_server:call(Server,{execute, SqlStmt}, Timeout);
erl_executeStmt(Server, Stmt, Timeout) ->
    exit({badarg,
	  erl_executeStmt, 2,
	  {"Not a string", Stmt}}).


%% erl_disconnect/[1, 2]
%% Closes the connection to a database.
%%
%% erl_disconnect(Server) ->
%% erl_disconnect(Server, Timeout) ->

erl_disconnect(Server) ->
    erl_disconnect(Server, ?DefaultTimeout).

erl_disconnect(Server, Timeout) ->
    CloseStr = "closdb",
    Ret = gen_server:call(Server,{closeconnection, CloseStr}, Timeout),
    i("erl_disconnect gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded disconnecting  
	{ok, Mes, RetCode} ->
	    ok;
	% Did not succeded disconnecting (may already be disconnected)
	{error, ErrMsg, ErrFunc} ->
	    {error, ErrMsg, ErrFunc};

	Other ->
	    {error, Other}
    end.


%%---------------------------------------------------------------------------
%%
%% Depricated functions
%%---------------------------------------------------------------------------

%% This function is deprecated and does nothing.
%% It is included only for compatibility reason.
%%--------------------------------------------------------------------------- 
sql_alloc_handle(Server, HandleType, RefInputHandle) ->
    sql_alloc_handle(Server, HandleType, RefInputHandle, ?DefaultTimeout).

sql_alloc_handle(Server, HandleType, RefInputHandle, Timeout) ->
    {0,void}. 

%% This function is deprecated and replaced by the function 
%% sqlBindColumn/[3,4].
%%---------------------------------------------------------------------------
sql_bind_col(Server, RefStmtHandle, ColNum, RefBuf) ->
    sql_bind_col(Server, RefStmtHandle, ColNum, RefBuf, ?DefaultTimeout).

sql_bind_col(Server, RefStmtHandle, ColNum, RefBuf, Timeout) ->
    Ret = sqlBindColumn(Server, ColNum, RefBuf, Timeout),
    case Ret of
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	Other ->
	    Other
    end.

%% This function is deprecated and replaced by the function
%% sqlCloseCursor/[1,2].
%%---------------------------------------------------------------------------
sql_close_cursor(Server, RefStmtHandle) ->
    sql_close_cursor(Server, RefStmtHandle, ?DefaultTimeout).

sql_close_cursor(Server, RefStmtHandle, Timeout) ->
    Ret = sqlCloseCursor(Server, Timeout),
    case Ret of
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	Other ->
	    Other
    end.

%% This function is deprecated and replaced by the function
%% sqlConnect/[4,5].
%%---------------------------------------------------------------------------
sql_connect(Server, RefConnHandle, DSN, UID, Auth) ->
    sql_connect(Server, RefConnHandle, DSN, UID, Auth, ?DefaultTimeout).

sql_connect(Server, RefConnHandle, DSN, UID, Auth, Timeout) ->
    Ret = sqlConnect(Server, DSN, UID, Auth, Timeout),
    case Ret of
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	Other ->
	    Other
    end.

%% This function is deprecated and repleced by the function
%% sqlDescribeCol/[2,3].
%%---------------------------------------------------------------------------
sql_describe_col(Server, RefStmtHandle, ColNum, BufLenColName) ->
    sql_describe_col(Server, RefStmtHandle, ColNum, BufLenColName, 
		     ?DefaultTimeout).

sql_describe_col(Server, RefStmtHandle, ColNum, BufLenColName, Timeout) ->
    Ret = sqlDescribeCol(Server, ColNum, Timeout),
    case Ret of
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	{ReturnCode, ColName, CNullable} ->
	    {ReturnCode, {ColName, -1}, -1, -1, -1, CNullable};
	Other ->
	    Other
    end.

%% This function is deprecated and replaced by the function
%% sqlDisConnect/[1,2].
%%---------------------------------------------------------------------------
sql_disconnect(Server, RefConnHandle) ->
    sql_disconnect(Server, RefConnHandle, ?DefaultTimeout).

sql_disconnect(Server, RefConnHandle, Timeout) ->
    Ret = sqlDisConnect(Server, Timeout),
    case Ret of
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	Other ->
	    Other
    end.	

%% This function is deprecated and does nothing.
%% It is included only for compatibility reason.
%%--------------------------------------------------------------------------
sql_driver_connect(Server, RefConnHandle, InConnStr,
		   BufLenOutConnStr, DrvCompletion) ->
    sql_driver_connect(Server, RefConnHandle, InConnStr,
		       BufLenOutConnStr, DrvCompletion, ?DefaultTimeout).

sql_driver_connect(Server, RefConnHandle, InConnStr,
		   BufLenOutConnStr, DrvCompletion, Timeout) ->
    {0, void}.


%% This function is deprecated and replaced by the function 
%% sqlEndTran/[2,3]
%%---------------------------------------------------------------------------
sql_end_tran(Server, HandleType, RefHandle, ComplType) ->
    sql_end_tran(Server, HandleType, RefHandle, ComplType, ?DefaultTimeout).

sql_end_tran(Server, HandleType, RefHandle, ComplType, Timeout) ->
    Ret = sqlEndTran(Server, ComplType, Timeout),
    case Ret of
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	Other ->
	    Other
    end.


%% This function is deprecated and replaced by the function
%% sqlExecDirect/[2,3].
%%---------------------------------------------------------------------------
sql_exec_direct(Server, RefStmtHandle, Stmt) ->
    sql_exec_direct(Server, RefStmtHandle, Stmt, ?DefaultTimeout).


sql_exec_direct(Server, RefStmtHandle, Stmt, Timeout) ->
    Ret = sqlExecDirect(Server, Stmt, Timeout),
    case Ret of
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	Other ->
	    Other
    end.

%% This function is deprecated and replaced by the function
%% sqlFetch/[1,2].
%%---------------------------------------------------------------------------
sql_fetch(Server, RefStmtHandle) ->
    sql_fetch(Server, RefStmtHandle, ?DefaultTimeout).

sql_fetch(Server, RefStmtHandle, Timeout) ->
    Ret = sqlFetch(Server, Timeout),
    case Ret of
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	Other ->
	    Other
    end.


%% This function is deprecated and does nothing.
%% It is included only for compatibility reason.
%%---------------------------------------------------------------------------
sql_free_handle(Server, HandleType, RefHandle) ->
    sql_free_handle(Server, HandleType, RefHandle, ?DefaultTimeout).

sql_free_handle(Server, HandleType, RefHandle, Timeout) ->
    0.

%% This function is deprecated and does nothing
%% It is included only for compatibility reason.
%%---------------------------------------------------------------------------
sql_get_connect_attr(Server, RefConnHandle, Attr, BufType) ->
    sql_get_connect_attr(Server, RefConnHandle, Attr, BufType, 
			 ?DefaultTimeout).

sql_get_connect_attr(Server, RefConnHandle, Attr, BufType, Timeout) ->
    {0, -1}.

%% This function is deprecated and does nothing.
%% It is included only for compatibility reason.
%%---------------------------------------------------------------------------
sql_get_diag_rec(Server, HandleType, RefHandle, RecNum, BufLenErrMsg) ->
    sql_get_diag_rec(Server, HandleType, RefHandle, RecNum, BufLenErrMsg,
		     ?DefaultTimeout).

sql_get_diag_rec(Server, HandleType, RefHandle, RecNum, BufLenErrMsg, 
		 Timeout) ->
   0.

%% This function is deprecated and replaced by the function
%% sqlNumResultCols/[1,2].
%%---------------------------------------------------------------------------
sql_num_result_cols(Server, RefStmtHandle) ->
    sql_num_result_cols(Server, RefStmtHandle, ?DefaultTimeout).

sql_num_result_cols(Server, RefStmtHandle, Timeout) ->
    Ret = sqlNumResultCols(Server, Timeout),
    case Ret of
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	 Other ->
	    Other
    end.    

%% This function is deprecated and replaced by the function
%% sql_rowCount/[1,2].
%%---------------------------------------------------------------------------
sql_row_count(Server, RefStmtHandle) ->
    sql_row_count(Server, RefStmtHandle, ?DefaultTimeout).

sql_row_count(Server, RefStmtHandle, Timeout) ->
    Ret = sqlRowCount(Server, Timeout),
    case Ret of
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	 Other ->
	    Other
    end.

%% This function is deprecated and replaced by the function.
%% sqlSetConnectAttr/[4, 5].
%%---------------------------------------------------------------------------
sql_set_connect_attr(Server, RefConnHandle, Attr, Value, BufType) ->
    sql_set_connect_attr(Server, RefConnHandle, Attr, Value, BufType,
			 ?DefaultTimeout).

sql_set_connect_attr(Server, RefConnHandle, Attr, Value, BufType, Timeout) ->
    Ret = sqlSetConnectAttr(Server,Attr,Value,Timeout),
    case Ret of
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	Other ->
	    Other
    end.

%% This function is deprecated and does nothing.
%% It is included only for compatibility reason.
%%---------------------------------------------------------------------------
sql_set_env_attr(Server, RefEnvHandle, Attr, Value, BufType) ->
    sql_set_env_attr(Server, RefEnvHandle, Attr, Value, BufType, 
		     ?DefaultTimeout).

sql_set_env_attr(Server, RefEnvHandle, Attr, Value, BufType, Timeout) ->
    0.

%% This function is deprecated and replaced by the function
%% columnRef/0. 
%%---------------------------------------------------------------------------
alloc_buffer(Server, BufCType, Size) ->
    alloc_buffer(Server, BufCType, Size, ?DefaultTimeout).

alloc_buffer(Server, BufCType, Size, Timeout) ->
    columnRef().

%% This function is deprecated and does nothing.
%% It is included only for compatibility reason.
%%---------------------------------------------------------------------------
dealloc_buffer(Server, RefBuf) ->
    dealloc_buffer(Server, RefBuf, ?DefaultTimeout).

dealloc_buffer(Server, RefBuf, Timeout) ->
    ok.

%% This function is deprecated and replaced by the function 
%% readData/[2,3].
%%---------------------------------------------------------------------------
read_buffer(Server, RefBuf) ->
    read_buffer(Server, RefBuf, ?DefaultTimeout).

read_buffer(Server, RefBuf, Timeout) ->
    Ret = readData(Server, RefBuf, Timeout),
    case Ret of
	{ok, Value} ->
	    {ok, {Value, -1}}; 
	Other ->
	    Other
    end.

%% This function is deprecated and does nothing.
%% It is included only for compatibility reason.
%%---------------------------------------------------------------------------
init_env(Server) ->
    init_env(Server, ?DefaultTimeout).

init_env(Server, Timeout) ->
    {ok, void}.


%% This function is deprecated and replaced by the function 
%% erl_connect/[2,3,4,5].
%%---------------------------------------------------------------------------
connect(Server, RefEnvHandle, ConnectStr) ->
    connect(Server, RefEnvHandle, ConnectStr, ?DefaultTimeout).

connect(Server, RefEnvHandle, ConnectStr, Timeout) ->
    Ret = erl_connect(Server, ConnectStr, Timeout),
    case Ret of
	ok ->
	    {ok, void};
	{error,ErrMsg, ErrCode} ->
	    {error, {connect, [{void, {ErrCode, ErrMsg, -1}}]}};
	Other ->
	    Other
    end.

connect(Server, RefEnvHandle, DSN, UID, PWD) ->
    connect(Server, RefEnvHandle, DSN, UID, PWD, ?DefaultTimeout).

connect(Server, RefEnvHandle, DSN, UID, PWD, Timeout) ->
    Ret = erl_connect(Server, DSN, UID, PWD, Timeout),
    case Ret of
	ok ->
	    {ok, void};
	{error,ErrMsg, ErrCode} ->
	    {error, {connect, [{void, {ErrCode, ErrMsg, -1}}]}};
	Other ->
	    Other
    end.

%% This function is deprecated and replaced by the function
%% erl_executeStmt/[2,3].
%%---------------------------------------------------------------------------
execute_stmt(Server, RefConnHandle, Stmt) ->
    execute_stmt(Server, RefConnHandle, Stmt, ?DefaultTimeout).

execute_stmt(Server, RefConnHandle, Stmt, Timeout) ->
    Ret = erl_executeStmt(Server, Stmt, Timeout),
    case Ret of
	{error,ErrMsg1, ErrMsg2} ->
	    {error,{execute_stmt, [{void, {-1, ErrMsg2, -1}}]}};
	{selected, ColumnList, RowData} ->
	    NewColumnList = ColumnList, % For new ColumnList 
	    NewRowData = bar(RowData),
	    {selected,NewColumnList, NewRowData};
	Other ->
	    Other
    end.
    
%% Bar convert a list in format [[El11, El12, ...], [El21, ...],....] 
%% to same lists but each element (El11,..) converted to a string. 
bar([])->
    [];
bar([[X|Xs]|Ys])->
    [lists:append(bar2(X),bar1(Xs))|bar(Ys)].
bar1([X])->
    bar2(X);
bar1([X|Xs]) ->
    lists:append(bar2(X),bar1(Xs)).
bar2([])->
    [" "];
bar2(X) when list(X) ->
    io_lib:format("~s",[X]);
bar2(X) when float(X) ->
    io_lib:format("~f",[X]);
bar2(X) ->
    io_lib:format("~p",[X]).

    
%% This function is deprecated and replaced by the function
%% erl_disconnect/[1,2].
%%---------------------------------------------------------------------------
disconnect(Server, RefConnHandle) ->
    disconnect(Server, RefConnHandle, ?DefaultTimeout).

disconnect(Server, RefConnHandle, Timeout) ->
    Ret = erl_disconnect(Server, Timeout),
    case Ret of
	ok ->
	    ok;
	{error, ErrMsg, ErrCode} ->
	    {error, {disconnect, [{void, {ErrCode, ErrMsg, -1}}]}};
	Other ->
	    Other
    end.


%% This function is deprecated and does nothing.
%% It is included only for compatibility reason.
%%-------------------------------------------------------------------
terminate_env(Server, RefEnvHandle) ->
    terminate_env(Server, RefEnvHandle, ?DefaultTimeout).

terminate_env(Server, RefEnvHandle, Timeout) ->
    ok.

%% This function is deprecated and does nothing.
%% It is included only for compatibility reason.
%%-------------------------------------------------------------------
display_size(SqlType, ColumnSize) when 
  integer(SqlType), integer(ColumnSize) ->
    -1.

%%%----------------------------------------------------------------------
%%% Debug fcns
%%% Return the internal state in a tagged tuple list.
%%%----------------------------------------------------------------------

get_state(Server) ->
    State = gen_server:call(Server, get_state),
    [{port, State#state.port},
     {connected, State#state.connected},
     {connection_string, State#state.connection_string},
     {from, State#state.from},
     {fetchdata, State#state.fetchdata},
     {columnvector, State#state.columnvector}].


%%%======================================================================
%%% Callback functions from gen_server
%%%======================================================================

%%----------------------------------------------------------------------
%% init/1
%%----------------------------------------------------------------------

init(Args) ->
    process_flag(trap_exit, true),
    % Default: Debug flag off
    put(dbge, false),
    {value,{client,ClientPid}} = lists:keysearch(client, 1,Args), 

    % Start the C program
    case os:find_executable(?ServerProg, ?ServerDir) of
	FileName when list(FileName)->
	    Port  = open_port({spawn,FileName},[{packet,4}, binary]),
	    State = #state{port=Port, clientpid = ClientPid},
	    {ok, State};
	false ->
	    {stop, {no_odbcserver, "Can't find executable."}}
    end.
		    

%%----------------------------------------------------------------------
%% handle_call/3
%%----------------------------------------------------------------------

%% Stop Call Back
handle_call(stop, _From, State) ->
    i("handle_call(stop)~n",[]),
    {stop, normal, ok, State};

%% ODBC Call Backs
handle_call(T, _, #state{from=From}=State) when tuple(T),
						From =/= undefined ->
    i("handle_call(From =/= undefined) -> From:~p T: ~p", [From,T]),
    {reply,busy,State};

handle_call({openconnection, Cmd, ConnectString}, From, State) ->
    i("handle_call(openconnection) -> entry", []),
    CmdString = lists:flatten([Cmd, ConnectString]),
    Port  = State#state.port,
    Reply = send(Port, CmdString),
    NewState = State#state{connected = true, 
			   connection_string = ConnectString,
			   from = From},
    {noreply, NewState};

handle_call({execute, SqlStmt}, From, #state{connected = false} = State) ->
    {reply,{error,not_connected},State};
handle_call({execute, SqlStmt}, From, State) ->
    Port = State#state.port,
    Port ! {self(), {command, term_to_binary(SqlStmt)}},
    {noreply,State#state{from=From}};

handle_call({execdir, CmdStr}, From,  #state{connected = false} = State) ->
    {reply,{error,not_connected},State};
handle_call({execdir, CmdStr}, From, State) ->
    Port = State#state.port,
    send(Port, CmdStr),
    {noreply,State#state{from=From}};

handle_call({closeconnection,CloseStr}, From, State) ->
    Port = State#state.port,
    NewState = State#state{connected = false, connection_string = []},
    Reply = send(Port, CloseStr),
    {noreply,NewState#state{from=From}};

handle_call({bindcolumn, CmdStr, ColNum, MemRef}, From,
	   #state{connected = false} = State) ->
    {reply,{error,not_connected},State};
handle_call({bindcolumn, CmdStr, ColNum, MemRef}, From, State) ->
    Port = State#state.port,
    ColumnVec = State#state.columnvector,
    NewColumnVec = insert({MemRef,ColNum}, ColumnVec),
    NewState = State#state{columnvector = NewColumnVec},
    NCmdStr = lists:flatten([CmdStr, integer_to_list(ColNum)]),
    send(Port, NCmdStr),
    {noreply,NewState#state{from=From}};

handle_call({closecursor, CmdStr}, From, #state{connected = false} = State) ->
    {reply, {error,not_connected}, State};
handle_call({closecursor, CmdStr}, From, State) ->
    i("handle_call(closecursor) -> entry with"
      "~n  CmdStr: ~s"
      "~n  From:   ~p", [CmdStr, From]),
    Port = State#state.port,
    ColumnVect    = State#state.columnvector,
    NewColumnVect = delete(ColumnVect),
    i("handle_call(closecursor) -> "
      "~n   NewColumnVect: ~p", [NewColumnVect]),
    NewState = State#state{columnvector = NewColumnVect, fetchdata = false},
    send(Port, CmdStr),
    {noreply, NewState#state{from = From}};

handle_call({describeColumn, CmdStr}, From, #state{connected = false} = 
	    State) ->
    {reply,{error,not_connected},State};
handle_call({describeColumn, CmdStr}, From, State) ->
    Port = State#state.port,
    send(Port, CmdStr),
    {noreply,State#state{from=From}};

handle_call({endTran, CmdStr}, From, #state{connected = false} = State) ->
    {reply,{error,not_connected},State};
handle_call({endTran, CmdStr}, From, State) ->
    Port = State#state.port,
    send(Port, CmdStr),
    {noreply,State#state{from=From}};

handle_call({fetch, CmdStr}, From, #state{connected = false} = State) ->
    i("handle_call(fetch) -> entry when not connected", []),
    {reply, {error,not_connected}, State};
handle_call({fetch, CmdStr}, From, State) ->
    i("handle_call(fetch) -> entry with"
      "~n  CmdStr: ~s"
      "~n  From:   ~p", [CmdStr, From]),
    Port = State#state.port,
    send(Port, CmdStr),
    {noreply, State#state{from = From, fetchdata = true}};

handle_call({numResultCols, CmdStr}, From, #state{connected = false} = 
	    State) ->
    {reply,{error,not_connected},State};
handle_call({numResultCols, CmdStr}, From, State) ->
    Port = State#state.port,
    send(Port, CmdStr),
    {noreply,State#state{from=From}};

handle_call({rowCount, CmdStr}, From, #state{connected = false} = State) ->
    {reply,{error,not_connected},State};
handle_call({rowCount, CmdStr}, From, State) ->
    Port = State#state.port,
    send(Port, CmdStr),
    {noreply,State#state{from=From}};

handle_call({setConnectAttr, CmdStr}, From, State) ->
    Port = State#state.port,
    send(Port, CmdStr),
    {noreply,State#state{from=From}};

handle_call({readbuffer, _, _}, _, #state{fetchdata = false} = State) ->
    i("handle_call(readbuffer) -> entry when fetchdata = falsewith", []),
    {reply, {error, no_data_fetched}, State};

handle_call({readbuffer, CmdStr, ColumnRef}, From, State) ->
    i("handle_call(readbuffer) -> entry with"
      "~n  CmdStr:    ~s"
      "~n  ColumnRef: ~p", [CmdStr, ColumnRef]),
    Port       = State#state.port,
    ColumnVect = State#state.columnvector,
    CNumber    = lockup(ColumnRef,ColumnVect),
    i("handle_call(readbuffer) -> CNumber: ~p", [CNumber]),
    case CNumber of
	not_found ->
	    {reply, {error, column_not_found}, State};
	_ ->
	    NCmdStr = lists:flatten([CmdStr, integer_to_list(CNumber)]),
	    send(Port, NCmdStr),
	    {noreply, State#state{from = From}}
    end;

%% Debug Call Back
handle_call({debuge, CmdStr}, From, State) ->
    case CmdStr of
	on ->
	    put(dbge, true);
	off ->
	    put(dbge, false)
    end,
    {reply,debugeok, State};
handle_call({debugc, CmdStr}, From, State) ->
    Port = State#state.port,
    send(Port, CmdStr),
    {reply,debugcok, State};

handle_call(get_state, _From, State) ->
    {reply, State, State};

% Unknown request
handle_call(Request, _From, State) ->
    error_msg("received unexpected request: ~p", [Request]),
    Reply = ok,
    {reply, Reply, State}.

%%----------------------------------------------------------------------
%% handle_cast/2
%%----------------------------------------------------------------------
handle_cast(Msg, State) ->
    error_msg("received unexpected message: ~p", [Msg]),
    {noreply, State}.

%%----------------------------------------------------------------------
%% handle_info/2
%%----------------------------------------------------------------------
handle_info({'EXIT', Pid, Reason}, State) when pid(Pid) ->
    error_msg("Linked process with Pid ~p got exit signal with reason ~p", 
			       [Pid, Reason]),
    {stop, {stopped, {'EXIT', Pid, Reason}}, State};

% EXIT from port, log exit and terminate.
handle_info({'EXIT', Port, Reason}, #state{port = Port} = State) ->
    error_msg("exit signal from port program:", [Reason]),
    Reason1 = {port_exit, Reason},
    {stop, Reason1, State};

handle_info({'EXIT', From, Reason}, State) ->
    error_msg("received unexpected exit signal from ~p: ~p", [From, Reason]),
    {noreply, State};

handle_info({Port, {data, BinData}}, #state{from = From} = State) 
  when From =/= undefined ->
    i("handle_info(data) -> entry with"
      "~n   Port:          ~p"
      "~n   size(BinData): ~p", [Port, size(BinData)]),
    case (catch binary_to_term(BinData)) of
	{'EXIT',Reason} ->
	    i("handle_info(Port) decode error: ~p~n", [Reason]),
	    gen_server:reply(From, {error,{binary_to_term, Reason}});
	Data ->
	    i("handle_info(Port) Data: ~p~n",[Data]),
	    gen_server:reply(From, Data)
    end,
    {noreply,State#state{from=undefined}};

handle_info({Port, {data, BinData}}, State) ->
    error_msg("unexpected data from port: ~p", [binary_to_term(BinData)]),
    {noreply,State};

% Other
handle_info(Info, State) ->
    error_msg("received unexpected info: ~p", [Info]),
    {noreply, State}.

%%----------------------------------------------------------------------
%% terminate/2 and code_change/3
%%----------------------------------------------------------------------

terminate(Reason, State) ->
    Port = State#state.port,
    Port ! {self(), close}.

code_change(Vsn, State, Extra) ->
    {ok, State}.

%%%======================================================================
%%% Internal functions
%%%======================================================================
%% Send a command to ODBC
send(Port, Data)->
    Port ! {self(), {command, term_to_binary(Data)}},
    ok.

%% This is a simple table. The table contains tuples.
%% Each tuple contain a memoryreferens and a corrsponding
%% columnnumber.
insert({Ref, ColumnNr}, []) ->
    [{Ref, ColumnNr}];
insert({Ref, ColumnNr}, Xs) ->
    [{Ref, ColumnNr}| Xs].

lockup(Ref, []) ->
    not_found;
lockup(Ref, [{Ref, ColumnNr} | Xs]) ->
    ColumnNr;
lockup(Ref, [{ORef, ColumnNr} | Xs] ) ->
    lockup(Ref, Xs).

delete(Xs) ->
    [].

error_msg(F, A) ->
    error_logger:error_msg("ODBC: " ++ F, A).


