
%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
%%%----------------------------------------------------------------------
%%% Purpose   : Spawn a worker for each connection initiated from client
%%%----------------------------------------------------------------------

-module(rshell_connection).

-export([start/0,
	 start/1,
         init/1]).

%         system_continue/3,
%         system_terminate/4,
%         system_code_change/4,

-include("rshell.hrl").


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% State for the server loop. It also keep the state for the encoding
%% and decoding but the code in this module knwo nothing about the
%% content of them.

-record(connection, {socket, user, cb_module, cb_data}).


%%%----------------------------------------------------------------------
%%% This module handle one connection using one protocol. No protocol
%%% specific handling is done in this module, instead this is in a
%%% separate call-back module. We define a protocol between this code
%%% and that module. We call the functions
%%%
%%%   CallBackModule:init(Socket) -> {ok,Actions,CallBackData}
%%%
%%%   CallBackModule:out(Bytes,OldCallBackData) -> {ok,Actions,CallBackData}
%%%
%%%   CallBackModule:in(Bytes,OldCallBackData) -> {ok,Actions,CallBackData}
%%%
%%% The init() function is called the first time to let the call-back
%%% module create the CallBackData term.
%%%
%%% The out() function is called when Erlang wants to send data to the
%%% client.
%%%
%%% The in() function is called when Erlang got data from the client.
%%% 
%%% 'Actions' is a list of things to do by the caller in this module.
%%% The actions are
%%%
%%%   {terminate,ReasonMessageString}  Send this string and terminate
%%%   {out,Bytes}                      Send these bytes to client
%%%   {in, Bytes}                      Send these bytes to user 
%%%
%%%----------------------------------------------------------------------

start() ->
    start([]).

start(Options) ->				% XXX: When any Options?
    Pid = proc_lib:spawn_link(?MODULE, init, [Options]),
    {ok, Pid}.

init(Socket) ->
    CallBackModule = rshell_telnet,
    process_flag(trap_exit, true),


    io:format(user,"Starting connection: ~p\n",[Socket]),
    User = user_drv:start(self(),{shell,start,[]}),
    io:format(user,"Spawned user_drv...\n",[]),

    {ok,Actions,CallBackData} = apply(CallBackModule,init,[Socket]),

    State = #connection{socket = Socket,
			user = User,
			cb_module = CallBackModule,
			cb_data   = CallBackData
		       },


    NewState = actions(Actions, State),
    io:format(user,"CB initiated ~p\n",[CallBackData]),

    dbg:p(self(),[m]),
%    dbg:p(User,all),
    loop(NewState).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Read the data and process it.

loop(#connection{socket = Sock, user = User} = State) ->
    receive
	{tcp, Sock, Bytes} ->
%	    io:format(user,"got ~p...\n",[Bytes]),
	    loop(in(Bytes,State));
	{tcp_closed, Sock} ->
	    io:format(user,"rshell_connection:loop(~p, ~p) socket closed~n",
		      [Sock, User]),
	    gen_tcp:close(Sock),		% Close to release resources
	    User ! {self(),eof},
	    exit(closed);
	{User, {command, [Code | Bytes]}} ->
	    Bytes1 = rshell_terminal:send_cmd(Code, Bytes),
	    loop(out(Bytes1,State));
	{'EXIT', User, Why} ->
	    io:format(user,"rshell_connection:loop(~p, ~p, ~p) user died ~p~n",
		      [Sock,User,State,Why]),
	    gen_tcp:close(Sock),		% Close and give up
	    exit(Why);
	Other ->
	    io:format(user,"rshell_connection:loop(~p, ~p, ~p) got msg ~p~n",
		      [Sock,User,State,Other]),
	    loop(State)
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Low level functions

in(Bytes, #connection{cb_module = CBModule, cb_data = CBData} = State) ->
    {ok,Actions,NewCBData} = apply(CBModule,in,[Bytes,CBData]),
    actions(Actions, State#connection{cb_data = NewCBData}).

out(Bytes, #connection{cb_module = CBModule, cb_data = CBData} = State) ->
    {ok,Actions,NewCBData} = apply(CBModule,out,[Bytes,CBData]),
    actions(Actions, State#connection{cb_data = NewCBData}).

actions([], State) ->
    State;
actions([Action | Actions], State) ->
    actions(Actions, actions(Action, State));
actions({in,Bytes}, State) ->
    State#connection.user ! {self(),{data,Bytes}},
    State;
actions({out,Bytes}, State) ->
    gen_tcp:send(State#connection.socket, Bytes),
    State.

%% FIXME terminate
