%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
%%%----------------------------------------------------------------------
%%% Purpose : Decoding of data recieved
%%%----------------------------------------------------------------------

-module(rshell_telnet).

-include("rshell.hrl").
-include("rshell_telnet.hrl").

-export([init/1,in/2,out/2]).

-import(lists,[flatten/1,reverse/1]).

-record(telnet,{in_state = normal, wait_for}).

%%%----------------------------------------------------------------------
%%% NOTE: This module should not send or receive messages or do any io
%%% operations, it is to be strictly functional.
%%%----------------------------------------------------------------------

%%%----------------------------------------------------------------------
%%% We send some options at startup that the client have to say
%%% DO to:
%%%
%%%   WILL SUPRESS GO AHEAD
%%%   WILL ECHO
%%%----------------------------------------------------------------------

init(_Socket) ->
    Actions = [
	       {out,[?IAC,?WILL,?ECHO]},
	       {out,[?IAC,?WILL,?SUPPRESS_GO_AHEAD]}
	      ],
    WaitFor = [
	       {do,?ECHO},
	       {do,?SUPPRESS_GO_AHEAD}
	      ],
    {ok,Actions,#telnet{wait_for = WaitFor}}.


%%%----------------------------------------------------------------------
%%% The telnet options are embedded into the character stream.
%%% We implement a small state machine to find them if they are
%%% received in separete data packets.
%%%
%%%   normal
%%%   iac            We got one iac
%%%   {iac,do}       We got one iac do
%%%   {iac,will}     We got one iac will
%%%   {iac,dont}     We got one iac dont
%%%   {iac,wont}     We got one iac wont
%%%----------------------------------------------------------------------

in(Bytes, State) ->
    {Actions, NewState} = in(Bytes, State, []),
    {ok,Actions,NewState}.


in([?IAC | Bytes], #telnet{in_state = normal} = State, Actions) ->
    in(Bytes, State#telnet{in_state = iac}, Actions);

in([?IAC | Bytes], #telnet{in_state = iac} = State, Actions) ->
    in(Bytes, State#telnet{in_state = normal}, [{in,[?IAC]} | Actions]);

in([?DO | Bytes], #telnet{in_state = iac} = State, Actions) ->
    in(Bytes, State#telnet{in_state = {iac,do}}, Actions);

in([?WILL | Bytes], #telnet{in_state = iac} = State, Actions) ->
    in(Bytes, State#telnet{in_state = {iac,will}}, Actions);

in([?DONT | Bytes], #telnet{in_state = iac} = State, Actions) ->
    in(Bytes, State#telnet{in_state = {iac,dont}}, Actions);

in([?WONT | Bytes], #telnet{in_state = iac} = State, Actions) ->
    in(Bytes, State#telnet{in_state = {iac,wont}}, Actions);

in([Option | Bytes], #telnet{in_state = {iac,Cmd}} = State, Actions) ->
    {NewAction,NewState} = handle_option(Cmd, Option, State),
    in(Bytes, NewState, [NewAction | Actions]);

in([Byte | Bytes], #telnet{in_state = normal} = State, Actions) ->
    in(Bytes, State, [{in,[Byte]} | Actions]);

in([], State, Actions) ->
    {reverse(Actions), State}.

handle_option(Cmd, Option, State) ->
    Action = 
	case lists:keysearch(Option,2,State#telnet.wait_for) of
	    {value,{Cmd,_Option}} ->
		[];				% OK!!!
	    {value,{_Cmd,_Option}} ->
		{terminate,"Client refuses to set character mode"};
	    false when Cmd == do ->
		{out,[?IAC,xnot(Cmd)]}
	end,
    {Action,State#telnet{in_state = normal}}.

xnot(do) ->
    ?WONT;
xnot(will) ->
    ?DONT;
xnot(dont) ->
    ?WONT;
xnot(wont) ->
    ?DONT.

%%%----------------------------------------------------------------------
%%% The telnet options are embedded into the character stream.
%%% We implement a small state machine to find them if they are
%%% received in separete data packets.
%%%----------------------------------------------------------------------


out(Bytes, State) ->
    {ok,{out,encode(Bytes)},State}.

%%%----------------------------------------------------------------------
%%% There are some telnet "standard functions" sent from the user that
%%% we want to handle in our way:
%%%
%%%   IP  Interrupt process. We want to put up our standard dialog.
%%%       We have to try out if this special handling is needed.
%%%
%%%   AO  Abort output (not required). Don't send the rest of the
%%%       output until we get a prompt or until a timeout maybe.
%%%
%%%   AYT Are you there (not required)? What to respond?
%%%
%%%   EC  Erase character (not required). Is the DEL or BS enough
%%%       or should we expect to get this one?
%%%
%%%   EL  Erase line (not required). Should we expect this?
%%%
%%% Read the section about urgent on page 9 of RFC854 again.....
%%%----------------------------------------------------------------------


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Find the telnet commands in the input stream and act on them. Send
% characters not part of a command to the erlang XXXX (shell/user???).

%worker_in([], State) ->
%    State;
%worker_in(Bytes, #worker{in_state = InState} = State) ->
%    {Decoded,NewInState,MoreBytes} =
%	case rshell_telnet:decode(Bytes,InState) of
%	    {done,StateAndChars} ->
%		StateAndChars;
%	    {Cmd,StateAndChars} ->
%		io:format(user,"Cmd: ~p~n",[Cmd]),
%		StateAndChars
%	end,
%    State#worker.user ! {self(), {data, Decoded}},
%    worker_in(MoreBytes, State#worker{in_state = NewInState}).




%%%----------------------------------------------------------------------
%%%
%%% The main purpose of this decoding module is to decode the telnet
%%% commands that are embedded into the character stream. It will
%%% also translate the sequence \r\000 to a single \r to make 'edlin'
%%% happy.
%%% 
%%% Telnet commands are two or three bytes long. The three byte
%%% sequences take an command id as the third byte
%%% 
%%%     IAC WILL CMD
%%%     IAC WONT CMD
%%%     IAC DO   CMD
%%%     IAC DONT CMD
%%%     IAC SB   CMD
%%% 
%%% The function decode():
%%% 
%%%     decode(Chars, State) -> {done, {DecodedChars, NewState, []}}
%%%               | {{sub,Cmd,CmdArg}, {DecodedChars, NewState, RemainingChars}}
%%%               | {{opt,Type,Cmd},   {DecodedChars, NewState, RemainingChars}}
%%%               | {{cmd,Cmd},        {DecodedChars, NewState, RemainingChars}}
%%% 
%%% We form a state machine to handle the case that we have part of the
%%% command.
%%% 
%%%     char			
%%%     iac			IAC
%%%     sb			IAC SB
%%%     {sb,Cmd,Chars}		IAC SB Cmd ....
%%%     {sb_iac,Chars}		IAC SB ..... IAC
%%%     {opt,Type}		IAC ( WILL | WONT | DO | DONT )
%%% 
%%% Note that the decode function has to be called repeatably until we have
%%% decoded all characters taking action upon the telnet commands we find
%%% as the appear.
%%% 
%%%----------------------------------------------------------------------

decode(Chars) ->
    decode(Chars, char).


decode(Chars, State) ->
    decode(State, flatten(Chars), []).


decode(State, [], Dec) ->			% We are done
    {done,{reverse(Dec),State,[]}};


decode(char, [?IAC | Chars], Dec) ->		% Begin reading a command
    decode(iac, Chars, Dec);

decode(char, [$\r, 0 | Chars], Dec) ->		% Strip \000 after \r
    decode(char, Chars, [$\r | Dec]);

decode(char,[Char | Chars], Dec) ->		% Plain characters
    decode(char,Chars, [Char | Dec]);


decode(iac, [?WILL | Chars], Dec) ->		% Read option request/reply
    decode({opt,will}, Chars, Dec);

decode(iac, [?WONT | Chars], Dec) ->
    decode({opt,wont}, Chars, Dec);

decode(iac, [?DO | Chars], Dec) ->
    decode({opt,do}, Chars, Dec);

decode(iac, [?DONT | Chars], Dec) ->
    decode({opt,dont}, Chars, Dec);

decode(iac, [?SB | Chars], Dec) ->		% Begin reading a sub command
    decode(sb, Chars, Dec);

decode(iac, [?IAC | Chars], Dec) ->		% Escape of 0xff
    decode(char, Chars, [?IAC | Dec]);

decode(iac, [Cmd | Chars], Dec) ->		% Some other command
    {{cmd,Cmd},{reverse(Dec),char,Chars}};


decode({opt,Type}, [Cmd | Chars], Dec) ->	% Return the option req/reply
    {{opt,Type,Cmd},{reverse(Dec),char,Chars}};


decode(sb, [Char | Chars], Dec) ->		% The sub cmd id
    decode({sb,Char,[]}, Chars, Dec);		% FIXME: Handle IAC IAC ?


decode({sb,Cmd,SubChars}, [?IAC | Chars], Dec) -> % Maybe the end of the sub cmd
    decode({sb_iac,Cmd,SubChars}, Chars, Dec);

decode({sb,Cmd,SubChars}, [Char | Chars], Dec) ->	% Char in sub cmd
    decode({sb,Cmd,[Char|SubChars]}, Chars, Dec);

decode({sb_iac,Cmd,SubChars}, [?IAC | Chars], Dec) -> % Not the end of the
    decode({sb_iac,Cmd,[?IAC|SubChars]}, Chars, Dec); % sub cmd

decode({sb_iac,Cmd,_SubChars}, [?SE | Chars], Dec) -> % The end of the sub cmd
    {{sub,Cmd,Chars},{reverse(Dec),char,Chars}};

decode({sub_iac,_SubChars}, [Char | Chars], Dec) -> % Really an error, skip
    decode(char, [Char | Chars], Dec).



%%%----------------------------------------------------------------------
%%%
%%% NOTE: This module is supposed to be free from side effects,
%%% there should be no send or receive statements.
%%%
%%% We get data as messages from XXXX and do some encoding before
%%% sending it to the telnet client.
%%%
%%%   We have to be in binary mode to send characters above 127.
%%%   (Currently not handled in this module)
%%%
%%%   A ascii 255 has to be doubled so that it is not misstaken
%%%   for a IAC.
%%%
%%%   A line is ended with \r\n so in general we have to precede
%%%   all \n with a \r.
%%%
%%%   A single \r is sent as \r\000.
%%%
%%%   The standard want us to wait until we got a complete line
%%%   before we send it. Some event, not defined in RFC 857, can
%%%   overrule this and make us send the data. We operate in
%%%   "character mode" and send things as soon as we get them.
%%%
%%%   To fully comply with the standard we have a problem if the
%%%   last character to send is a \r. We don't know if it is worth
%%%   to wait a few micro seconds to se if the next character is
%%%   \n in case we have a new line, \r\n. If \r was followed by 
%%%   something else it should be sent as \r\000.
%%%
%%%   We ignore this problem and assumes that we get either
%%%   \r\n at once or we got a single \r not to be followed by
%%%   a \n later, i.e. we send a \r\000.
%%%
%%%----------------------------------------------------------------------

encode(Chars) ->
    reverse(encode(flatten(Chars), [])).
    

encode([?IAC | Cs], Enc) ->
    encode(Cs, [?IAC, ?IAC | Enc]);

encode([$\r, $\n | Cs], Enc) ->			% MSDOS EOL, just pass
    encode(Cs, [$\n, $\r | Enc]);

encode([$\n | Cs], Enc) ->			% Unix, add a CR before LF
    encode(Cs, [$\n, $\r | Enc]);

encode([$\r], Enc) ->				% We are forced to decide,
    [0, $\r | Enc];				% guess it is not end of line

encode([C | Cs], Enc) ->
    encode(Cs, [C | Enc]);

encode([], Enc) ->
    Enc.

