%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id $
%%
-module(dets_v9).

%% Dets files, implementation part. This module handles version 9.
%% To be called from dets.erl only.

-export([constants/0, mark_dirty/1, read_file_header/2,
         check_file_header/2, do_perform_save/1, initiate_file/11,
         prep_table_copy/9, init_freelist/2, fsck_input/3,
         bulk_input/3, output_objs/4, fast_init/2, write_cache/1,
         may_grow/3, find_object/2, slot_objs/2, scan_objs/7,
         db_hash/2, no_slots/1, table_parameters/1]).

-export([file_info/1, v_segments/1]).

-export([cache_segps/3]).

-compile({inline, [{write_segment_file,6}]}). 
-compile({inline, [{sz2pos,1},{adjsz,1},{to_objects,3},{scan_skip,8}]}).
-compile({inline, [{skip_bytes,6},{make_object,4}]}).
-compile({inline, [{segp_cache,2},{get_segp,1},{get_arrpart,1}]}).
-compile({inline, [{h,1},{actual_seg_size,0}]}).

-include("dets.hrl").

%%  The layout of the file is :
%%
%%   bytes   decsription
%%  ---------------------- File header
%%    4      FreelistsPointer
%%    4      Cookie
%%    4      ClosedProperly (pos=8)
%%    4      Type (pos=12)
%%    4      Version (pos=16)
%%    4      M
%%    4      Next
%%    4      KeyPos
%%    4      NoObjects
%%    4      NoKeys
%%    4      MinNoSlots
%%    4      MaxNoSlots
%%    4      MaxObjSize
%%    4      N
%%    256    Reserved for future versions. Initially zeros.
%%  ------------------ end of file header
%%    4*256  SegmentArray Pointers.
%%  ------------------ This is BASE.
%%    4*512  SegmentArray Part 1
%%    ...    More SegmentArray Parts
%%    8*256  First segment
%%    ???    Objects (free and alive)
%%    4*512  Further SegmentArray Part.
%%    ???    Objects (free and alive)
%%    8*256  Further segment.
%%    ???    Objects (free and alive)
%%    ... more objects, segment array parts, and segments ...
%%  -----------------------------
%%    ???    Free lists
%%  -----------------------------
%%    4      File size, in bytes. 

%%  Before we can find an object we must find the slot where the
%%  object resides. Each slot is a (possibly empty) list (or chain) of
%%  objects that hash to the same slot. If the value stored in the
%%  slot is zero, the slot chain is empty. If the slot value is
%%  non-zero, the value points to a position in the file where the
%%  collection of objects resides. Each collection has the following
%%  layout:
%%
%%   bytes  decsription
%%  --------------------
%%    4     Size of the area allocated for the collection (8+Sz)
%%    4     Status  (FREE or ACTIVE). These two are the Object Header.
%%    Sz    A binary containing the objects per key, sorted on key
%% 
%%  When repairing or converting a file, the status field is used.
%%
%%  The binary containing the objects per key of a table of type 'set'
%%  has the following layout:
%%
%%   bytes  decsription
%%  --------------------
%%    4     Size of the object of the first key (4+OSz1)
%%    OSz1  The object of the first key
%%    ...
%%    4     Size of the object of the ith key (4+OSzi)
%%    OSzi  The object of the ith key
%% 
%%  The binary containing the objects per key of a table of type 'bag'
%%  or 'duplicate_bag' has the following layout:
%%
%%   bytes    decsription
%%  ----------------------
%%    4       Size of the objects of the first key (4 + OSz1_1+...+OSz1_j+...)
%%    4       Size of the first object of the first key (4+OSz1_1)
%%    OSz1_1  The first object of the first key
%%    ...
%%    4       Size of the jth object of the first key (4+OSz1_j)
%%    OSz1_j  The jth object of the first key
%%    ...
%%    4       Size of the objects of the ith key (4 + OSzi_1+...+OSzi_k+...)
%%    4       Size of the first object of the ith key (4+OSzi_1)
%%    OSzi_1  The first object of the ith key
%%    ...
%%    4       Size of the kth object of the ith key (4+OSzi_k)
%%    OSzi_k  The kth object of the ith key
%%    ...
%%
%%  The objects of a key are placed in time order, that is, the older
%%  objects come first. If a new object is inserted, it is inserted
%%  last.
%%
%%
%%
%%|---------------|
%%|      head     |
%%|       	  |
%%|               |
%%|_______________|
%%|               |--|
%%|___part ptr 1__|  |
%%|               |  | segarr part 1
%%|___part ptr 2__|  V______________|
%%|               |  |   p1         |
%%|               |  |______________|--|
%%|     ....      |  |   p2         |  |
%%     (256)         |______________|  |
%%                   |              |  |
%%                   |     ....     |  | segment 1
%%                   |    (512)     |  V __slot 0 ____|
%%                                     |   size       |
%%                                     |   pointer    |--|
%%                                     |___slot 1 ____|  |
%%                                     |              |  |
%%                                     |   ....       |  |  objects in slot 0
%%                                         (256)         V  segment 1
%%                                                       |___________|
%%                                                       |  size     |
%%                                                       |___________|
%%                                                       |  status   |
%%                                                       |___________|
%%                                                       |           |
%%                                                       |   object  |
%%                                                       |   collec. |
%%                                                       |___________|

%%%
%%% File header
%%%

-define(RESERVED, 256).        % Reserved for future use.

-define(HEADSZ, 56).           % The size of the file header, in bytes.
-define(HEADEND, (?HEADSZ+?RESERVED)). % End of header and reserved area.
-define(SEGSZ, 512).           % Size of a segment, in words. SZOBJP*SEGSZP.
-define(SEGSZP, 256).          % Size of a segment, in number of pointers.
-define(SEGPARTSZ, 512).       % Size of segment array part, in words.
-define(SEGARRSZ, 256).        % Maximal number of segment array parts..
-define(SEGARRADDR(PartN), (?HEADEND + (4 * (PartN)))).
-define(SEGPARTADDR(P,SegN), ((P) + (4 * ((SegN) rem ?SEGPARTSZ)))).
-define(BASE, ?SEGARRADDR(?SEGARRSZ)).
-define(MAXSLOTS, (?SEGARRSZ * ?SEGPARTSZ * ?SEGSZP)).

%% BIG is used for hashing. BIG must be greater than the maximum
%% number of slots, currently 32 M (MAXSLOTS).
-define(BIG, 16#3ffffff). % 64 M

%% Hard coded positions into the file header:
-define(FREELIST_POS, 0).
-define(CLOSED_PROPERLY_POS, 8).
-define(D_POS, 20).
-define(NO_KEYS_POS, (?D_POS + 16)).

%%% Dets file versions up to 8 are handled in dets_v8. This module
%%% handles version 9, introduced in R8.

-define(FILE_FORMAT_VERSION, 9).

-define(NOT_PROPERLY_CLOSED,0).
-define(CLOSED_PROPERLY,1).

%% Size of object pointer, in words. SEGSZ = SZOBJP * SEGSZP.
-define(SZOBJP, 2).

-define(OHDSZ, 8).          % The size of the object header, in bytes.
-define(STATUS_POS, 4).     % Position of the status field.

-define(OHDSZ_v8, 12).      % The size of the version 8 object header.

%% The size of each object is a multiple of 16.
%% BUMP is used when repairing files.
-define(BUMP, 16).

-define(HASH_TAG, '$hash').

%%-define(DEBUGF(X,Y), io:format(X, Y)).
-define(DEBUGF(X,Y), void).

%% {Bump}
constants() ->
    {?BUMP, ?BASE}.

%% -> ok | throw({NewHead,Error})
mark_dirty(Head) ->
    Dirty = [{?CLOSED_PROPERLY_POS, <<?NOT_PROPERLY_CLOSED:32>>}],
    dets_utils:pwrite(Head, Dirty),
    dets_utils:sync(Head),
    dets_utils:position(Head, Head#head.freelists_p),
    dets_utils:truncate(Head, cur).

%% -> {ok, head()} | throw(Error) | throw(badarg)
prep_table_copy(Fd, Tab, Fname, Type, Kp, Ram, CacheSz, Auto, Parms) ->
    case Parms of
	{?HASH_TAG,?FILE_FORMAT_VERSION,N,M,Next,Min,Max}
	        when integer(N), integer(M), integer(Next), 
		     integer(Min), integer(Max) ->
	    case hash_invars(N, M, Next, Min, Max) of
		false ->
		    throw(badarg);
		true ->
		    init_file(Fd, Tab, Fname, Type, Kp, Min, Max, 
			      Ram, CacheSz, Auto, false, M, N, Next)
	    end;
	_ ->
	    throw(badarg)
    end.

%% -> {ok, head()} | throw(Error)
%% The File header and the SegmentArray Pointers are written here.
%% SegmentArray Parts are also written, but the segments are are not
%% initialized on file unless DoInitSegments is 'true'. (When
%% initializing a file by calling init_table, some time is saved by
%% not writing the segments twice.)
initiate_file(Fd, Tab, Fname, Type, Kp, MinSlots0, MaxSlots0, 
	      Ram, CacheSz, Auto, DoInitSegments) ->
    MaxSlots1 = lists:min([MaxSlots0, ?MAXSLOTS]),
    MinSlots1 = lists:min([MinSlots0, MaxSlots1]),
    MinSlots = slots2(MinSlots1),
    MaxSlots = slots2(MaxSlots1),
    M = Next = MinSlots,
    N = 0,
    init_file(Fd, Tab, Fname, Type, Kp, MinSlots, MaxSlots, Ram, CacheSz,
	      Auto, DoInitSegments, M, N, Next).

init_file(Fd, Tab, Fname, Type, Kp, MinSlots, MaxSlots, Ram, CacheSz,
	  Auto, DoInitSegments, M, N, Next) ->
    Ftab = dets_utils:init_alloc(?BASE),

    Head0 = #head{
      m  = M,
      next = Next,
      fptr = Fd,
      no_objects = 0,
      no_keys = 0,
      maxobjsize = 0,
      n = N,
      type = Type,
      update_mode = dirty,
      freelists = Ftab,
      auto_save = Auto,
      hash_bif = phash,
      keypos = Kp,
      min_no_slots = MinSlots,
      max_no_slots = MaxSlots,
      
      ram_file = Ram, 
      filename = Fname, 
      name = Tab,
      cache = dets_utils:new_cache(CacheSz),
      version = ?FILE_FORMAT_VERSION,
      bump = ?BUMP,
      base = ?BASE,
      mod = ?MODULE
     },

    FreeListsPointer = 0,
    Header = file_header(Head0, FreeListsPointer, ?NOT_PROPERLY_CLOSED),

    W0 = {0, [Header, 
	      <<0:?RESERVED/unit:8,          %% Reserved area.
	        0:(4*?SEGARRSZ)/unit:8>>]},  %% SegmentArray Pointers

    %% Remove cached pointers to segment array parts and segments:
    lists:foreach(fun({I1,I2}) when integer(I1), integer(I2) -> ok;
		     ({K,V}) -> put(K, V)
		  end, erase()),

    %% Initialize array parts. 
    %% All parts before segments, for the sake of repair and initialization.
    Zero = seg_zero(),
    {Head1, Ws1} = init_parts(Head0, 0, no_parts(MinSlots), Zero, []),
    NoSegs = no_segs(MinSlots),

    {Head, WsI, WsP} = init_segments(Head1, 0, NoSegs, Zero, [], []),
    Ws2 = if
	      DoInitSegments == true -> WsP ++ WsI;
	      true -> WsP
	 end,
    dets_utils:pwrite(Fd, Fname, [W0 | lists:append(Ws1) ++ Ws2]),
    true = hash_invars(Head),
    {ok, Head}.

slots2(NoSlots) ->
    segs2slots(no_segs(NoSlots)).

init_parts(Head, PartNo, NoParts, Zero, Ws) when PartNo < NoParts ->
    PartPos = ?SEGARRADDR(PartNo),
    {NewHead, W, _Part} = alloc_part(Head, Zero, PartPos),
    init_parts(NewHead, PartNo+1, NoParts, Zero, [W | Ws]);
init_parts(Head, _PartNo, _NoParts, _Zero, Ws) ->
    {Head, Ws}.

%% -> {Head, SegInitList, OtherList}; 
%% SegPtrList = SegInitList = pwrite_list().
init_segments(Head, SegNo, NoSegs, SegZero, WsP, WsI) when SegNo < NoSegs ->
    {NewHead, WI, Ws} = allocate_segment(Head, SegZero, SegNo),
    init_segments(NewHead, SegNo+1, NoSegs, SegZero, Ws ++ WsP, [WI | WsI]);
init_segments(Head, _SegNo, _NoSegs, _SegZero, WsP, WsI) ->
    {Head, WsI, WsP}.

%% -> {NewHead, SegInit, [SegPtr | PartStuff]}
allocate_segment(Head, SegZero, SegNo) ->
    PartPos = ?SEGARRADDR(SegNo div ?SEGPARTSZ),
    case get_arrpart(PartPos) of
	undefined ->
	    %% may throw error:
	    {Head1, [InitArrPart, ArrPartPointer], Part} = 
		alloc_part(Head, SegZero, PartPos),
	    {NewHead, InitSegment, [SegPointer]} =
                alloc_seg(Head1, SegZero, SegNo, Part),
	    {NewHead, InitSegment, [SegPointer, InitArrPart, ArrPartPointer]};
	Part ->
            alloc_seg(Head, SegZero, SegNo, Part)
    end.

alloc_part(Head, PartZero, PartPos) ->
    %% may throw error:
    {NewHead, Part} = dets_utils:alloc(Head, adjsz(4 * ?SEGPARTSZ)),
    arrpart_cache(PartPos, Part),
    InitArrPart = {Part, PartZero}, % same size as segment
    ArrPartPointer = {PartPos, <<Part:32>>},
    {NewHead, [InitArrPart, ArrPartPointer], Part}.

alloc_seg(Head, SegZero, SegNo, Part) ->
    %% may throw error:
    {NewHead, Segment} = dets_utils:alloc(Head, adjsz(4 * ?SEGSZ)), 
    InitSegment = {Segment, SegZero},
    Pos = ?SEGPARTADDR(Part, SegNo),
    segp_cache(Pos, Segment),
    SegPointer = {Pos, <<Segment:32>>},
    {NewHead, InitSegment, [SegPointer]}.

%% Read free lists (using a Buddy System) from file. 
init_freelist(Head, true) ->
    Pos = Head#head.freelists_p,
    free_lists_from_file(Head, Pos).

%% -> {ok, Fd, fileheader()} | throw(Error)
read_file_header(Fd, FileName) ->
    {ok, Bin} = dets_utils:pread_close(Fd, FileName, 0, ?HEADSZ),
    <<FreeList:32,   Cookie:32,  CP:32,         Type2:32,
      Version:32,    M:32,       Next:32,       Kp:32,
      NoObjects:32,  NoKeys:32,  MinNoSlots:32, MaxNoSlots:32,
      MaxObjSize:32, N:32>> = Bin,
    {ok, EOF} = dets_utils:position_close(Fd, FileName, eof),
    {ok, <<FileSize:32>>} = dets_utils:pread_close(Fd, FileName, EOF-4, 4),
    FH = #fileheader{freelist = FreeList,
		     cookie = Cookie,
		     closed_properly = CP,
		     type = dets_utils:code_to_type(Type2),
		     version = Version,
		     m = M,
		     next = Next,
		     keypos = Kp,
		     no_objects = NoObjects,
		     no_keys = NoKeys,
		     min_no_slots = MinNoSlots,
		     max_no_slots = MaxNoSlots,
		     maxobjsize = MaxObjSize,
		     trailer = FileSize,
		     eof = EOF,
		     n = N,
		     mod = ?MODULE},
    {ok, Fd, FH}.

%% -> {ok, head(), ExtraInfo} | {error, Reason} (Reason lacking file name)
%% ExtraInfo = true
check_file_header(FH, Fd) ->
    Test = 
	if
	    FH#fileheader.cookie /= ?MAGIC ->
		{error, not_a_dets_file};
	    FH#fileheader.type == badtype ->
		{error, invalid_type_code};
	    FH#fileheader.version /= ?FILE_FORMAT_VERSION -> 
                {error, bad_version};
	    FH#fileheader.trailer /= FH#fileheader.eof ->
		{error, not_closed};
	    FH#fileheader.closed_properly == ?CLOSED_PROPERLY ->
		{ok, true};
	    FH#fileheader.closed_properly == ?NOT_PROPERLY_CLOSED ->
		{error, not_closed};
	    true ->
		{error, not_a_dets_file}
	end,
    case Test of
	{ok, ExtraInfo} ->
	    H = #head{
	      m = FH#fileheader.m,
	      next = FH#fileheader.next,
	      fptr = Fd,
	      no_objects = FH#fileheader.no_objects,
	      no_keys = FH#fileheader.no_keys,
	      maxobjsize = FH#fileheader.maxobjsize,
	      n = FH#fileheader.n,
	      type = FH#fileheader.type,
	      update_mode = saved,
	      auto_save = infinity,             % not saved on file
	      fixed = false,			% not saved on file
	      freelists_p = FH#fileheader.freelist,
	      hash_bif = phash,
	      keypos = FH#fileheader.keypos},
	    {ok, H, ExtraInfo};
	Error ->
	    Error
    end.

cache_segps(Fd, FileName, M) ->
    NoParts = no_parts(M),
    ArrStart = ?SEGARRADDR(0),
    {ok, Bin} = dets_utils:pread_close(Fd, FileName, ArrStart, 4 * NoParts),
    cache_arrparts(Bin, ?HEADEND, Fd, FileName).

cache_arrparts(<<ArrPartPos:32, B/binary>>, Pos, Fd, FileName) ->
    arrpart_cache(Pos, ArrPartPos),
    {ok, ArrPartBin} = dets_utils:pread_close(Fd, FileName, 
                                              ArrPartPos, 
                                              ?SEGPARTSZ*4),
    cache_segps(ArrPartBin, ArrPartPos),
    cache_arrparts(B, Pos+4, Fd, FileName);
cache_arrparts(B, _Pos, _Fd, _FileName) when size(B) == 0 ->
    ok.

cache_segps(<<0:32,_/binary>>, _P) ->
    ok;
cache_segps(<<S:32,B/binary>>, P) ->
    segp_cache(P, S),
    cache_segps(B, P+4);
cache_segps(B, _P) when size(B) == 0 ->
    ok.

no_parts(NoSlots) ->
    ((NoSlots - 1)  div (?SEGSZP * ?SEGPARTSZ)) + 1.

no_segs(NoSlots) ->
    ((NoSlots - 1)  div ?SEGSZP) + 1.

segs2slots(NoSegs) ->
    NoSegs * ?SEGSZP.

%%%
%%% Repair, conversion and initialization of a dets file.
%%%

%%% bulk_input/3. Initialization, the general case (any stream of objects).
%%% output_objs/4. Initialization (general case) and repair.
%%% fast_init/2. Initialization when input is a v9 table.

bulk_input(Head, InitFun, _Cntrs) ->
    bulk_input(Head, InitFun, make_ref(), 0).

bulk_input(Head, InitFun, Ref, Seq) ->
    fun(close) ->
	    _ = (catch InitFun(close));
       (read) ->
	    case catch {Ref, InitFun(read)} of
		{Ref, end_of_input} ->
		    end_of_input;
		{Ref, {L0, NewInitFun}} when list(L0), function(NewInitFun) ->
		    Kp = Head#head.keypos,
		    case catch bulk_objects(L0, Head, Kp, Seq, []) of
			{'EXIT', _Error} ->
			    _ = (catch NewInitFun(close)),
			    {error, invalid_objects_list};
			{L, NSeq} ->
			    {L, bulk_input(Head, NewInitFun, Ref, NSeq)}
		    end;
		{Ref, Value} ->
		    {error, {init_fun, Value}};
		Error ->
		    throw({thrown, Error})
	    end
    end.

bulk_objects([T | Ts], Head, Kp, Seq, L) ->
    BT = term_to_binary(T),
    Key = element(Kp, T),
    bulk_objects(Ts, Head, Kp, Seq+1, [make_object(Head, Key, Seq, BT) | L]);
bulk_objects([], _Head, _Kp, Seq, L) ->
    {L, Seq}.

-define(FSCK_SEGMENT, 1).
-define(FSCK_SEGMENT2, 10000).

-define(VEMPTY, {}).
-define(VSET(I, V, E), setelement(I, V, E)).
-define(VGET(I, V), element(I, V)).
-define(VEXT(S, V, T), 
	list_to_tuple(tuple_to_list(V) ++ lists:duplicate(S-size(V), T))).

%% {LogSize,NoObjects} in Cntrs is replaced by
%% {LogSize,Position,{FileName,FileDescriptor},NoCollections}.
%% There is also an object {no, NoObjects, NoKeys}.
-define(COUNTERS, no).
-define(OBJ_COUNTER, 2).
-define(KEY_COUNTER, 3).

output_objs(OldV, Head, SlotNums, Cntrs) when OldV =< 9 ->
    fun(close) ->
            %% Make sure that the segments are initialized in case
            %% init_table has been called.
	    Cache = ?VEMPTY,
            Acc = [], % This is the only way Acc can be empty.
            true = ets:insert(Cntrs, {?FSCK_SEGMENT,0,[],0}),
	    true = ets:insert(Cntrs, {?COUNTERS, 0, 0}),
            Fun = output_objs2(foo, Acc, OldV, Head, Cache, Cntrs,
			       SlotNums, bar),
            Fun(close);
       ([]) ->
	    output_objs(OldV, Head, SlotNums, Cntrs);
       (L) ->
	    %% Information about number of objects per size is not
	    %% relevant for version 9. It is the number of collections
	    %% that matters.
            true = ets:delete_all_objects(Cntrs),
	    true = ets:insert(Cntrs, {?COUNTERS, 0, 0}),
	    Es = bin2term(L, OldV, Head#head.keypos),
	    %% The cache is a tuple indexed by the (log) size. An element
	    %% is [BinaryObject].
	    Cache = ?VEMPTY,
	    {NE, NAcc, NCache} = output_slots(Es, Head, Cache, Cntrs, 0, 0),
	    output_objs2(NE, NAcc, OldV, Head, NCache, Cntrs, SlotNums, 1)
    end.

output_objs2(E, Acc, OldV, Head, Cache, FileT, SlotNums, 0) ->
    NCache = write_all_sizes(Cache, FileT, Head, more),
    %% Number of handled file_sorter chunks before writing:
    Max = lists:min([size(NCache),10]), 
    output_objs2(E, Acc, OldV, Head, NCache, FileT, SlotNums, Max);
output_objs2(E, Acc, OldV, Head, Cache, FileT, SlotNums, ChunkI) ->
    fun(close) ->
            {_, [], Cache1} = 
                 if
		     Acc == [] -> {foo, [], Cache};
		     true -> output_slot(Acc, Head, Cache, [], FileT, 0, 0)
		 end,
	    _NCache = write_all_sizes(Cache1, FileT, Head, no_more),
	    SegSz = actual_seg_size(),
	    {_, SegEnd} = dets_utils:alloc(Head, adjsz(SegSz)),
	    [{?COUNTERS,NoObjects,NoKeys}] = ets:lookup(FileT, ?COUNTERS),
	    Head1 = Head#head{no_objects = NoObjects, no_keys = NoKeys},
	    true = ets:delete(FileT, ?COUNTERS),
	    {NewHead, NL, _MaxSz} = allocate_all_objects(Head1, FileT),
	    %% It is not known until all objects have been collected
	    %% how many object collections there are per size. Now
	    %% that is known and the absolute positions of the object
	    %% collections can be calculated.
            segment_file(FileT, NewHead, NL, SegEnd),
	    {MinSlots, EstNoSlots, MaxSlots} = SlotNums,
	    if 
		EstNoSlots == bulk_init ->
		    {ok, 0, NewHead};
		true ->
		    EstNoSegs = no_segs(EstNoSlots),
		    MinNoSegs = no_segs(MinSlots),
		    MaxNoSegs = no_segs(MaxSlots),
		    NoSegs = no_segs(NoKeys),
		    Diff = abs(NoSegs - EstNoSegs),
		    if 
			Diff > 5, NoSegs =< MaxNoSegs, NoSegs >= MinNoSegs  ->
			    {try_again, NoKeys};
			true ->
			    {ok, 0, NewHead}
		    end
	    end;
       (L) ->
	    Es = bin2term(L, OldV, Head#head.keypos),
	    {NE, NAcc, NCache} = 
		output_slots(E, Es, Acc, Head, Cache, FileT, 0, 0),
	    output_objs2(NE, NAcc, OldV, Head, NCache, FileT, SlotNums,
			 ChunkI-1)
    end.

fast_init(Head, InitFun) ->
    FileT = ets:new(dets_init, []),
    Cache = ?VEMPTY,
    true = ets:insert(FileT, {?COUNTERS, 0, 0}),
    Input = fast_input(InitFun, FileT, Head, make_ref(), Cache, 0),
    Output = fast_output(Head, FileT),
    TmpDir = filename:dirname(Head#head.filename),
    Reply = (catch file_sorter:sort(Input, Output, 
				    [{format, binary},{tmpdir, TmpDir},
				     {header, 1}])), % make_slot/4: 13 bytes
    L = ets:tab2list(FileT),
    ets:delete(FileT),
    {Reply, lists:reverse(lists:sort(L))}.

fast_input(InitFun, FileT, Head, Ref, Cache, ASz) ->
    fun(close) ->
            _ = (catch InitFun(close));
       (read) ->
	    case catch {Ref, InitFun(read)} of
		{Ref, end_of_input} ->
		    _ = fast_write_all_sizes(Cache, FileT, Head, no_more),
		    end_of_input;
		{Ref, {L0, NInitFun}} when list(L0), function(NInitFun) ->
		    case catch make_slots(L0, Cache, FileT, Head, ASz) of
			{'EXIT', _} ->
			    _ = (catch NInitFun(close)),
			    {error, invalid_objects_list};
			{NCache, SegBs, NASz} ->
			    Fun = fast_input(NInitFun, FileT, Head, Ref, 
					     NCache, NASz),
			    {SegBs, Fun}
		    end;
		{Ref, Value} ->
		    {error, {init_fun, Value}};
		Error ->
		    throw({thrown, Error})
	    end
    end.

make_slots([], Cache, _FileT, _Head, ASz) ->
    {Cache, [], ASz};
make_slots([E | _] = Es, Cache, FileT, Head, ASz) when Head#head.type == set ->
    Key = element(Head#head.keypos, E),
    Slot = db_hash(Key, Head),
    No = length(Es),
    {C_SegBs, NSlot, Sz, Ws, NASz} = 
	make_set_slots(Es, [Cache | []], Head, Slot, 0, [], ASz),
    make_slots_end(C_SegBs, FileT, Head, NSlot, Sz, No, No, Ws, NASz);
make_slots([E | _] = Es, Cache, FileT, Head, ASz) ->
    Key = element(Head#head.keypos, E),
    Slot = db_hash(Key, Head),
    No = length(Es),
    {C_SegBs, NSlot, Sz, KNo, Ws, NASz} = 
	make_bag_slots(Es, [Cache | []], Head, Slot, 0, 1, Key, [], 0, [], ASz),
    make_slots_end(C_SegBs, FileT, Head, NSlot, Sz, KNo, No, Ws, NASz).

make_set_slots([E | Es], C_SegBs, H, Slot, Sz, Ws, ASz) ->
    BT = term_to_binary(E),
    BSz = size(BT) + 4,
    W = [<<BSz:32>> | BT],
    case db_hash(element(H#head.keypos, E), H) of
	Slot ->
	    make_set_slots(Es, C_SegBs, H, Slot, Sz + BSz, [Ws | W], ASz);
	Slot2 ->
	    NC_SegBs = make_slot(Ws, C_SegBs, Slot, Sz),
	    make_set_slots(Es, NC_SegBs, H, Slot2, BSz, W, ASz + Sz)
    end;
make_set_slots([], C_SegBs, _H, Slot, Sz, Ws, ASz) ->
    {C_SegBs, Slot, Sz, Ws, ASz}.

make_bag_slots([E | Es], C_SegBs, H, Slot, KSz, KNo, K, KWs, SSz, SWs, ASz) ->
    BT = term_to_binary(E),
    BSz = size(BT) + 4,
    W = [<<BSz:32>> | BT],
    case element(H#head.keypos, E) of
	K ->
            make_bag_slots(Es, C_SegBs, H, Slot, KSz + BSz, KNo, K, 
			   [KWs | W], SSz, SWs, ASz);
        K2 ->
            NKNo = KNo+1,
            case db_hash(K2, H) of
                Slot ->
		    KSz1 = KSz + 4,
		    NKWs = [<<KSz1:32>> | KWs],
		    NSSz = SSz + KSz1,
		    make_bag_slots(Es, C_SegBs, H, Slot, BSz, NKNo, K2, 
				   W, NSSz, [SWs | NKWs], ASz);
                Slot2 ->
		    KSz1 = KSz + 4,
		    NSSz = SSz + KSz1,
		    NSWs = [SWs, <<KSz1:32>> | KWs],
		    NC_SegBs = make_slot(NSWs, C_SegBs, Slot, NSSz),
                    make_bag_slots(Es, NC_SegBs, H, Slot2, BSz, NKNo, 
				   K2, W, 0, [], ASz + NSSz)
            end
    end;
make_bag_slots([], C_SegBs, _H, Slot, KSz, KNo, _Key, KWs, SSz, SWs, ASz) ->
    KSz1 = KSz + 4,
    NSWs = [SWs, <<KSz1:32>> | KWs],
    {C_SegBs, Slot, SSz + KSz1, KNo, NSWs, ASz}.

make_slots_end(C_SegBs, FT, H, Slot, SSz, KNo, No, Ws, ASz) ->
    _ = ets:update_counter(FT, ?COUNTERS, {?OBJ_COUNTER,No}),
    _ = ets:update_counter(FT, ?COUNTERS, {?KEY_COUNTER,KNo}),
    [C1 | NSegBs] = make_slot(Ws, C_SegBs, Slot, SSz),
    NASz = ASz + SSz,
    if 
	NASz > 10 * ?CHUNK_SIZE -> % Padding not included...
	    NC = fast_write_all_sizes(C1, FT, H, more),
	    {NC, NSegBs, 0};
	true ->
	    {C1, NSegBs, NASz}
    end.

make_slot(Ws, [Cache | SegBs], Slot, Size) ->
    BSize = Size + ?OHDSZ,
    LSize = sz2pos(BSize),
    Size2 = ?POW(LSize-1),
    Pad = <<0:(Size2-BSize)/unit:8>>,
    BinObject = [<<BSize:32, ?ACTIVE:32>>, Ws | Pad],
    if
	LSize > size(Cache) ->
	    C1 = ?VEXT(LSize, Cache, [0 | []]),
	    NCache = ?VSET(LSize, C1, [Size2 | [BinObject]]),
	    [NCache | [<<Slot:32,BSize:32,0:32,LSize:8>> | SegBs]];
	true ->
	    [Addr | L] = ?VGET(LSize, Cache),
	    NCache = ?VSET(LSize, Cache, [Addr + Size2 | [BinObject | L]]),
	    [NCache | [<<Slot:32,BSize:32,Addr:32,LSize:8>> | SegBs]]
    end.

fast_output(Head, FileT) ->
    fun(close) ->
	    {ok, 0, Head};
       (L) ->
	    SegSz = actual_seg_size(),
	    {_, SegEnd} = dets_utils:alloc(Head, adjsz(SegSz)),
	    [{?COUNTERS,NoObjects,NoKeys}] = ets:lookup(FileT, ?COUNTERS),
	    Head1 = Head#head{no_objects = NoObjects, no_keys = NoKeys},
	    true = ets:delete(FileT, ?COUNTERS),
            true = ets:insert(FileT, {?FSCK_SEGMENT,0,[],0}),
	    {NewHead, NL0, MaxSz} = allocate_all_objects(Head1, FileT),
            [{?FSCK_SEGMENT,SegAddr,[],0} | NL] = NL0,
            true = ets:delete_all_objects(FileT),
            lists:foreach(fun(X) -> true = ets:insert(FileT, X) end, NL),
	    Bases = lists:foldl(fun({LSz,P,_D,_N}, A) -> setelement(LSz,A,P) end,
				erlang:make_tuple(MaxSz, 0), NL),
	    case file:position(NewHead#head.fptr, SegAddr) of
		{ok, SegAddr} ->
		    NewSegAddr = write_segment_file(L, Bases, Head, [], 
						    SegAddr, SegAddr),
		    fast_output2(NewHead, Bases, NewSegAddr, SegAddr, SegEnd);
		Error ->
		    catch dets_utils:file_error(Error, NewHead#head.filename)
	    end
    end.

fast_output2(Head, Bases, SegAddr, SS, SegEnd) ->
    fun(close) ->
	    FinalZ = SegEnd - SegAddr,
	    dets_utils:write(Head, dets_utils:make_zeros(FinalZ)),
	    {ok, 0, Head};
       (L) ->
            NewSegAddr = write_segment_file(L, Bases, Head, [], SegAddr, SS),
            fast_output2(Head, Bases, NewSegAddr, SS, SegEnd)
    end.

%% Inlined.
write_segment_file([<<Slot:32,BSize:32,AddrToBe:32,LSize:8>> | Bins], 
		   Bases, Head, Ws, SegAddr, SS) ->
    %% Should call slot_position/1, but since all segments are
    %% allocated in a sequence, the position of a slot can be
    %% calculated faster.
    Pos = SS + ?SZOBJP*4 * Slot, % Same as {Pos, _} = slot_position(Slot).
    write_segment_file(Bins, Bases, Head, Ws, SegAddr, SS, Pos, 
		       BSize, AddrToBe, LSize);
write_segment_file([], _Bases, Head, Ws, SegAddr, _SS) ->
    dets_utils:write(Head, Ws),
    SegAddr.

write_segment_file(Bins, Bases, Head, Ws, SegAddr, SS, Pos, BSize, 
		   AddrToBe, LSize) when Pos == SegAddr ->
    Addr = AddrToBe + element(LSize, Bases),
    NWs = [Ws | <<BSize:32,Addr:32>>],
    write_segment_file(Bins, Bases, Head, NWs, SegAddr + ?SZOBJP*4, SS);
write_segment_file(Bins, Bases, Head, Ws, SegAddr, SS, Pos, BSize, 
		   AddrToBe, LSize) when Pos - SegAddr < 100 ->
    Addr = AddrToBe + element(LSize, Bases),
    NoZeros = Pos - SegAddr,
    NWs = [Ws | <<0:NoZeros/unit:8,BSize:32,Addr:32>>],
    NSegAddr = SegAddr + NoZeros + ?SZOBJP*4,
    write_segment_file(Bins, Bases, Head, NWs, NSegAddr, SS);
write_segment_file(Bins, Bases, Head, Ws, SegAddr, SS, Pos, BSize, 
		   AddrToBe, LSize) ->
    Addr = AddrToBe + element(LSize, Bases),
    NoZeros = Pos - SegAddr,
    NWs = [Ws, dets_utils:make_zeros(NoZeros) | <<BSize:32,Addr:32>>],
    NSegAddr = SegAddr + NoZeros + ?SZOBJP*4,
    write_segment_file(Bins, Bases, Head, NWs, NSegAddr, SS).

fast_write_all_sizes(Cache, FileT, Head, More) ->
    CL = tuple_to_list(Cache),
    L = lists:map(fun([_Addr | L0]) -> L0 end, CL),
    _ = write_all_sizes2(L, FileT, Head, More),
    NL = lists:map(fun([Addr | _]) -> [Addr] end, CL),
    list_to_tuple(NL).

%% By allocating bigger objects before smaller ones, holes in the
%% buddy system memory map are avoided. Unfortunately, the segments
%% are always allocated first, so if there are objects bigger than a
%% segment, there is a hole to handle. (Haven't considered placing the
%% segments among other objects of the same size.)
allocate_all_objects(Head, FileT) ->
    DTL = lists:reverse(lists:sort(ets:tab2list(FileT))),
    MaxSz = element(1, hd(DTL)),
    SegSize = actual_seg_size(),
    {Head1, HSz, HN, HA} = alloc_hole(MaxSz, Head, SegSize),
    {Head2, NL} = allocate_all(Head1, DTL, []),
    Head3 = free_hole(Head2, HSz, HN, HA),
    {Head3, NL, MaxSz}.

alloc_hole(LSize, Head, SegSz) when ?POW(LSize-1) > SegSz ->
    Size = ?POW(LSize-1),
    {_, SegAddr} = dets_utils:alloc(Head, adjsz(SegSz)),
    {_, Addr} = dets_utils:alloc(Head, adjsz(Size)),
    N = (Addr - SegAddr) div SegSz,
    Head1 = dets_utils:alloc_many(Head, SegSz, N, SegAddr),
    {Head1, SegSz, N, SegAddr};
alloc_hole(_MaxSz, Head, _SegSz) ->
    {Head, 0, 0, 0}.

free_hole(Head, _Size, 0, _Addr) ->
    Head;
free_hole(Head, Size, N, Addr) ->
    Head1 = dets_utils:free(Head, Addr, adjsz(Size)),
    free_hole(Head1, Size, N-1, Addr+Size).

%% One (temporary) file for each buddy size, write all objects of that
%% size to the file.
allocate_all(Head, [{?FSCK_SEGMENT,_,Data,_}], L) ->
    %% And one file for the segments...
    %% Note that space for the array parts and the segments has
    %% already been allocated, but the segments have not been
    %% initialized on disk.
    NoParts = no_parts(Head#head.m),
    %% All parts first, ensured by init_segments/6.
    Addr = ?BASE + NoParts * 4 * ?SEGPARTSZ,
    {Head, [{?FSCK_SEGMENT,Addr,Data,0} | L]};
allocate_all(Head, [{LSize,_,Data,NoCollections} | DTL], L) ->
    Size = ?POW(LSize-1),
    {_Head, Addr} = dets_utils:alloc(Head, adjsz(Size)),
    NewHead = dets_utils:alloc_many(Head, Size, NoCollections, Addr),
    E = {LSize,Addr,Data,NoCollections},
    allocate_all(NewHead, DTL, [E | L]).

bin2term(Bin, 9, Kp) ->
    bin2term1(Bin, Kp, []);
bin2term(Bin, 8, Kp) ->
    bin2term_v8(Bin, Kp, []).

bin2term1([<<Slot:32, Seq:32, BinTerm/binary>> | BTs], Kp, L) ->
    Term = binary_to_term(BinTerm),
    Key = element(Kp, Term),
    bin2term1(BTs, Kp, [{Slot, Key, Seq, Term, BinTerm} | L]);
bin2term1([], _Kp, L) ->
    lists:reverse(L).

bin2term_v8([<<Slot:32, BinTerm/binary>> | BTs], Kp, L) ->
    Term = binary_to_term(BinTerm),
    Key = element(Kp, Term),
    bin2term_v8(BTs, Kp, [{Slot, Key, foo, Term, BinTerm} | L]);
bin2term_v8([], _Kp, L) ->
    lists:reverse(L).

write_all_sizes(Cache, FileT, Head, More) ->
    NCL = write_all_sizes2(tuple_to_list(Cache), FileT, Head, More),
    list_to_tuple(NCL).

write_all_sizes2(CacheL, _FileT, _Head, _More) when CacheL == [] ->
    CacheL;
write_all_sizes2(CacheL0, FileT, Head, More) ->
    CacheL = lists:reverse(CacheL0),
    Sz = length(CacheL),
    case ets:info(FileT, size) of
	1 when More == no_more -> % COUNTERS only...
	    all_sizes(CacheL, Sz, FileT);
	_ ->
	    write_sizes(CacheL, Sz, FileT, Head)
    end.

all_sizes(CL, _Sz, _FileT) when CL == [] ->
    CL;
all_sizes([C | CL], Sz, FileT) when C == [] ->
    [C | all_sizes(CL, Sz-1, FileT)];
all_sizes([C0 | CL], Sz, FileT) ->
    C = lists:reverse(C0),
    NoCollections = length(C),
    true = ets:insert(FileT, {Sz,0,C,NoCollections}),
    [[] | all_sizes(CL, Sz-1, FileT)].

write_sizes(CL, _Sz, _FileT, _Head) when CL == [] ->
    CL;
write_sizes([C | CL], Sz, FileT, Head) when C == [] ->
    [C | write_sizes(CL, Sz-1, FileT, Head)];
write_sizes([C | CL], Sz, FileT, Head) ->
    {FileName, Fd} = 
	case ets:lookup(FileT, Sz) of
	    [] ->
		temp_file(Head, FileT, Sz);
	    [{_,_,{FN,F},_}] ->
		{FN, F}
	end,
    NoCollections = length(C),
    _ = ets:update_counter(FileT, Sz, {4,NoCollections}),
    case file:write(Fd, lists:reverse(C)) of
	ok ->
	    [[] | write_sizes(CL, Sz-1, FileT, Head)];
	Error ->
	    dets_utils:file_error(FileName, Error)
    end.

output_slots([E | Es], Head, Cache, FileT, NoKeys, NoObjs) ->
    output_slots(E, Es, [E], Head, Cache, FileT, NoKeys, NoObjs);
output_slots([], _Head, Cache, FileT, NoKeys, NoObjs) ->
    _ = ets:update_counter(FileT, ?COUNTERS, {?OBJ_COUNTER,NoObjs}),
    _ = ets:update_counter(FileT, ?COUNTERS, {?KEY_COUNTER,NoKeys}),
    {not_a_tuple, [], Cache}.

output_slots(E, [E1 | Es], Acc, Head, Cache, FileT, NoKeys, NoObjs) 
                       when element(1, E) == element(1, E1) ->
    output_slots(E1, Es, [E1 | Acc], Head, Cache, FileT, NoKeys, NoObjs);
output_slots(E, [], Acc, _Head, Cache, FileT, NoKeys, NoObjs) ->
    _ = ets:update_counter(FileT, ?COUNTERS, {?OBJ_COUNTER,NoObjs}),
    _ = ets:update_counter(FileT, ?COUNTERS, {?KEY_COUNTER,NoKeys}),
    {E, Acc, Cache};
output_slots(_E, L, Acc, Head, Cache, FileT, NoKeys, NoObjs) ->
    output_slot(Acc, Head, Cache, L, FileT, NoKeys, NoObjs).

output_slot(Es, Head, Cache, L, FileT, NoKeys, NoObjs) ->
    Slot = element(1, hd(Es)),
    {Bins, Size, No, KNo} = prep_slot(lists:sort(Es), Head),
    NNoKeys = NoKeys + KNo,
    NNoObjs = NoObjs + No,

    %% First the object collection.
    BSize = Size + ?OHDSZ,
    LSize = sz2pos(BSize),
    Size2 = ?POW(LSize-1),
    Pad = <<0:(Size2-BSize)/unit:8>>,
    BinObject = [<<BSize:32, ?ACTIVE:32>>, Bins | Pad],
    Cache1 = 
	if
	    LSize > size(Cache) ->
                C1 = ?VEXT(LSize, Cache, []),
		?VSET(LSize, C1, [BinObject]);
	    true ->
		CL = ?VGET(LSize, Cache),
		?VSET(LSize, Cache, [BinObject | CL])
	end,

    %% Then the pointer to the object collection.
    %% Cannot yet determine the absolute pointers; segment_file/5 does that.
    PBin = <<Slot:32,BSize:32,LSize:8>>,
    PL = ?VGET(?FSCK_SEGMENT, Cache1),
    NCache = ?VSET(?FSCK_SEGMENT, Cache1, [PBin | PL]),
    output_slots(L, Head, NCache, FileT, NNoKeys, NNoObjs).

prep_slot(L, Head) when Head#head.type =/= set ->
    prep_slot(L, Head, []);
prep_slot([{_Slot,Key,_Seq,_T,BT} | L], _Head) ->
    prep_set_slot(L, Key, BT, 0, 0, 0, []).

prep_slot([{_Slot, Key, Seq, T, _BT} | L], Head, W) ->
    prep_slot(L, Head, [{Key, {Seq, {insert,T}}} | W]);
prep_slot([], Head, W) ->
    WLs = sofs:to_external(sofs:relation_to_family(sofs:relation(W, 2))),
    {[], Bins, Size, No, KNo, _} = 
	eval_slot(WLs, [], Head#head.type, [], [], 0, 0, 0, false),
    {Bins, Size, No, KNo}.

%% Optimization, prep_slot/3 would work for set tables as well.
prep_set_slot([{_,K,_Seq,_T1,BT1} | L], K, _BT, Sz, NoKeys, NoObjs, Ws) ->
    prep_set_slot(L, K, BT1, Sz, NoKeys, NoObjs, Ws);
prep_set_slot([{_,K1,_Seq,_T1,BT1} | L], _K, BT, Sz, NoKeys, NoObjs, Ws) ->
    BSize = size(BT) + 4,
    NWs = [Ws,<<BSize:32>>|BT],
    prep_set_slot(L, K1, BT1, Sz+BSize, NoKeys+1, NoObjs+1, NWs);
prep_set_slot([], _K, BT, Sz, NoKeys, NoObjs, Ws) ->
    BSize = size(BT) + 4,
    {[Ws, <<BSize:32>> | BT], Sz + BSize, NoKeys+1, NoObjs+1}.

segment_file(FileT, Head, FileData, SegEnd) ->
    I = 2,
    true = ets:delete_all_objects(FileT),    
    lists:foreach(fun(X) -> true = ets:insert(FileT, X) end, FileData),
    [{?FSCK_SEGMENT,SegAddr,Data,0} | FileData1] = FileData,
    NewData = 
	case Data of
	    {InFile,In0} ->
		{OutFile, Out} = temp_file(Head, FileT, I),
		file:close(In0),
		{ok, In} = dets_utils:open(InFile, [raw,binary,read]),
		{ok, 0} = dets_utils:position(In, InFile, bof),
		seg_file(SegAddr, SegAddr, In, InFile, Out, OutFile, FileT, 
			 SegEnd),
		file:close(In),
		file:delete(InFile),
		{OutFile,Out};
	    Objects ->
		{LastAddr, B} = seg_file(Objects, SegAddr, SegAddr, FileT, []),
		FinalZ = SegEnd - LastAddr,
		[B | dets_utils:make_zeros(FinalZ)]
	end,
    %% Restore the positions.
    true = ets:delete_all_objects(FileT),
    %% To get the segments copied first by dets:fsck_copy/4, use a big
    %% number here, FSCK_SEGMENT2.
    lists:foreach(fun(X) -> true = ets:insert(FileT, X) end, 
		  [{?FSCK_SEGMENT2,SegAddr,NewData,0} | FileData1]),
    ok.
    
seg_file(Addr, SS, In, InFile, Out, OutFile, FileT, SegEnd) ->
    case dets_utils:read_n(In, 4500) of
	eof ->
	    FinalZ = SegEnd - Addr,
	    dets_utils:fwrite(Out, OutFile, dets_utils:make_zeros(FinalZ));
	Bin ->
	    {NewAddr, L} = seg_file(Bin, Addr, SS, FileT, []),
	    ok = dets_utils:fwrite(Out, OutFile, L),
	    seg_file(NewAddr, SS, In, InFile, Out, OutFile, FileT, SegEnd)
    end.

seg_file(<<Slot:32,BSize:32,LSize:8,T/binary>>, Addr, SS, FileT, L) ->
    seg_file_item(T, Addr, SS, FileT, L, Slot, BSize, LSize);
seg_file([<<Slot:32,BSize:32,LSize:8>> | T], Addr, SS, FileT, L) ->
    seg_file_item(T, Addr, SS, FileT, L, Slot, BSize, LSize);
seg_file(Bin, Addr, _SS, _FileT, L) when size(Bin) == 0; Bin == [] ->
    {Addr, lists:reverse(L)}.

seg_file_item(T, Addr, SS, FileT, L, Slot, BSize, LSize) ->
    %% Should call slot_position/1, but since all segments are
    %% allocated in a sequence, the position of a slot can be
    %% calculated faster.
    SlotPos = SS + ?SZOBJP*4 * Slot, % {SlotPos, _} = slot_position(Slot)
    NoZeros = SlotPos - Addr,
    PSize = NoZeros+?SZOBJP*4,
    Inc = ?POW(LSize-1),
    CollP = ets:update_counter(FileT, LSize, Inc) - Inc,
    PointerBin = if 
		     NoZeros =:= 0 ->
			 <<BSize:32, CollP:32>>;
		     NoZeros > 100 ->
			 [dets_utils:make_zeros(NoZeros) | 
			  <<BSize:32, CollP:32>>];
		     true ->
			 <<0:NoZeros/unit:8, BSize:32, CollP:32>>
                 end,
    seg_file(T, Addr + PSize, SS, FileT, [PointerBin | L]).

temp_file(Head, FileT, N) ->
    TmpName = lists:concat([Head#head.filename, '.', N]),
    {ok, Fd} = dets_utils:open(TmpName, [raw, binary, write]),
    %% The file table is consulted when cleaning up.
    true = ets:insert(FileT, {N,0,{TmpName,Fd},0}),
    {TmpName, Fd}.

%% Does not close Fd.
fsck_input(Head, Fd, Cntrs) ->
    State0 = fsck_read(?BASE, Fd, [], 0),
    fsck_input(Head, State0, Fd, Cntrs).

fsck_input(Head, State, Fd, Cntrs) ->
    fun(close) ->
	    ok;
       (read) ->
	    case State of
		done ->
		    end_of_input;
		{done, L, _Seq} ->
		    R = count_input(Head, Cntrs, L),
		    {R, fsck_input(Head, done, Fd, Cntrs)};
		{cont, L, Bin, Pos, Seq} ->
		    R = count_input(Head, Cntrs, L),
		    NewState = fsck_read(Bin, Pos, Fd, Head, [], Seq),
		    {R, fsck_input(Head, NewState, Fd, Cntrs)}
	    end
    end.

%% The ets table Cntrs is used for counting objects per size.
count_input(Head, Cntrs, L) when Head#head.version == 8 ->
    count_input1(Cntrs, L, []);
count_input(_Head, _Cntrs, L) ->
    lists:reverse(L).

count_input1(Cntrs, [[LogSz | B] | Ts], L) ->
    case catch ets:update_counter(Cntrs, LogSz, 1) of
	N when integer(N) -> ok;
	_Badarg -> true = ets:insert(Cntrs, {LogSz, 1})
    end,
    count_input1(Cntrs, Ts, [B | L]);
count_input1(_Cntrs, [], L) ->
    L.

fsck_read(Pos, F, L, Seq) ->
    case file:position(F, Pos) of
	{ok, _} ->
	    read_more_bytes([], 0, Pos, F, L, Seq);
	_Error ->
	    {done, L, Seq}
    end.

fsck_read(Bin, Pos, F, Head, L, Seq) ->
    case fsck_objs(Bin, Head#head.keypos, Head, L, Seq) of
        {more, NewBin, Sz, NL, NSeq} ->
            read_more_bytes(NewBin, Sz, Pos, F, NL, NSeq);
        {new, Skip, NL, NSeq} ->
            NewPos = Pos + Skip,
            fsck_read(NewPos, F, NL, NSeq)
    end.

read_more_bytes(B, Min, Pos, F, L, Seq) ->
    Max = if 
	      Min < ?CHUNK_SIZE -> ?CHUNK_SIZE; 
	      true -> Min 
	  end,
    case dets_utils:read_n(F, Max) of
	eof ->
	    {done, L, Seq};
	Bin ->
	    NewPos = Pos + size(Bin),
	    {cont, L, list_to_binary([B, Bin]), NewPos, Seq}
    end.

fsck_objs(Bin = <<Sz:32, Status:32, Tail/binary>>, Kp, Head, L, Seq) ->
    if 
	Status == ?ACTIVE ->
	    Sz1 = Sz-?OHDSZ,
	    case Tail of
		<<BinTerm:Sz1/binary, Tail2/binary>> ->
		    case catch bin2keybins(BinTerm, Head) of
			{'EXIT', _Reason} ->
			    %% The whole collection of objects is skipped.
			    skip_bytes(Bin, ?BUMP, Kp, Head, L, Seq);
			BOs ->
			    {NL, NSeq} = make_objects(BOs, Seq, Kp, Head, L),
			    Skip = ?POW(sz2pos(Sz)-1) - Sz,
			    skip_bytes(Tail2, Skip, Kp, Head, NL, NSeq)
		    end;
		_ ->
                    {more, Bin, Sz, L, Seq}
	    end;
	true -> 
	    skip_bytes(Bin, ?BUMP, Kp, Head, L, Seq)
    end;
fsck_objs(Bin, _Kp, _Head, L, Seq) ->
    {more, Bin, 0, L, Seq}.
    
make_objects([{K,BT}|Os], Seq, Kp, Head, L) when Head#head.version == 8 ->
    LogSz = dets_v8:sz2pos(size(BT)+?OHDSZ_v8),
    Slot = dets_v8:db_hash(K, Head),
    Obj = [LogSz | <<Slot:32, LogSz:8, BT/binary>>],
    make_objects(Os, Seq, Kp, Head, [Obj | L]);
make_objects([{K,BT} | Os], Seq, Kp, Head, L) ->
    Obj = make_object(Head, K, Seq, BT),
    make_objects(Os, Seq+1, Kp, Head, [Obj | L]);
make_objects([], Seq, _Kp, _Head, L) ->
    {L, Seq}.

%% Inlined.
make_object(Head, Key, Seq, BT) ->
    Slot = db_hash(Key, Head),
    <<Slot:32, Seq:32, BT/binary>>.

%% Inlined.
skip_bytes(Bin, Skip, Kp, Head, L, Seq) ->
    case Bin of
	<<_:Skip/binary, Tail/binary>> ->
	    fsck_objs(Tail, Kp, Head, L, Seq);
	_ ->
            {new, Skip - size(Bin), L, Seq}
    end.

%%%
%%% End of repair, conversion and initialization of a dets file.
%%%

%% -> {NewHead, ok} | throw({Head, Error})
do_perform_save(H) ->
    {ok, FreeListsPointer} = dets_utils:position(H, eof),
    H1 = H#head{freelists_p = FreeListsPointer},
    {FLW, FLSize} = free_lists_to_file(H1),
    FileSize = FreeListsPointer + FLSize + 4,
    ok = dets_utils:write(H, [FLW | <<FileSize:32>>]),
    Header = file_header(H1, FreeListsPointer, ?CLOSED_PROPERLY),
    W = {0, Header},
    dets_utils:pwrite(H1, [W]).

file_header(Head, FreeListsPointer, ClosedProperly) ->
    Cookie = ?MAGIC,
    TypeCode = dets_utils:type_to_code(Head#head.type),
    Version = ?FILE_FORMAT_VERSION,
    <<FreeListsPointer:32,
      Cookie:32,
      ClosedProperly:32,
      TypeCode:32,
      Version:32,
      (Head#head.m):32, 
      (Head#head.next):32, 
      (Head#head.keypos):32,
      (Head#head.no_objects):32,
      (Head#head.no_keys):32,
      (Head#head.min_no_slots):32,
      (Head#head.max_no_slots):32,
      (Head#head.maxobjsize):32,
      (Head#head.n):32>>.

%% Going through some trouble to avoid creating one single binary for
%% the free lists. If the free lists are huge, binary_to_term and
%% term_to_binary could otherwise stop the emulator for quite some time.

-define(MAXFREEOBJ, 4096).
-define(ENDFREE, 12345).

free_lists_to_file(H) ->
    FL = dets_utils:get_freelists(H),
    free_list_to_file(FL, H, 1, size(FL), [], 0).

free_list_to_file(_Ftab, _H, Pos, Sz, Ws, WsSz) when Pos > Sz ->
    {[Ws | <<(4+?OHDSZ):32, ?FREE:32, ?ENDFREE:32>>], WsSz+4+?OHDSZ};
free_list_to_file(Ftab, H, Pos, Sz, Ws, WsSz) ->
    Max = (?MAXFREEOBJ - 4 - ?OHDSZ) div 4,
    F = fun(N, L, W, S) when N == 0 -> {N, L, W, S};
	   (N, L, W, S) ->
		{L1, N1, More} =
		    if 
			N > Max ->
			    {lists:sublist(L, Max), Max,
			     {N-Max, lists:nthtail(Max, L)}};
			true ->
			    {L, N, no_more}
		    end,
                Size = N1*4 + 4 + ?OHDSZ,
		Header = <<Size:32, ?FREE:32, Pos:32>>,
		NW = [W, Header | L1],
		case More of
		    no_more ->
			{0, [], NW, S+Size};
		    {NN, NL} ->
			ok = dets_utils:write(H, NW),
			{NN, NL, [], S+Size}
		end
	end,
    {NWs,NWsSz} = dets_utils:tree_to_bin(element(Pos, Ftab), F, Max, Ws, WsSz),
    free_list_to_file(Ftab, H, Pos+1, Sz, NWs, NWsSz).

free_lists_from_file(H, Pos) ->
    dets_utils:position(H#head.fptr, H#head.filename, Pos),
    FL = dets_utils:empty_free_lists(),
    case catch bin_to_tree([], H, start, FL, -1, []) of
	{'EXIT', _} ->
	    throw({error, {bad_freelists, H#head.filename}});
	Reply ->
	    Reply
    end.

bin_to_tree(Bin, H, LastPos, Ftab, A0, L) ->
    case Bin of 
        <<_Size:32,?FREE:32,?ENDFREE:32,_/binary>> when L == [] ->
            Ftab;
        <<_Size:32,?FREE:32,?ENDFREE:32,_/binary>> ->
            setelement(LastPos, Ftab, dets_utils:list_to_tree(L));
        <<Size:32,?FREE:32,Pos:32,T/binary>> when size(T) >= Size-4-?OHDSZ ->
	    {NFtab, L1, A1} = 
		if
		    Pos =/= LastPos, LastPos =/= start ->
			Tree = dets_utils:list_to_tree(L),
			{setelement(LastPos, Ftab, Tree), [], -1};
		    true ->
			{Ftab, L, A0}
		    end,
	    {NL, B2, A2} = bin_to_tree1(T, Size-?OHDSZ-4, A1, L1),
	    bin_to_tree(B2, H, Pos, NFtab, A2, NL);
        _ ->
            Bin2 = dets_utils:read_n(H#head.fptr, ?MAXFREEOBJ),
            bin_to_tree(list_to_binary([Bin | Bin2]), H, LastPos, Ftab, A0, L)
    end.

bin_to_tree1(<<A1:32,A2:32,A3:32,A4:32,T/binary>>, Size, A, L) 
         when Size >= 16, A < A1, A1 < A2, A2 < A3, A3 < A4 ->
    bin_to_tree1(T, Size-16, A4, [A4, A3, A2, A1 | L]);
bin_to_tree1(<<A1:32,T/binary>>, Size, A, L) when Size >= 4, A < A1 ->
    bin_to_tree1(T, Size - 4, A1, [A1 | L]);
bin_to_tree1(B, 0, A, L) ->
    {L, B, A}.

%% -> [term()] | throw({Head, Error})
slot_objs(H, Slot) when Slot >= H#head.next ->
    '$end_of_table';
slot_objs(H, Slot) ->
    {ok, _Pointer, Objects} = slot_objects(H, Slot),
    Objects.

%% Inlined.
h(I) -> erlang:phash(I, ?BIG) - 1.

db_hash(Key, Head) ->
    H = h(Key),
    Hash = H rem Head#head.m,
    if
	Hash < Head#head.n ->
	    H rem (2 * Head#head.m);
	true ->
	    Hash
    end.

no_slots(Head) ->
    {Head#head.min_no_slots, Head#head.next, Head#head.max_no_slots}.

table_parameters(Head) ->
    {?HASH_TAG, Head#head.version, Head#head.n, Head#head.m, Head#head.next,
     Head#head.min_no_slots, Head#head.max_no_slots}.

%% Allow quite a lot when reading object collections.
-define(MAXCOLL, (10 * ?CHUNK_SIZE)).

%% Re-hashing a segment, starting with SlotStart.
%%
%% On the average, half of the keys of the slot are put in a new slot.
%% If the old slot is i, then the new slot is i+m. The new slots
%% reside in a newly allocated segment.
%%
%% -> {NewHead, ok} | throw({Head, Error})
re_hash(Head, SlotStart) ->
    {FromSlotPos, _} = slot_position(SlotStart),
    {ToSlotPos, _} = slot_position(SlotStart + Head#head.m),
    RSpec = [{FromSlotPos, 4 * ?SEGSZ}],
    {ok, [FromBin]} = dets_utils:pread(RSpec, Head),
    split_bins(FromBin, Head, FromSlotPos, ToSlotPos, [], [], 0).

split_bins(FB, Head, _Pos1, _Pos2, _ToRead, _L, 0) when size(FB) == 0 ->
    {Head, ok};
split_bins(FB, Head, Pos1, Pos2, ToRead, L, _SoFar) when size(FB) == 0 ->
    re_hash_write(Head, ToRead, L, Pos1, Pos2);
split_bins(FB, Head, Pos1, Pos2, ToRead, L, SoFar) ->
    <<Sz1:32, P1:32, FT/binary>> = FB,
    <<B1:?OHDSZ/binary, _/binary>> = FB,
    NSoFar = SoFar + Sz1,
    NPos1 = Pos1 + ?SZOBJP*4,
    NPos2 = Pos2 + ?SZOBJP*4,
    if
	NSoFar > ?MAXCOLL, ToRead =/= [] ->
	    {NewHead, ok} = re_hash_write(Head, ToRead, L, Pos1, Pos2),
	    split_bins(FB, NewHead, Pos1, Pos2, [], [], 0);
	Sz1 == 0 ->
	    E = {skip,B1},
	    split_bins(FT, Head, NPos1, NPos2, ToRead, [E | L], NSoFar);
        true ->
	    E = {Sz1,P1,B1,Pos1,Pos2},
	    NewToRead = [{P1,Sz1} | ToRead],
	    split_bins(FT, Head, NPos1, NPos2, NewToRead, [E | L], NSoFar)
    end.

re_hash_write(Head, ToRead, L, Pos1, Pos2) ->
    {ok, Bins} = dets_utils:pread(ToRead, Head),
    Z = <<0:32, 0:32>>,
    {Head1, BinFS, BinTS, WsB} = re_hash_slots(Bins, L, Head, Z, [],[],[]),
    WPos1 = Pos1 - ?SZOBJP*4*length(L),
    WPos2 = Pos2 - ?SZOBJP*4*length(L),
    ToWrite = [{WPos1,BinFS}, {WPos2, BinTS} | WsB],
    dets_utils:pwrite(Head1, ToWrite).

re_hash_slots(Bins, [{skip,B1} | L], Head, Z, BinFS, BinTS, WsB) ->
    re_hash_slots(Bins, L, Head, Z, [B1 | BinFS], [Z | BinTS], WsB);
re_hash_slots([FB | Bins], [E | L], Head, Z, BinFS, BinTS, WsB) ->
    {Sz1,P1,B1,Pos1,Pos2} = E,
    KeyObjs = case catch per_key(Head, FB) of
		  {'EXIT', _Error} ->
		      throw(dets_utils:corrupt_reason(Head, bad_object));
		  Else ->
		      Else
	      end,
    case re_hash_split(KeyObjs, Head, [], 0, [], 0) of
	{_, _, [], 0} ->
	    re_hash_slots(Bins, L, Head, Z, [B1 | BinFS], [Z | BinTS], WsB);
	{[], 0, _ML, _MSz} -> %% Optimization.
	    re_hash_slots(Bins, L, Head, Z, [Z | BinFS], [B1 | BinTS], WsB);
	{KL, KSz, ML, MSz} ->
	    {Head1, FS1, Ws1} = updated(Head, P1, Sz1, KSz, Pos1, KL, true),
	    {NewHead, TS2, Ws2} = updated(Head1, 0, 0, MSz, Pos2, ML, true),
	    NewBinFS = case FS1 of
			   [{Pos1,Bin1}] -> [Bin1 | BinFS];
			   [] -> BinFS
		       end,
	    NewBinTS = case TS2 of
			   [{Pos2,Bin2}] -> [Bin2 | BinTS];
			   [] -> BinTS
		       end,
	    NewWsB = Ws2 ++ Ws1 ++ WsB,
	    re_hash_slots(Bins, L, NewHead, Z, NewBinFS, NewBinTS, NewWsB)
    end;
re_hash_slots([], [], Head, _Z, BinFS, BinTS, WsB) ->
    {Head, BinFS, BinTS, lists:reverse(WsB)}.

re_hash_split([E | KeyObjs], Head, KL, KSz, ML, MSz) ->
    {Key,Sz,Bin,_Item,_Objs} = E,
    New = h(Key) rem (2 * Head#head.m),
    if
	New >= Head#head.m ->
	    re_hash_split(KeyObjs, Head, KL, KSz, [Bin | ML], MSz + Sz);
	true ->
	    re_hash_split(KeyObjs, Head, [Bin | KL], KSz + Sz, ML, MSz)
    end;
re_hash_split([], _Head, KL, KSz, ML, MSz) ->
    {lists:reverse(KL), KSz, lists:reverse(ML), MSz}.

%% -> {NewHead, [LookedUpObject], pwrite_list()} | throw({NewHead, Error})
write_cache(Head) ->
    C = Head#head.cache,
    case dets_utils:is_empty_cache(C) of
	true -> {Head, [], []};
	false ->
	    {NewC, MaxInserts, PerKey} = dets_utils:reset_cache(C),
	    %% MaxNoInsertedKeys is an upper limit on the number of new keys.
	    MaxNoInsertedKeys = lists:min([MaxInserts, length(PerKey)]),
	    Head1 = Head#head{cache = NewC},
	    case may_grow(Head1, MaxNoInsertedKeys, once) of
		{Head2, ok} ->
		    eval_work_list(Head2, PerKey);
		HeadError ->
		    throw(HeadError)
	    end
    end.

%% -> {NewHead, ok} | {NewHead, Error}
may_grow(Head, _N, _How) when Head#head.fixed =/= false ->
    {Head, ok};
may_grow(Head, _N, _How) when Head#head.next >= Head#head.max_no_slots ->
    {Head, ok};
may_grow(Head, N, How) ->
    Extra = lists:min([2*?SEGSZP, Head#head.no_keys + N - Head#head.next]),
    case catch may_grow1(Head, Extra, How) of
	{error, _Reason} = Error -> % alloc may throw error
	    dets_utils:corrupt(Head, Error);
	Reply ->
	    Reply
    end.

may_grow1(Head, Extra, many_times) when Extra > ?SEGSZP ->
    Reply = grow(Head, 1, undefined),
    case Reply of
	{_, ok} -> self() ! ?DETS_CALL(self(), may_grow);
	_ -> ok
    end,
    Reply;
may_grow1(Head, Extra, _How) ->    
    grow(Head, Extra, undefined).

%% -> {Head, ok} | throw({Head, Error})
grow(Head, Extra, _SegZero) when Extra =< 0 ->
    {Head, ok};
grow(Head, Extra, undefined) ->
    grow(Head, Extra, seg_zero());
grow(Head, Extra, SegZero) ->
    #head{n = N, next = Next, m = M} = Head,
    SegNum = Next div ?SEGSZP,    
    {Head0, W, Ws1} = allocate_segment(Head, SegZero, SegNum),
    %% re_hash/2 will overwrite the segment, but initialize it anyway...
    {Head1, ok} = dets_utils:pwrite(Head0, [W | Ws1]),
    %% If re_hash fails, segp_cache has been called, but it does not matter.
    {Head2, ok} = re_hash(Head1, N),
    NewHead =
	if 
	    N + ?SEGSZP == M ->
		Head2#head{n = 0, next = Next + ?SEGSZP, m = 2 * M};
	    true ->
		Head2#head{n = N + ?SEGSZP, next = Next + ?SEGSZP}
	end,
    true = hash_invars(NewHead),
    grow(NewHead, Extra - ?SEGSZP, SegZero).

hash_invars(H) ->
    hash_invars(H#head.n, H#head.m, H#head.next, H#head.min_no_slots, 
		H#head.max_no_slots).

-define(M8(X), (((X) band (?SEGSZP - 1)) == 0)).
hash_invars(N, M, Next, Min, Max) ->
        ?M8(N) and ?M8(M) and ?M8(Next) and ?M8(Min) and ?M8(Max)
    and (0 =< N) and (N =< M) and (N =< 2*Next) and (M =< Next)
    and (Next =< 2*M) and (0 =< Min) and (Min =< Next) and (Next =< Max) 
    and (Min =< M).

seg_zero() ->
    <<0:(4*?SEGSZ)/unit:8>>.

find_object(Head, Object) ->
    Key = element(Head#head.keypos, Object),
    Slot = db_hash(Key, Head),
    find_object(Head, Object, Slot).    

find_object(H, _Obj, Slot) when Slot >= H#head.next ->
    false;
find_object(H, Obj, Slot) ->
    case catch slot_objects(H, Slot) of
	{ok, Pointer, Objects} ->
	    case lists:member(Obj, Objects) of
		true -> {ok, Pointer};
		false -> false
	    end;
	_ -> false
    end.

%% -> {ok, BucketP, Objects} | throw({Head, Error})
slot_objects(Head, Slot) ->
    {SlotPos, _Size} = slot_position(Slot),    
    case dets_utils:ipread(Head#head.fptr, SlotPos, infinity) of 
	{ok, {BucketSz, Pointer, <<BucketSz:32, _St:32, KeysObjs/binary>>}} ->
	    case catch bin2objs(KeysObjs, Head#head.type, []) of
		{'EXIT', _Error} ->
		    throw(dets_utils:corrupt_reason(Head, bad_object));
		{Objs, _NoObjs} ->
		    {ok, Pointer, Objs}
	    end;
        [] ->
	    {ok, 0, []};
	_Bad -> % eof or bad badly formed binary
	    throw(dets_utils:corrupt_reason(Head, bad_object))
    end.

%%%
%%% Cache routines depending on the dets file format.
%%%

%% -> {Head, [LookedUpObject], pwrite_list()} | throw({Head, Error})
eval_work_list(Head, [{Key,[{_Seq,{lookup,Pid}}]}]) ->
    {SlotPos, _} = slot_position(db_hash(Key, Head)),
    Objs = case dets_utils:ipread(Head#head.fptr, SlotPos, infinity) of 
	       {ok, {_BucketSz, _Pointer, Bin}} ->
		   case lists:keysearch(Key, 1, per_key(Head, Bin)) of
		       false ->
			   [];
		       {value, {_Key,_KS,_KB,O,Os}} ->
			   get_objects([O | binobjs2terms(Os)])
		   end;
	       [] ->
		   [];
	       _Bad -> % eof or bad badly formed binary
		   throw(dets_utils:corrupt_reason(Head, bad_object))
	   end,
    {Head, [{Pid,Objs}], []};
eval_work_list(Head, PerKey) ->
    SWLs = tag_with_slot(PerKey, Head, []),
    P1 = sofs:to_external(sofs:relation_to_family(sofs:relation(SWLs, 2))),
    {PerSlot, SlotPositions} = remove_slot_tag(P1, [], []),
    {ok, Bins} = dets_utils:pread(SlotPositions, Head),
    read_buckets(PerSlot, SlotPositions, Bins, Head, [], [], [], [], 0, 0, 0).

tag_with_slot([{K,_} = WL | WLs], Head, L) ->
    tag_with_slot(WLs, Head, [{db_hash(K, Head), WL} | L]);
tag_with_slot([], _Head, L) ->
    L.

remove_slot_tag([{S,SWLs} | SSWLs], Ls, SPs) ->
    remove_slot_tag(SSWLs, [SWLs | Ls], [slot_position(S) | SPs]);
remove_slot_tag([], Ls, SPs) ->
    {Ls, SPs}.

read_buckets([WLs | SPs], [{P1,_8} | Ss], [<<_Zero:32,P2:32>> | Bs], Head,
	      PWLs, ToRead, LU, Ws, NoObjs, NoKeys, SoFar) when P2 == 0 ->
    {NewHead, NLU, NWs, No, KNo} = 
	eval_bucket_keys(WLs, P1, 0, 0, [], Head, Ws, LU),
    NewNoObjs = No + NoObjs,
    NewNoKeys = KNo + NoKeys,
    read_buckets(SPs, Ss, Bs, NewHead, PWLs, ToRead, NLU, NWs, 
		 NewNoObjs, NewNoKeys, SoFar);
read_buckets([WorkLists| SPs], [{P1,_8} | Ss], [<<Size:32,P2:32>> | Bs], Head,
	     PWLs, ToRead, LU, Ws, NoObjs, NoKeys, SoFar) 
                                 when SoFar + Size < ?MAXCOLL; ToRead == [] ->
    NewToRead = [{P2, Size} | ToRead],
    NewPWLs = [{P1,P2,WorkLists} | PWLs],
    NewSoFar = SoFar + Size,
    read_buckets(SPs, Ss, Bs, Head, NewPWLs, NewToRead, LU, Ws, 
		 NoObjs, NoKeys, NewSoFar);
read_buckets(SPs, Ss, Bs, Head, PWLs, ToRead, LU, Ws, NoObjs, NoKeys, SoFar) 
                                                             when SoFar > 0 ->
    {ok, Bins} = dets_utils:pread(ToRead, Head),
    case catch eval_buckets(Bins, PWLs, Head, LU, Ws, 0, 0) of
	{ok, NewHead, NLU, [], 0, 0} -> 
            read_buckets(SPs, Ss, Bs, NewHead, [], [], NLU, [], 
			 NoObjs, NoKeys, 0);
	{ok, Head1, NLU, NWs, No, KNo} -> 
	    NewNoObjs = NoObjs + No,
	    NewNoKeys = NoKeys + KNo,
            {NewHead, ok} = dets_utils:pwrite(Head1, lists:reverse(NWs)),
            read_buckets(SPs, Ss, Bs, NewHead, [], [], NLU, [], 
			 NewNoObjs, NewNoKeys, 0);
	_Error  -> 
	    throw(dets_utils:corrupt_reason(Head, bad_object))
    end;
read_buckets([], [], [], Head, [], [], LU, Ws, NoObjs, NoKeys, 0) ->
    {NewHead, NWs} = update_no_keys(Head, Ws, NoObjs, NoKeys),
    {NewHead, LU, lists:reverse(NWs)}.

eval_buckets([Bin | Bins], [SP | SPs], Head, LU, Ws, NoObjs, NoKeys) ->
    {P1, Pos, WLs} = SP,
    KeyObjs = per_key(Head, Bin),
    {NewHead, NLU, NWs, No, KNo} = 
	eval_bucket_keys(WLs, P1, Pos, size(Bin), KeyObjs, Head, Ws, LU),
    eval_buckets(Bins, SPs, NewHead, NLU, NWs, NoObjs + No, NoKeys + KNo);
eval_buckets([], [], Head, LU, Ws, NoObjs, NoKeys) ->
    {ok, Head, LU, Ws, NoObjs, NoKeys}.

eval_bucket_keys(WLs, SlotPos, Pos, OldSize, KeyObjs, Head, Ws, LU) ->
    {NLU, Bins, BSize, No, KNo, Ch} = 
         eval_slot(WLs, KeyObjs, Head#head.type, LU, [], 0, 0, 0, false),
    {NewHead, W1, W2} = updated(Head, Pos, OldSize, BSize, SlotPos, Bins, Ch),
    {NewHead, NLU, W1++W2++Ws, No, KNo}.

updated(Head, Pos, OldSize, BSize, SlotPos, Bins, Ch) ->
    BinsSize = BSize + ?OHDSZ,
    if 
	Pos == 0, BSize == 0 ->
	    {Head, [], []};
	Pos == 0, BSize > 0 ->
	    {NewHead, NewPos} = dets_utils:alloc(Head, adjsz(BinsSize)),
	    W1 = {NewPos, [<<BinsSize:32, ?ACTIVE:32>> | Bins]},
	    W2 = {SlotPos, <<BinsSize:32, NewPos:32>>},
	    {NewHead, [W2], [W1]};
	Pos =/= 0, BSize == 0 ->
	    NewHead = dets_utils:free(Head, Pos, adjsz(OldSize)),
	    W1 = {Pos+?STATUS_POS, <<?FREE:32>>},
	    W2 = {SlotPos, <<0:32, 0:32>>},
	    {NewHead, [W2], [W1]};
	Pos =/= 0, BSize > 0, Ch == false ->
	    {Head, [], []};
	Pos =/= 0, BSize > 0 ->
	    %% Doubtful. The scan function has to be careful since
	    %% partly scanned objects may be overwritten.
	    Overwrite = if
			    OldSize == BinsSize -> same;
			    true -> sz2pos(OldSize) == sz2pos(BinsSize)
			end,
	    if 
		Overwrite == same ->
		    W1 = {Pos+?OHDSZ, Bins},
		    {Head, [], [W1]};
		Overwrite == true ->
		    W1 = {Pos, [<<BinsSize:32, ?ACTIVE:32>> | Bins]},
		    %% Pos is already there, but return {SlotPos, <8 bytes>}.
		    W2 = {SlotPos, <<BinsSize:32, Pos:32>>},
		    {Head, [W2], [W1]};
		true ->
		    Head1 = dets_utils:free(Head, Pos, adjsz(OldSize)),
		    {NewHead, NewPos} = 
			dets_utils:alloc(Head1, adjsz(BinsSize)),
		    W0 = {NewPos, [<<BinsSize:32, ?ACTIVE:32>> | Bins]},
		    W2 = {SlotPos, <<BinsSize:32, NewPos:32>>},
		    W1 = if 
			      Pos =/= NewPos ->
                                  %% W0 first.
				  [W0, {Pos+?STATUS_POS, <<?FREE:32>>}];
			      true -> 
				  [W0]
			  end,
		    {NewHead, [W2], W1}
	    end
    end.

eval_slot([{Key,Commands} | WLs] = WLs0, KOs, Type, LU, Ws, No, KNo,BSz, Ch) ->
    case KOs of
        [{K,KS,KB,O,Os} | KOs1] when K == Key ->
            Old = [O | binobjs2terms(Os)],
            {NLU, NWs, Sz, No1, KNo1, NCh} = 
		eval_key(Key, Commands, Old, Type, KB, KS, LU, Ws, Ch),
            eval_slot(WLs, KOs1, Type, NLU, NWs, No1 + No, 
		      KNo1 + KNo, Sz + BSz, NCh);
        [{K,Size,KeyBin,_,_} | KOs1] when K < Key ->
            eval_slot(WLs0, KOs1, Type, LU, [Ws | KeyBin], No, 
		      KNo, Size + BSz, Ch);
        _ ->
            {NLU, NWs, Sz, No1, KNo1, NCh} = 
		eval_key(Key, Commands, [], Type, [], 0, LU, Ws, Ch),
            eval_slot(WLs, KOs, Type, NLU, NWs, No1 + No, 
		      KNo1 + KNo, Sz + BSz, NCh)
    end;
eval_slot([], [{_Key,Size,KeyBin,_,_} | KOs], Type, LU, Ws, No, KNo,BSz, Ch) ->
    eval_slot([], KOs, Type, LU, [Ws | KeyBin], No, KNo, Size + BSz, Ch);
eval_slot([], [], _Type, LU, Ws, No, KNo, BSz, Ch) ->
    {LU, Ws, BSz, No, KNo, Ch}.

eval_key(_K, [{_Seq,{lookup,Pid}}], [], _Type, _KeyBin, _KeySz, LU, Ws, Ch) ->
    NLU = [{Pid, []} | LU],
    {NLU, Ws, 0, 0, 0, Ch};    
eval_key(_K, [{_Seq,{lookup,Pid}}], Old, _Type, KeyBin, KeySz, LU, Ws, Ch) ->
    Objs = get_objects(lists:keysort(2, Old)),  
    NLU = [{Pid, Objs} | LU],
    {NLU, [Ws | KeyBin], KeySz, 0, 0, Ch};    
eval_key(K, Comms, Orig, Type, KeyBin, KeySz, LU, Ws, Ch) ->
    Old = lists:sort(Orig),
    case eval_key1(Comms, [], Old, Type, K, LU, Ws, 0, Orig) of
	{ok, NLU} when Old == [] ->
	    {NLU, Ws, 0, 0, 0, Ch};
	{ok, NLU} ->
	    {NLU, [Ws | KeyBin], KeySz, 0, 0, Ch};
	{NLU, NWs, NSz, No} when Old == [], NSz > 0 ->
	    {NLU, NWs, NSz, No, 1, true};
	{NLU, NWs, NSz, No} when Old =/= [], NSz == 0 ->
	    {NLU, NWs, NSz, No, -1, true};
	{NLU, NWs, NSz, No} ->
	    {NLU, NWs, NSz, No, 0, true}
    end.

%% First find 'delete_key' and 'lookup' commands, and handle the 'set' type.
eval_key1([{_Seq,{insert,Term}} | L], Cs, [{Term,_,_}] = Old, Type, K,
	  LU, Ws, No, Orig) when Type == set ->
    eval_key1(L, Cs, Old, Type, K, LU, Ws, No, Orig);
eval_key1([{Seq,{insert,Term}} | L], Cs, Old, Type, K, LU, Ws, No, Orig) 
                                         when Type == set ->
    NNo = No + 1 - length(Old),
    eval_key1(L, Cs, [{Term,Seq,insert}], Type, K, LU, Ws, NNo, Orig);
eval_key1([{_Seq,{lookup,Pid}} | L], Cs, Old, Type, Key, LU, Ws, No, Orig) ->
    {ok, New, NewNo} = eval_comms(Cs, Old, Type, No),
    Objs = get_objects(lists:keysort(2, New)),
    NLU = [{Pid, Objs} | LU],
    if 
	L == [] ->
	    eval_end(New, NLU, Type, Ws, NewNo, Orig);
	true ->
	    NewOld = lists:sort(New),
	    eval_key1(L, Cs, NewOld, Type, Key, NLU, Ws, NewNo, Orig)
    end;
eval_key1([{_Seq,delete_key} | L], _Cs, Old, Type, K, LU, Ws, No, Orig) ->
    NewNo = No - length(Old),
    eval_key1(L, [], [], Type, K, LU, Ws, NewNo, Orig);
eval_key1([{_Seq,{delete_object,Term}} | L], Cs, [{Term,_,_}], Type, K,
	  LU, Ws, No, Orig) when Type == set ->
    eval_key1(L, Cs, [], Type, K, LU, Ws, No-1, Orig);
eval_key1([{_Seq,{delete_object,_T}}| L], Cs, Old1, Type, K, LU, 
	      Ws, No, Orig) when Type == set ->
    eval_key1(L, Cs, Old1, Type, K, LU, Ws, No, Orig);
eval_key1([{Seq,{Comm,Term}} | L], Cs, Old, Type, K, LU, Ws, No, Orig) 
                                         when Type =/= set ->
    eval_key1(L, [{Term,Seq,Comm} | Cs], Old, Type, K, LU, Ws, No, Orig);
eval_key1([], Cs, Old, Type, _Key, LU, Ws, No, Orig) when Type == set ->
    [] = Cs,
    eval_end(Old, LU, Type, Ws, No, Orig);
eval_key1([], Cs, Old, Type, _Key, LU, Ws, No, Orig) ->
    {ok, New, NewNo} = eval_comms(Cs, Old, Type, No),
    eval_end(New, LU, Type, Ws, NewNo, Orig).

eval_comms([], L, Type, No) when Type == set ->
    {ok, L, No};
eval_comms(Cs, Old, Type, No) ->
    Commands = lists:sort(Cs),
    case Type of
	bag -> eval_bag(Commands, Old, [], No);
	duplicate_bag -> eval_dupbag(Commands, Old, [], No)
    end.

eval_end(New0, LU, Type, Ws, NewNo, Orig) ->
    New = lists:keysort(2, New0),
    NoChange = if
		   length(New) =/= length(Orig) -> false;
		   true -> 
		       same_terms(Orig, New)
	       end,
    if 
        NoChange == true ->
	    %% The key's objects have not changed.
	    {ok, LU};
	New == [] ->
	    {LU, Ws, 0, NewNo};
	true -> 
	    {Ws1, Sz} = make_bins(New, [], 0),
	    if 
		Type == set ->
		    {LU, [Ws | Ws1], Sz, NewNo};
		true -> 
		    NSz = Sz + 4,
		    {LU, [Ws, <<NSz:32>> | Ws1], NSz, NewNo}
	    end
    end.

same_terms([E1 | L1], [E2 | L2]) when element(1, E1) == element(1, E2) ->
    same_terms(L1, L2);
same_terms(L1, L2) ->
    (L1 == []) and (L2 == []).

make_bins([{_Term,_Seq,B} | L], W, Sz) when binary(B) ->
    make_bins(L, [W | B], Sz + size(B));
make_bins([{Term,_Seq,insert} | L], W, Sz) ->
    B = term_to_binary(Term),
    BSize = size(B) + 4,
    make_bins(L, [W, [<<BSize:32>> | B]], Sz + BSize);
make_bins([], W, Sz) ->
    {W, Sz}.

get_objects([{T,_S,_BT} | L]) ->
    [T | get_objects(L)];
get_objects([]) ->
    [].

eval_bag([{Term1,_Seq1,delete_object} | L], Old, New, No) 
                       when Old == []; Term1 < element(1, hd(Old)) ->
    eval_bag(L, Old, New, No);
eval_bag([{Term,_Seq1,insert} = N | L], Old, New, No) 
                       when Old == []; Term < element(1, hd(Old)) ->
    bag_object(L, Old, New, No, [N], Term);
eval_bag(L, [{Term2,_S,_B} = O | Old], New, No) 
                       when L == []; element(1, hd(L)) > Term2 ->
    eval_bag(L, Old, [O | New], No);
eval_bag([{Term,_,insert} = N | L], [{Term,_,_} | Old], New, No) ->
    bag_object(L, Old, New, No-1, [N], Term);
eval_bag([{Term,_,delete_object} | L], [{Term,_,_} | Old], New, No) ->
    bag_object(L, Old, New, No-1, [], Term);
eval_bag([], [], New, No) ->
    {ok, New, No}.

bag_object([{Term,_,insert} = N | L], Old, New, No, _N, Term) ->
    bag_object(L, Old, New, No, [N], Term);
bag_object([{Term,_,delete_object} | L], Old, New, No, _N, Term) ->
    bag_object(L, Old, New, No, [], Term);
bag_object(L, Old, New, No, [], _Term) ->
    eval_bag(L, Old, New, No);
bag_object(L, Old, New, No, [N], _Term) ->
    eval_bag(L, Old, [N | New], No+1).

eval_dupbag([{Term1,_Seq1,delete_object} | L], Old, New, No) 
                       when Old == []; Term1 < element(1, hd(Old)) ->
    eval_dupbag(L, Old, New, No);
eval_dupbag([{Term,_Seq1,insert} = N | L], Old, New, No) 
                       when Old == []; Term < element(1, hd(Old)) ->
    dup_object(L, Old, New, No+1, Term, [N]);
eval_dupbag(L, [{Term2,_S,_B} = O | Old], New, No) 
                       when L == []; element(1, hd(L)) > Term2 ->
    eval_dupbag(L, Old, [O | New], No);
eval_dupbag([{Term,_,_} | _] = L, [{Term,_,_} = Obj | Old], New, No) ->
    old_dup_object(L, Old, New, No, Term, [Obj]);
eval_dupbag([], [], New, No) ->
    {ok, New, No}.
    
old_dup_object(L, [{Term,_,_} = Obj | Old], New, No, Term, N) ->
    old_dup_object(L, Old, New, No, Term, [Obj | N]);
old_dup_object(L, Old, New, No, Term, N) ->
    dup_object(L, Old, New, No, Term, N).

dup_object([{Term,_,insert} = Obj | L], Old, New, No, Term, Q) ->
    dup_object(L, Old, New, No+1, Term, [Obj | Q]);
dup_object([{Term,_Seq,delete_object} | L], Old, New, No, Term, Q) ->
    %% All objects are deleted.
    NewNo = No - length(Q),
    dup_object(L, Old, New, NewNo, Term, []);
dup_object(L, Old, New, No, _Term, Q) ->
    eval_dupbag(L, Old, Q ++ New, No).

%% Update no_keys on the file too, if the number of segments that
%% dets:fsck/6 uses for estimate has changed.
update_no_keys(Head, Ws, 0, 0) -> {Head, Ws};
update_no_keys(Head, Ws, DeltaObjects, DeltaKeys) ->
    NoKeys = Head#head.no_keys,
    NewNoKeys = NoKeys + DeltaKeys,
    NWs = 
	if 
	    NewNoKeys > Head#head.max_no_slots ->
		Ws;
	    NoKeys div ?SEGSZP == NewNoKeys div ?SEGSZP ->
		Ws;
	    true ->
		[{?NO_KEYS_POS, <<NewNoKeys:32>>} | Ws]
	end,
    NewNoObject = Head#head.no_objects + DeltaObjects,
    {Head#head{no_objects = NewNoObject, no_keys = NewNoKeys}, NWs}.

slot_position(S) ->
    SegNo = S div ?SEGSZP,
    PartPos = ?SEGARRADDR(SegNo div ?SEGPARTSZ),
    Part = get_arrpart(PartPos),
    Pos = ?SEGPARTADDR(Part, SegNo),
    Segment = get_segp(Pos),
    SegObjSize = 4 * ?SZOBJP,
    FinalPos = Segment + (SegObjSize * (S rem ?SEGSZP)),
    {FinalPos, SegObjSize}.

%% Inlined.
actual_seg_size() ->
    ?SEGSZ*4.

%% Inlined.
segp_cache(Pos, Segment) ->
    put(Pos, Segment).

%% Inlined.
get_segp(Pos) ->
    get(Pos).

arrpart_cache(Pos, ArrPart) ->
    put(Pos, ArrPart).

%% Inlined.
get_arrpart(Pos) ->
    get(Pos).

sz2pos(N) ->
    1 + dets_utils:log2(N).

%% Inlined. Compensates for the bug in dets_utils:sz2pos/1.
adjsz(N) ->
    N-1.

scan_objs(Bin, From, To, L, Ts, -1, _C) ->
    {stop, Bin, From, To, L, Ts};
scan_objs(B = <<Size:32, St:32, KO/binary>>, From, To, L, Ts, R, C) ->
    if 
	St == ?ACTIVE;
	St == ?FREE -> % deleted after scanning started
	    Size1 = Size-?OHDSZ,
	    case KO of
		<<KeysObjs:Size1/binary, T/binary>> ->
                    case catch to_objects(KeysObjs, C, Ts) of
                        {'EXIT', _} ->
                            bad_object;
                        {NTs, NoObjs} -> 
			    Skip = ?POW(sz2pos(Size)-1) - Size,
                            F2 = From + Size,
			    NR = if 
				     R < 0, R + NoObjs > -1 ->
					 -1;
                                     R < 0 ->
                                         R + NoObjs;
				     true ->
					 R + Size + Skip
				 end,
			    scan_skip(T, F2, To, Skip, L, NTs, NR, C)
		    end;
		_ ->
                    {more, B, From, To, L, Ts, R, Size}
	    end;
	true -> % a segment
	    scan_skip(B, From, To, actual_seg_size(), L, Ts, R, C)
    end;
scan_objs(B, From, To, L, Ts, R, _C) ->
    {more, B, From, To, L, Ts, R, 0}.

%% Inlined.
scan_skip(Bin, From, To, Skip, L, Ts, R, C) when From + Skip == To ->
    scan_next_allocated(Bin, From, L, Ts, R, C);
scan_skip(Bin, From, To, Skip, L, Ts, R, C) ->
    SkipPos = From + Skip,
    case Bin of
	<<_:Skip/binary, Tail/binary>> ->
	    scan_objs(Tail, SkipPos, To, L, Ts, R, C);
	_ ->
            {read, SkipPos, To, L, Ts, R}
    end.

scan_next_allocated(_Bin, _From, L, Ts, _R, _C) when size(L) == 0 ->
    {finished, Ts, L};
scan_next_allocated(Bin, From0, <<From:32, To:32, L/binary>>, Ts, R, C) ->
    Skip = From - From0,
    scan_skip(Bin, From0, To, Skip, L, Ts, R, C).

%% Inlined.
to_objects(Bin, {binary, _, set}, Ts) ->
    <<ObjSz0:32, T/binary>> = Bin,
    ObjSz = ObjSz0-4,
    case size(T) of
	SizeT when SizeT > ObjSz ->
	    <<Obj:ObjSz/binary, NObjSz:32, T1/binary>> = T,
	    bin2bins1(T1, [Obj], Ts, 2, NObjSz-4);
	ObjSz ->
	    {[T | Ts], 1}
    end;
to_objects(Bin, {binary, _, _Bag}, Ts) ->
    bin2bins2(Bin, [], Ts, 0);
to_objects(Bin, {_Repr, _, Type}, Ts) -> 
    bin2objs(Bin, Type, Ts).

%%%%%%%%%%%%%%%%%  DEBUG functions %%%%%%%%%%%%%%%%

file_info(FH) ->
    #fileheader{closed_properly = CP, keypos = Kp,
                m = M, next = Next, n = N, version = Version,
                type = Type, no_objects = NoObjects, no_keys = NoKeys} 
        = FH,
    if
        CP == 0 ->
            {error, not_closed};
        FH#fileheader.cookie /= ?MAGIC ->
            {error, not_a_dets_file};
        FH#fileheader.version /= ?FILE_FORMAT_VERSION ->
            {error, bad_version};
        true ->
            {ok, [{closed_properly,CP},{keypos,Kp},{m, M},{n,N},
		  {next,Next},{no_objects,NoObjects},{no_keys,NoKeys},
                  {type,Type},{version,Version}]}
    end.

v_segments(H) ->
    v_parts(H, 0, 0).

v_parts(_H, ?SEGARRSZ, _SegNo) ->
    done;
v_parts(H, PartNo, SegNo) ->
    Fd = H#head.fptr,
    PartPos = dets_utils:read_4(Fd, ?SEGARRADDR(PartNo)),
    if
	PartPos == 0 ->
	    done;
	true ->
	    PartBin = dets_utils:pread_n(Fd, PartPos, ?SEGPARTSZ*4),
	    v_segments(H, PartBin, PartNo+1, SegNo)
    end.

v_segments(H, B, PartNo, SegNo) when size(B) == 0 ->
    v_parts(H, PartNo, SegNo);
v_segments(_H, <<0:32,_/binary>>, _PartNo, _SegNo) ->
    done;
v_segments(H, <<Seg:32,T/binary>>, PartNo, SegNo) ->
    io:format("<~w>SEGMENT ~w~n", [Seg, SegNo]),
    v_segment(H, SegNo, Seg, 0),
    v_segments(H, T, PartNo, SegNo+1).

v_segment(_H, _, _SegPos, ?SEGSZP) ->
    done;
v_segment(H, SegNo, SegPos, SegSlot) ->
    Slot = SegSlot + (SegNo * ?SEGSZP),
    BucketP = SegPos + (4 * ?SZOBJP * SegSlot),
    case catch read_bucket(H#head.fptr, BucketP, H#head.type) of
	{'EXIT', Reason} -> 
	    dets_utils:vformat("** dets: Corrupt or truncated dets file~n", 
			       []), 
	    io:format("~nERROR ~p~n", [Reason]);
	[] ->  %% don't print empty buckets
	    true;
	{Size, CollP, Objects} ->
	    io:format("   <~w>~w: <~w:~p>~w~n", 
		      [BucketP, Slot, CollP, Size, Objects])
    end,
    v_segment(H, SegNo, SegPos, SegSlot+1).

%% -> [] | {Pointer, [object()]} | throw(EXIT)
read_bucket(Fd, Position, Type) ->
    case dets_utils:ipread(Fd, Position, infinity) of
	{ok, {Size, Pointer, <<Size:32, _Status:32, KeysObjs/binary>>}} ->
	    {Objs, _NoObjs} = bin2objs(KeysObjs, Type, []),
	    {Size, Pointer, Objs};
	[] ->
	    []
    end.
     
-define(SEQSTART, -(1 bsl 26)).

%% -> [{Key,SizeOfWholeKey,WholeKeyBin,FirstObject,OtherObjects}] |throw(EXIT)
%% FirstObject = {Term, Seq, Binary}
%% Seq < 0 (and ascending).
per_key(Head, <<BinSize:32, ?ACTIVE:32, Bin/binary>> = B) ->
    true = size(B) == BinSize,
    if 
	Head#head.type == set ->
	    per_set_key(Bin, Head#head.keypos, []);
	true ->
	    per_bag_key(Bin, Head#head.keypos, [])
    end.

per_set_key(<<Size:32, T/binary>> = B, KeyPos, L) ->
    <<KeyBin:Size/binary, R/binary>> = B,
    Size1 = Size - 4,   
    <<Obj:Size1/binary, _/binary>> = T,
    Term = binary_to_term(Obj),
    Key = element(KeyPos, Term),
    Item = {Term, ?SEQSTART, KeyBin},
    per_set_key(R, KeyPos, [{Key,Size,KeyBin,Item,[]} | L]);
per_set_key(B, _KeyPos, L) when size(B) == 0 ->
    lists:reverse(L).

per_bag_key(<<Size:32, ObjSz:32, T/binary>> = B, KeyPos, L) ->
    <<KeyBin:Size/binary, R/binary>> = B,
    ObjSz1 = ObjSz - 4, 
    Size1 = Size - ObjSz - 4,
    <<Obj:ObjSz1/binary, KeyObjs:Size1/binary, _/binary>> = T,
    <<_Size:32, Bin:ObjSz/binary, _/binary>> = B,
    Term = binary_to_term(Obj),
    Key = element(KeyPos, Term),
    Item = {Term, ?SEQSTART, Bin},
    per_bag_key(R, KeyPos, [{Key,Size,KeyBin,Item,KeyObjs} | L]);
per_bag_key(B, _KeyPos, L) when size(B) == 0 ->
    lists:reverse(L).


binobjs2terms(<<ObjSz:32, _/binary>> = B) ->
    binobjs2terms(B, ObjSz, size(B)-ObjSz, ?SEQSTART+1, []);
binobjs2terms(B) when B == [] ->
    B;
binobjs2terms(B) when size(B) == 0 ->
    [].

binobjs2terms(Bin, ObjSz, Size, N, L) when Size == 0 ->
    ObjSz1 = ObjSz-4,
    <<_ObjSz:32, Obj:ObjSz1/binary>> = Bin,
    lists:reverse(L, [{binary_to_term(Obj), N, Bin}]);
binobjs2terms(Bin, ObjSz, Size, N, L) ->
    <<B:ObjSz/binary, T/binary>> = Bin,
    ObjSz1 = ObjSz-4,
    <<_ObjSz:32, Obj:ObjSz1/binary>> = B,
    <<NObjSz:32, _/binary>> = T,
    Item = {binary_to_term(Obj), N, B},
    binobjs2terms(T, NObjSz, Size-NObjSz, N+1, [Item | L]).


bin2objs(KeysObjs, set, Ts) ->
    <<ObjSz:32, T/binary>> = KeysObjs,
    bin2objs(T, ObjSz-4, size(KeysObjs)-ObjSz, [], Ts, 0);
bin2objs(KeysObjs, _Type, Ts) ->
    bin2objs2(KeysObjs, [], Ts, 0).

bin2objs2(<<Size:32, ObjSz:32, T/binary>>, L, Ts, NoObjs) ->
    bin2objs(T, ObjSz-4, Size-ObjSz-4, L, Ts, NoObjs);
bin2objs2(B, L, Ts, NoObjs) when size(B) == 0 ->
    {lists:reverse(L, Ts), NoObjs}.

bin2objs(Bin, ObjSz, Size, L, Ts, NoObjs) when Size == 0 ->
    <<Obj:ObjSz/binary, T/binary>> = Bin,
    bin2objs2(T, [binary_to_term(Obj) | L], Ts, NoObjs + 1);
bin2objs(Bin, ObjSz, Size, L, Ts, NoObjs) ->
    <<Obj:ObjSz/binary, NObjSz:32, T/binary>> = Bin,
    bin2objs(T, NObjSz-4, Size-NObjSz, [binary_to_term(Obj) | L], Ts, NoObjs+1).


bin2bins1(Bin, L, Ts, NoObjs, ObjSz) when size(Bin) > ObjSz ->
    <<Obj:ObjSz/binary, NObjSz:32, T/binary>> = Bin,
    bin2bins1(T, [Obj | L], Ts, NoObjs + 1, NObjSz-4);
bin2bins1(Bin, [Bin1], Ts, NoObjs, _ObjSz) ->
    {[Bin1, Bin | Ts], NoObjs};
bin2bins1(Bin, L, Ts, NoObjs, _ObjSz) ->
    {lists:reverse(L, [Bin | Ts]), NoObjs}.

bin2bins2(<<Size:32, ObjSz:32, T/binary>>, L, Ts, NoObjs) ->
    bin2bins(T, L, Ts, NoObjs, Size-ObjSz-4, ObjSz-4);
bin2bins2(B, L, Ts, NoObjs) when size(B) == 0 ->
    {lists:reverse(L, Ts), NoObjs}.

bin2bins(Bin, L, Ts, NoObjs, Size, ObjSz) when Size == 0 ->
    <<Obj:ObjSz/binary, T/binary>> = Bin,
    bin2bins2(T, [Obj | L], Ts, NoObjs + 1);
bin2bins(Bin, L, Ts, NoObjs, Size, ObjSz) ->
    <<Obj:ObjSz/binary, NObjSz:32, T/binary>> = Bin,
    bin2bins(T, [Obj | L], Ts, NoObjs + 1, Size-NObjSz, NObjSz-4).


bin2keybins(KeysObjs, Head) when Head#head.type == set ->
    <<ObjSz:32, T/binary>> = KeysObjs,
    bin2keybins(T, Head#head.keypos, ObjSz-4, size(KeysObjs)-ObjSz, []);
bin2keybins(KeysObjs, Head) ->
    bin2keybins2(KeysObjs, Head#head.keypos, []).

bin2keybins2(<<Size:32, ObjSz:32, T/binary>>, Kp, L) ->
    bin2keybins(T, Kp, ObjSz-4, Size-ObjSz-4, L);
bin2keybins2(B, _Kp, L) when size(B) == 0 ->
    lists:reverse(L).

bin2keybins(Bin, Kp, ObjSz, Size, L) when Size == 0 ->
    <<Obj:ObjSz/binary, T/binary>> = Bin,
    Term = binary_to_term(Obj),
    bin2keybins2(T, Kp, [{element(Kp, Term),Obj} | L]);
bin2keybins(Bin, Kp, ObjSz, Size, L) ->
    <<Obj:ObjSz/binary, NObjSz:32, T/binary>> = Bin,
    Term = binary_to_term(Obj),
    bin2keybins(T, Kp, NObjSz-4, Size-NObjSz, [{element(Kp,Term),Obj} | L]).
